/* Custom BDD Routines by Alan J. Hu, copyright Alan Hu 1997 */


#include "bddint.h"


static
bdd
#if defined(__STDC__)
bdd_bounded_and_step(bdd_manager bddm, bdd f, bdd g, long *size_bound)
#else
bdd_bounded_and_step(bddm, f, g, size_bound)
     bdd_manager bddm;
     bdd f;
     bdd g;
     long *size_bound;
#endif
{
  bdd_indexindex_type top_indexindex;
  bdd f1, f2;
  bdd g1, g2;
  bdd temp1, temp2;
  bdd result;

  BDD_SETUP(f);
  BDD_SETUP(g);

  /*fprintf(stderr,"[%d]",*size_bound); fflush(stderr);*/
  if (*size_bound <= 0) return (bdd)0; /* AJH */

  if (BDD_IS_CONST(f))
    {
      if (f == BDD_ZERO(bddm))
	return (f);
      BDD_TEMP_INCREFS(g);
      return (g);
    }
  /* f is not constant. */
  if (BDD_IS_CONST(g))
    {
      if (g == BDD_ZERO(bddm))
	return (g);
      BDD_TEMP_INCREFS(f);
      return (f);
    }
  /* f and g are not constant. */
  if (BDD_SAME_OR_NEGATIONS(f, g))
    {
      if (f == g)
	{
	  BDD_TEMP_INCREFS(f);
	  return (f);
	}
      return (BDD_ZERO(bddm));
    }
  /* f and g are not constant and are not equal or negations. */
  if (BDD_OUT_OF_ORDER(f, g))
    BDD_SWAP(f, g);
  if (bdd_lookup_in_cache31(bddm, CACHE_TYPE_ITE, (long)f, (long)g, (long)BDD_ZERO(bddm), (long *)&result))
    return (result);
  BDD_TOP_VAR2(top_indexindex, bddm, f, g);
  BDD_COFACTOR(top_indexindex, f, f1, f2);
  BDD_COFACTOR(top_indexindex, g, g1, g2);
  temp1=bdd_bounded_and_step(bddm, f1, g1, size_bound);
  if (temp1==(bdd)0) return (bdd)0;
  temp2=bdd_bounded_and_step(bddm, f2, g2, size_bound);
  if (temp2==(bdd)0) {
      BDD_SETUP(temp1);
      BDD_TEMP_DECREFS(temp1);
      return (bdd)0;
  }
  {
      long old_count = bddm->unique_table.entries;
      result=bdd_find(bddm, top_indexindex, temp1, temp2);
      if (old_count < bddm->unique_table.entries) {
	    (*size_bound)--; /* New node must have been allocated. --AJH */
      }
  }
  bdd_insert_in_cache31(bddm, CACHE_TYPE_ITE, (long)f, (long)g, (long)BDD_ZERO(bddm), (long)result);
  if (*size_bound < 0) {
	BDD_SETUP(result);
	BDD_TEMP_DECREFS(result);
	return (bdd)0;
  } else return (result);
}


static
bdd
#if defined(__STDC__)
bdd_bounded_xnor_step(bdd_manager bddm, bdd f, bdd g, long *size_bound)
#else
bdd_bounded_xnor_step(bddm, f, g, size_bound)
     bdd_manager bddm;
     bdd f;
     bdd g;
     long *size_bound;
#endif
{
  int outneg;
  bdd_indexindex_type top_indexindex;
  bdd f1, f2;
  bdd g1, g2;
  bdd temp1, temp2;
  bdd result;

  BDD_SETUP(f);
  BDD_SETUP(g);

  /*fprintf(stderr,"[%d]",*size_bound); fflush(stderr);*/
  if (*size_bound <= 0) return (bdd)0; /* AJH */

  if (BDD_IS_CONST(f))
    {
      BDD_TEMP_INCREFS(g);
      if (f == BDD_ONE(bddm))
	return (g);
      return (BDD_NOT(g));
    }
  if (BDD_IS_CONST(g))
    {
      BDD_TEMP_INCREFS(f);
      if (g == BDD_ONE(bddm))
	return (f);
      return (BDD_NOT(f));
    }
  /* f and g are not constant. */
  if (BDD_SAME_OR_NEGATIONS(f, g))
    {
      if (f == g)
	return (BDD_ONE(bddm));
      return (BDD_ZERO(bddm));
    }
  /* f and g are not constant, not same, not negations. */
  if (BDD_OUT_OF_ORDER(f, g))
    BDD_SWAP(f, g);
  if (BDD_IS_OUTPOS(g))
    outneg=0;
  else
    {
      outneg=1;
      g=BDD_NOT(g);
    }
  /* g is an uncomplemented output pointer. */
  if (bdd_lookup_in_cache31(bddm, CACHE_TYPE_ITE, (long)f, (long)g, (long)BDD_NOT(g), (long *)&result))
    return (outneg ? BDD_NOT(result) : result);
  BDD_TOP_VAR2(top_indexindex, bddm, f, g);
  BDD_COFACTOR(top_indexindex, f, f1, f2);
  BDD_COFACTOR(top_indexindex, g, g1, g2);
  temp1=bdd_bounded_xnor_step(bddm, f1, g1, size_bound);
  if (temp1==(bdd)0) return (bdd)0;
  temp2=bdd_bounded_xnor_step(bddm, f2, g2, size_bound);
  if (temp2==(bdd)0) {
      BDD_SETUP(temp1);
      BDD_TEMP_DECREFS(temp1);
      return (bdd)0;
  }
  {
      long old_count = bddm->unique_table.entries;
      result=bdd_find(bddm, top_indexindex, temp1, temp2);
      if (old_count < bddm->unique_table.entries) {
	    (*size_bound)--; /* New node must have been allocated. --AJH */
      }
  }
  bdd_insert_in_cache31(bddm, CACHE_TYPE_ITE, (long)f, (long)g, (long)BDD_NOT(g), (long)result);
  if (*size_bound < 0) {
	BDD_SETUP(result);
	BDD_TEMP_DECREFS(result);
	return (bdd)0;
  } else return (outneg ? BDD_NOT(result) : result);
}


bdd
#if defined(__STDC__)
bdd_bounded_ite_step(bdd_manager bddm, bdd f, bdd g, bdd h, long *size_bound)
#else
bdd_bounded_ite_step(bddm, f, g, h, size_bound)
     bdd_manager bddm;
     bdd f;
     bdd g;
     bdd h;
     long *size_bound;
#endif
{
  int outneg;
  bdd_indexindex_type top_indexindex;
  bdd f1, f2;
  bdd g1, g2;
  bdd h1, h2;
  bdd temp1, temp2;
  bdd result;

  BDD_SETUP(f);
  BDD_SETUP(g);
  BDD_SETUP(h);

  /*fprintf(stderr,"[%d]",*size_bound); fflush(stderr);*/
  if (*size_bound <= 0) return (bdd)0; /* AJH */

  if (BDD_IS_CONST(f))
    {
      if (f == BDD_ONE(bddm))
	{
	  BDD_TEMP_INCREFS(g);
	  return (g);
	}
      BDD_TEMP_INCREFS(h);
      return (h);
    }
  /* f is not constant. */
  if (BDD_SAME_OR_NEGATIONS(f, g))
    {
      if (f == g)
	g=BDD_ONE(bddm);
      else
	g=BDD_ZERO(bddm);
      BDD_RESET(g);
    }
  if (BDD_SAME_OR_NEGATIONS(f, h))
    {
      if (f == h)
	h=BDD_ZERO(bddm);
      else
	h=BDD_ONE(bddm);
      BDD_RESET(h);
    }
  if (BDD_IS_CONST(g))
    {
      if (BDD_IS_CONST(h))
	{
	  if (g == h)
	    return (g);
	  BDD_TEMP_INCREFS(f);
	  if (g == BDD_ONE(bddm))
	    return (f);
	  return (BDD_NOT(f));
	}
      if (g == BDD_ZERO(bddm))
	return (bdd_bounded_and_step(bddm, BDD_NOT(f), h, size_bound));
      temp1 = bdd_bounded_and_step(bddm, BDD_NOT(f), BDD_NOT(h), size_bound);
      if (temp1==(bdd)0) return (bdd)0;
      return (BDD_NOT(temp1));
    }
  else if (BDD_SAME_OR_NEGATIONS(g, h))
    {
      if (g == h)
	{
	  BDD_TEMP_INCREFS(g);
	  return (g);
	}
      return (bdd_bounded_xnor_step(bddm, f, g, size_bound));
    }
  else if (BDD_IS_CONST(h))
    if (h == BDD_ZERO(bddm))
      return (bdd_bounded_and_step(bddm, f, g, size_bound));
    else {
      temp1 = bdd_bounded_and_step(bddm, f, BDD_NOT(g), size_bound);
      if (temp1==(bdd)0) return (bdd)0;
      return (BDD_NOT(temp1));
    }
  /* No special cases; it's a real if-then-else. */
  if (!BDD_IS_OUTPOS(f))
    {
      f=BDD_NOT(f);
      BDD_SWAP(g, h);
    }
  /* f is now an uncomplemented output pointer. */
  if (BDD_IS_OUTPOS(g))
    outneg=0;
  else
    {
      outneg=1;
      g=BDD_NOT(g);
      h=BDD_NOT(h);
    }
  /* g is now an uncomplemented output pointer. */
  if (bdd_lookup_in_cache31(bddm, CACHE_TYPE_ITE, (long)f, (long)g, (long)h, (long *)&result))
    return (outneg ? BDD_NOT(result) : result);
  BDD_TOP_VAR3(top_indexindex, bddm, f, g, h);
  BDD_COFACTOR(top_indexindex, f, f1, f2);
  BDD_COFACTOR(top_indexindex, g, g1, g2);
  BDD_COFACTOR(top_indexindex, h, h1, h2);
  temp1=bdd_bounded_ite_step(bddm, f1, g1, h1, size_bound);
  if (temp1==(bdd)0) return (bdd)0;
  temp2=bdd_bounded_ite_step(bddm, f2, g2, h2, size_bound);
  if (temp2==(bdd)0) {
      BDD_SETUP(temp1);
      BDD_TEMP_DECREFS(temp1);
      return (bdd)0;
  }
  {
      long old_count = bddm->unique_table.entries;
      result=bdd_find(bddm, top_indexindex, temp1, temp2);
      if (old_count < bddm->unique_table.entries) {
	    (*size_bound)--; /* New node must have been allocated. --AJH */
      }
  }
  bdd_insert_in_cache31(bddm, CACHE_TYPE_ITE, (long)f, (long)g, (long)h, (long)result);
  if (*size_bound < 0) {
	BDD_SETUP(result);
	BDD_TEMP_DECREFS(result);
	return (bdd)0;
  } else return (outneg ? BDD_NOT(result) : result);
}


/* bdd_bounded_ite(bddm, f, g, h, size) returns the BDD for
				"if f then g else h". */
/* HOWEVER, it tries not to exceed size bound.  If the BDD being built
	becomes too large, it aborts the operation and returns 0.
	Note that the size bound is only used approximately, to keep
	things fast (and also because I can't figure out a reliable
	way to determine if bdd_find allocated a new node without
	modifying bdd_find, which I don't want to do).  Basically, if
	the number of BDD nodes increased after a call to bdd_find, I
	know that a new BDD node was created, so I charge that node
	against the size bound.  However, cache hits and special cases
	can all result in large sub-BDDs not being charged to the
	size bound (since no new nodes were allocated), and a garbage
	collection during a bdd_find can result in failing to count a
	new node allocated during that bdd_find call.
	(In addition, I'm only checking the size bound upon entering a
	recursive call, so the size bound may be exceeded for a while
	before being detected (if we're backing out of a long chain
	of recursion.) (Not anymore 12 Jul 94)
	The guarantee, though, is that if the bounded routine aborts,
	the unbounded version would have created a BDD that exceeded the
	size bound.  (The errors are all one-sided.)
	-- AJH 11 Jul 94 */

bdd
#if defined(__STDC__)
bdd_bounded_ite(bdd_manager bddm, bdd f, bdd g, bdd h, long size_bound)
#else
bdd_bounded_ite(bddm, f, g, h, size_bound)
     bdd_manager bddm;
     bdd f;
     bdd g;
     bdd h;
     long size_bound;
#endif
{
  long size_foo = size_bound;
  bdd bdd_foo;

  if (bdd_check_arguments(3, f, g, h))
    {
      FIREWALL(bddm);
      bdd_foo = bdd_bounded_ite_step(bddm, f, g, h, &size_foo);
      if (bdd_foo==(bdd)0) {
	  /*fprintf(stderr,"Size Bound Exceeded.\n");*/
	  return (bdd)0;
      }
      RETURN_BDD(bdd_foo);
    }
  return ((bdd)0);
}
