/* Copyright 1991 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Header File Version 0.0 */
/* By Alan J. Hu */
/* July 22, 1991 */


#include <stdio.h>
#include <strings.h>
#include "bdduser.h"

#define TRUE 1
#define FALSE 0

/* Lex and Yacc Interface */

extern int yylineno;
extern char yytext[];
/*extern unsigned char yytext[];*/ /* Needed on Alpha, 26 Nov 1996 --AJH */
void lexinit();


/* *** Main Data Structures *** */


typedef struct record_list_node	*record_list;
typedef struct type_desc_node	*type_desc;
typedef struct string_list_node	*string_list;
typedef struct prop_list_node	*prop_list;
typedef struct uneval_prop_node	*uneval_prop;
typedef struct bit_vector_node	*bit_vector;
typedef struct vectorize_list_node	*vectorize_list;
typedef struct bit_vector_list_node	*bit_vector_list;
typedef struct variable_list_node	*variable_list;
typedef struct variable_desig_node	*variable_desig;
typedef struct var_accessor_node	*var_accessor;
typedef struct predicate_node	*predicate;
typedef struct id_list_node	*id_list;


struct record_list_node	{
				char		*field_name;
				type_desc	field_type;
				record_list	next;
			};

struct type_desc_node	{
				int		type;
				int		interleaved;
				int		bdd_count;
				int		bits;
				string_list	output_labels;
				int		array_lb, array_ub;
				type_desc	subtype;
				record_list	field_list;
			};

struct string_list_node	{
				char		*data;
				string_list	next;
			};

struct prop_list_node	{
				uneval_prop	data;
				prop_list	next;
			};

struct uneval_prop_node	{
				int		type;
				bdd	bdd;
				uneval_prop	prop1, prop2, prop3;
				prop_list	prop_list1;
				bit_vector_list	vec_list1;
				bit_vector	vector1, vector2;
				predicate	pred1;
				variable_desig	var1;
				variable_list	quant_vars;
				int		lineno;	/* For error messages.*/
				int		count; /* For repeat command */
			};

struct bit_vector_node	{
				int		type;
				bit_vector	vector1, vector2;
				bit_vector_list	vec_list1;
				vectorize_list	vectorize_list1;
				variable_desig	var1;
				uneval_prop	prop1;
				int		size;
				int		is_constant;
				int		constant_value;
				int		formal_order;
				bdd	*bit_list;
			};

struct vectorize_list_node	{
					int		type;
					uneval_prop	prop_data;
					bit_vector	vector_data;
					vectorize_list	next;
				};

struct bit_vector_list_node	{
					bit_vector	data;
					bit_vector_list	next;
				};

struct variable_list_node	{
					variable_desig	data;
					variable_list	next;
				};

struct variable_desig_node	{
					type_desc	base_type;
					int		bitcount;
					int		fdv; /* dependent? */
					bit_vector	func_def;
						/* Definition if fdv. */
					int		bfv_offset;
						/* offset to start of this var
						   in a bfv */
					bdd	*bits;
					int		phase;
					var_accessor	modifiers;
				};

struct var_accessor_node	{
					int		type;
					bit_vector	array_index;
					char		*field_select;
					var_accessor	next;
				};

struct predicate_node	{
				int		is_lambda;
				id_list		formal_list;
				uneval_prop	bound_prop;
			};

struct id_list_node	{
				char		*data;
				id_list		next;
			};


void pline();
void yyerror(char *s);
char *strpreserve(char *s);

/* Symbol Table Functions in eversymtab.c */

void define_type(char *id, type_desc its_type);
void define_variable(char *id, type_desc its_type);
void define_dependent_variable(char *id,
				type_desc its_type,
				bit_vector definition);
void define_proposition(char *id, uneval_prop value);
void define_predicate(char *id, predicate value);
void define_bitvector(char *id, bit_vector value);

type_desc	get_type_descriptor(char *id);
variable_desig	get_variable_descriptor(char *id);
uneval_prop	get_proposition_descriptor(char *id);
predicate	get_predicate_descriptor(char *id);
bit_vector	get_bit_vector(char *id);

void push_formals(id_list formal_list);
void pop_scope();

/* End of Symbol Table Functions */

/* Variable Ordering Functions in evervar.c */

void interleave_variables(bdd *base, type_desc remaining_type);

/* End of Variable Ordering Functions */


/* Goes forward from startstates, trying to reach the goal states. */
void printtrace(uneval_prop startstates,
		uneval_prop nextstate,
		uneval_prop goal);

/* Goes backward from invariant, seeing if any startstate can escape. */
void checkinvariant(uneval_prop startstates,
		    uneval_prop nextstates,
		    uneval_prop invariant);

/* Goes backward, but uses Nicolas's ideas to use upper approximations
   for better efficiency. */
void check_with_approx(uneval_prop startstates,
		       uneval_prop nextstate,
		       uneval_prop properties,
		       uneval_prop assumptions);
/* Goes backward from invariant, but tries to project out each iteration
   onto the user-specified support_sets. */
void check_with_support(uneval_prop startstates,
			uneval_prop nextstate,
			uneval_prop properties,
			bit_vector_list support_set_specifiers);

/* Goes forward, but uses implicit disjunctions. */
void check_invariant_fwd(uneval_prop startstates,
			 uneval_prop nextstate,
			 uneval_prop properties);


type_desc	build_bit_field_type(int size, string_list labels);
type_desc	build_array_type(int interleave,
				    int lb,
				    int ub,
				    type_desc element_type);
type_desc	build_record_type(int interleave, record_list fields);

string_list	append_to_string_list(char *datum, string_list old_list);

record_list	append_to_record_list(char *id,
					type_desc field_type,
					record_list old_list);

prop_list	append_to_prop_list(uneval_prop new_item, prop_list old_list);

vectorize_list	append_prop_to_vectorize_list(uneval_prop new_item,
						vectorize_list old_list);
vectorize_list	append_bit_vector_to_vectorize_list(bit_vector new_item,
						vectorize_list old_list);

bit_vector_list	append_to_bit_vector_list(bit_vector new_item,
					bit_vector_list old_list);

variable_list	append_to_variable_list(variable_desig new_item,
					variable_list old_list);

id_list append_to_ID_list(char *new_item, id_list old_list);

uneval_prop instantiate_predicate(predicate pred, bit_vector_list args);
uneval_prop evaluate_proposition(uneval_prop p, int force); /* in evereval.c */
uneval_prop and_propositions(prop_list props);
uneval_prop or_propositions(prop_list props);
uneval_prop not_proposition(uneval_prop p);
uneval_prop xor_propositions(uneval_prop a, uneval_prop b);
uneval_prop equiv_propositions(uneval_prop a, uneval_prop b);
uneval_prop implies_propositions(uneval_prop a, uneval_prop b);
uneval_prop forall_proposition(variable_list quant_vars, uneval_prop p);
uneval_prop exists_proposition(variable_list quant_vars, uneval_prop p);
uneval_prop hide_previous_uneval(uneval_prop p);
uneval_prop hide_current_uneval(uneval_prop p);
uneval_prop hide_next_uneval(uneval_prop p);
uneval_prop if_then_else_uneval(uneval_prop ifp,
				uneval_prop thenp,
				uneval_prop elsep);
uneval_prop reachable_uneval(uneval_prop start, uneval_prop next);
uneval_prop reachgoal_uneval(uneval_prop goal,
				uneval_prop start,
				uneval_prop next);
uneval_prop AG_uneval(uneval_prop p, uneval_prop next);
uneval_prop EG_uneval(uneval_prop p, uneval_prop next);
uneval_prop AF_uneval(uneval_prop p, uneval_prop next);
uneval_prop EF_uneval(uneval_prop p, uneval_prop next);
uneval_prop AX_uneval(uneval_prop p, uneval_prop next);
uneval_prop EX_uneval(uneval_prop p, uneval_prop next);
uneval_prop AU_uneval(uneval_prop p, uneval_prop untilq, uneval_prop next);
uneval_prop EU_uneval(uneval_prop p, uneval_prop untilq, uneval_prop next);
uneval_prop simrel_uneval(uneval_prop next1, uneval_prop next2);
uneval_prop becomes_uneval(variable_desig lhs, bit_vector rhs);
uneval_prop constrain_uneval(variable_desig lhs, uneval_prop rhs);
uneval_prop compose_uneval(prop_list nextstates);
uneval_prop repeat_uneval(int count, uneval_prop nextstate);
uneval_prop gt_uneval(bit_vector lhs, bit_vector rhs);
uneval_prop ge_uneval(bit_vector lhs, bit_vector rhs);
uneval_prop eq_uneval(bit_vector lhs, bit_vector rhs);
uneval_prop le_uneval(bit_vector lhs, bit_vector rhs);
uneval_prop lt_uneval(bit_vector lhs, bit_vector rhs);
uneval_prop and_compress_uneval(bit_vector x);
uneval_prop or_compress_uneval(bit_vector x);

bit_vector evaluate_bit_vector(bit_vector b, int force); /* in evereval.c */
bit_vector make_integer_bit_vector(int width, int i);
bit_vector make_variable_bit_vector(variable_desig v);
bit_vector vectorize(vectorize_list list);
bit_vector add_bit_vectors(bit_vector_list operands);
bit_vector sub_bit_vectors(bit_vector a, bit_vector b);
bit_vector left_shift_bit_vectors(bit_vector operand, bit_vector amount);
bit_vector right_shift_bit_vectors(bit_vector operand, bit_vector amount);
bit_vector left_rotate_bit_vectors(bit_vector operand, bit_vector amount);
bit_vector right_rotate_bit_vectors(bit_vector operand, bit_vector amount);
bit_vector bitwise_equivalence(bit_vector a, bit_vector b);
bit_vector bitwise_implies(bit_vector a, bit_vector b);
bit_vector bitwise_xor(bit_vector a, bit_vector b);
bit_vector bitwise_and(bit_vector_list operands);
bit_vector bitwise_or(bit_vector_list operands);
bit_vector bitwise_not(bit_vector a);
bit_vector bit_vector_if_then_else(uneval_prop cond,
					bit_vector thenpart,
					bit_vector elsepart);

variable_desig	add_phase_to_var_desig(int phase, variable_desig x);
variable_desig	array_index_var_desig(variable_desig var, bit_vector index);
variable_desig	record_access_var_desig(variable_desig var, char *fieldname);

predicate	build_macro_predicate(id_list formals, uneval_prop x);
predicate	build_lambda_predicate(id_list formals, uneval_prop x);



/* Returns true iff all the variable designators in variable_list
   are OK to quantify on. */
int quantifiable_variables(variable_list the_vars);

extern bdd_manager gbm;

extern uneval_prop true_proposition, false_proposition;
extern uneval_prop previous_current_eq, current_next_eq, previous_next_eq;
extern bdd *prev_vars, *cur_vars, *next_vars;
extern int prev_vars_assoc, cur_vars_assoc, next_vars_assoc;
extern int global_bdd_count;
extern int global_fdv_count;	/* number of functionally dependent bits */

extern string_list global_variable_names;	/* in eversymtab.c */



/* Routines for building multisets of BFVs (MOBs) to handle functionally
	dependent variables (FDVs) in evermob.c. */
void build_MOB(uneval_prop next_state_rel); /* Builds MOB for next_state_rel*/
bdd check_MOB(bdd reached); /* Returns set of states that can
						lead to violation. */

/* Monitoring Routines */

void reset_memory_meter();
long read_memory_meter();
int memory_grew();
int user_cpu_time();
long count_bdd_nodes(uneval_prop p);
double enumerate_reachable_states(bdd reachable,
				  int unconstrained_bits);
char *bdd_naming_fn_full(bdd_manager bddm,
			 bdd bdd_var,
			 pointer env);
void print_bdd_support(FILE *fp, char *(*naming_fn)(), bdd x);

/* evermem.c Memory Allocation Declarations -- 7 December 1996 */

/*
extern void *malloc(unsigned size);
extern void free(void *ptr);
extern void *realloc(void *ptr, unsigned size);
extern void *calloc(unsigned nelem, unsigned elsize);
*/
