/* Copyright 1991 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Lexer Version 0.0 */
/* By Alan J. Hu */
/* July 22, 1991 */

/****************************************************************/
/*  CHANGE LOG  */
/*  I guess I've been remiss here. */
/*  26 March 1994 AJH -- Added tokens checkwithsupport and using.
/*  22 December 1996 AJH -- Added token repeat.
/*  11 August 1997 AJH -- Added rotate operators
/****************************************************************/

/* Increase size of position array.  --AJH */
%p 4000
%a 3000

%{
#include "ever.h"
#include "y.tab.h"

%}


%start		norm comment

LETTER		[a-zA-Z_]
DIGIT		[0-9]
ID		({LETTER}({LETTER}|{DIGIT})*)
INTEGER		({DIGIT}{DIGIT}*)

%%
<norm>[ \t\n\014]		;		/* ignore whitespace */

<norm>[();[\]\.'] 	return(yytext[0]);
<norm>"^p"		return(PREV);
<norm>"^c"		return(CUR);
<norm>"^n"		return(NEXT);
<norm>"interleaved"	return(INTERLEAVED);
<norm>"bits"		return(BITS);
<norm>"array"		return(ARRAY);
<norm>"record"		return(RECORD);

<norm>"TRUE"		return(T);
<norm>"FALSE"		return(F);
<norm>"and"		return(AND);
<norm>"or"		return(OR);
<norm>"not"		return(NOT);
<norm>"xor"		return(XOR);
<norm>"equiv"		return(EQUIV);
<norm>"implies"		return(IMPLIES);
<norm>"forall"		return(FORALL);
<norm>"exists"		return(EXISTS);
<norm>"hideprev"	return(HIDEPREV);
<norm>"hidecur"		return(HIDECUR);
<norm>"hidenext"	return(HIDENEXT);
<norm>"if"		return(IF);

<norm>"reachable"	return(REACHABLE);
<norm>"reachgoal"	return(REACHGOAL);

<norm>"AG"		return(AG);
<norm>"EG"		return(EG);
<norm>"AF"		return(AF);
<norm>"EF"		return(EF);
<norm>"AX"		return(AX);
<norm>"EX"		return(EX);
<norm>"AU"		return(AU);
<norm>"EU"		return(EU);

<norm>"simrel"		return(SIMREL);
<norm>"compose"		return(COMPOSE);
<norm>"becomes"		return(BECOMES);
<norm>"constrain"	return(CONSTRAIN);
<norm>"repeat"		return(REPEAT);

<norm>"eval"		return(EVAL);

<norm>"deftype"		return(DEFTYPE);
<norm>"defvar"		return(DEFVAR);
<norm>"defdepvar"	return(DEFDEPVAR);
<norm>"defprop"		return(DEFPROP);
<norm>"defpred"		return(DEFPRED);
<norm>"defvec"		return(DEFVEC);
<norm>"lambda"		return(LAMBDA);

<norm>"printstring"	return(PRINTSTRING);
<norm>"printtrace"	return(PRINTTRACE);
<norm>"printprop"	return(PRINTPROP);
<norm>"printsize"	return(PRINTSIZE);
<norm>"checkinvariant"	return(CHECKINVARIANT);
<norm>"ctlmodelcheck"	return(CTLMODELCHECK);
<norm>"checkwithapprox"	return(CHECKWITHAPPROX);
<norm>"assuming"	return(ASSUMING);
<norm>"checkwithsupport"	return(CHECKWITHSUPPORT);
<norm>"using"		return(USING);
<norm>"checkinvariantfwd"	return(CHECKINVFWD);

<norm>"include"		return(INCLUDE);

<norm>"end"		return(END);

<norm>"PrevCurEq"	return(PREVCUREQ);
<norm>"CurNextEq"	return(CURNEXTEQ);
<norm>"PrevNextEq"	return(PREVNEXTEQ);

<norm>"vector"		return(VECTOR);
<norm>"add"		return(ADD);
<norm>"sub"		return(SUB);
<norm>"<<"		return(LEFTSHIFT);
<norm>">>"		return(RIGHTSHIFT);
<norm>"<<<"		return(LEFTROTATE);
<norm>">>>"		return(RIGHTROTATE);
<norm>"bequiv"		return(BEQUIV);
<norm>"bimplies"	return(BIMPLIES);
<norm>"bxor"		return(BXOR);
<norm>"band"		return(BAND);
<norm>"bor"		return(BOR);
<norm>"bnot"		return(BNOT);
<norm>"bif"		return(BIF);

<norm>"gt"		return(GT);
<norm>"ge"		return(GE);
<norm>"eq"		return(EQ);
<norm>"le"		return(LE);
<norm>"lt"		return(LT);
<norm>"andcompress"	return(ANDCOMPRESS);
<norm>"orcompress"	return(ORCOMPRESS);

<norm>\"[^"]*\"		{
			   /* Kill the last double-quote */
			   yytext[yyleng-1]='\0';
			   yylval.string = strpreserve(yytext+1);
			   return(STRING);
			}

<norm>{ID}		{
			   yylval.string = strpreserve(yytext);
			   return(ID);
			}

<norm>{INTEGER}		{
			   yylval.string = strpreserve(yytext);
			   return(INTEGER);
			}

"--"			{
			   BEGIN comment;
			}

<comment>(.)		;	/* discard everything within a comment */
<comment>[\n]		{
			  BEGIN norm; /* newline ends a comment */
			}

<norm>.			{ /* No other rule applies, so it's a bad token */
		   	   fprintf(stderr,"Bad input char `%c' on line %d\n",
			      yytext[0],yylineno);
			}
%%

/* lexinit - starts the Lexical Analyzer off in the right start condition */
void lexinit()
{
   BEGIN norm;
}

yywrap()
{
    if ((yyin != stdin)) {
	close (yyin);
	yyin = stdin;
	printf("Command> ");
	return(0);
    };
    return (1);
}


