/* Copyright 1991 by Alan J. Hu */
/* All rights reserved. */

/* EVER YACC Grammar Version 0.0 */
/* By Alan J. Hu */
/* July 22, 1991 */

/***********************  CHANGE LOG ***********************/
/*
Aug 15, 1991 -- AJH:  Changed predicate arguments to be bit vectors.
Sep 28, 1991 -- AJH:  Added syntax for interleaved variable ordering.
Nov 7, 1992  -- AJH:  Revised to use David Long's BDD package.
Jan 4, 1993  -- AJH:  Added command for backward invariant checking
Oct 28, 1993 -- AJH:  Added command to use Nicolas' approximation stuff
Mar 26, 1994 -- AJH:  Added command to do backward invariant checking
			using user-specified support sets.
Dec 22, 1996 -- AJH:  Added repeat operator to do rule iteration.
Aug 11, 1997 -- AJH:  Added rotate operators.
*/
/**********************************************************/

%{
#include "ever.h"

extern FILE *yyin;

%}

%start commandlist

%union	yystacktype {
  char			*string;
  int			integer;
  type_desc		typedescriptor;
  string_list		optlabellist;
  record_list		recordlist;
  prop_list		proplist;
  uneval_prop		unevalprop;
  bit_vector		bitvector;
  vectorize_list	vectorizelist;
  bit_vector_list	bitvectorlist;
  variable_list		variablelist;
  variable_desig	variabledesig;
  predicate		predicate;
  id_list		idlist;
}

%type <typedescriptor> typedescriptor
%type <optlabellist> optlabellist
%type <recordlist> recordlist
%type <proplist> unevalproplist
%type <unevalprop> unevalprop
%type <bitvector> bitvector arraydesig
%type <vectorizelist> vectorizelist
%type <bitvectorlist> bitvectorlist actuallist
%type <variablelist> variablelist
%type <variabledesig> variabledesignator unphasedvardesig
%type <predicate> predicate
%type <idlist> formallist
%type <string> recorddesig
%type <integer> optbitwidth optinterleave
%nonassoc <string> ID STRING INTEGER



%token	PREV CUR NEXT
%token	INTERLEAVED BITS ARRAY RECORD
%token	T F
%token	AND OR NOT XOR EQUIV IMPLIES FORALL EXISTS HIDEPREV HIDECUR HIDENEXT IF
%token	REACHABLE REACHGOAL
%token	AG EG AF EF AX EX AU EU
%token	SIMREL COMPOSE BECOMES CONSTRAIN REPEAT
%token	EVAL
%token	DEFTYPE DEFVAR DEFDEPVAR DEFPROP DEFPRED DEFVEC LAMBDA
%token	PRINTSTRING PRINTTRACE PRINTPROP PRINTSIZE CHECKINVARIANT CTLMODELCHECK
%token	CHECKWITHAPPROX ASSUMING
%token	CHECKWITHSUPPORT USING
%token	CHECKINVFWD
%token	INCLUDE
%token	END
%token	PREVCUREQ CURNEXTEQ PREVNEXTEQ
%token	VECTOR ADD SUB LEFTSHIFT RIGHTSHIFT LEFTROTATE RIGHTROTATE
%token	BEQUIV BIMPLIES BXOR BAND BOR BNOT BIF
%token	GT GE EQ LE LT ANDCOMPRESS ORCOMPRESS

%%

commandlist :	commandlist command
		{
		  if (yyin == stdin) printf("Command> ");
		}  
	|	 /* empty */
		{
		  if (yyin == stdin) printf("Command> ");
		}
	;

command	:	INCLUDE STRING ';'
		{
		  /* *** Doesn't deal with recursive loads *** */
		  yyin = fopen($2,"r"); 
		  if (yyin==NULL) { 
			fprintf( stderr, "*** Error opening `%s' ***\n",$2);
			fflush(stderr);
			/* perhaps less-than-perfect recovery */
			yyin=stdin;
		  } else printf("Taking input from %s...\n",$2);
		  free($2);
		}
	|	END ';'
		{
		  return 0;	/* Exit the parser. */
		}
	|	DEFTYPE ID typedescriptor ';'
		{
		  define_type($2,$3);
		}
	|	DEFVAR ID typedescriptor ';'
		{
		  define_variable($2,$3);
		}
	|	DEFDEPVAR ID typedescriptor bitvector ';'
		{
		  define_dependent_variable($2,$3,$4);
		}
	|	DEFPROP ID unevalprop ';'
		{
		  define_proposition($2,$3);
		}
	|	DEFPRED ID predicate ';'
		{
		  define_predicate($2,$3);
		}
	|	DEFVEC ID bitvector ';'
		{
		  define_bitvector($2,$3);
		}
	|	PRINTSTRING STRING ';'
		{
		  printf("%s\n",$2);
		}
	|	PRINTTRACE unevalprop unevalprop unevalprop ';'
		{
		  /* $2 is start, $3 is nextstate, $4 is goal */
		  printtrace($2,$3,$4);
		}
	|	CHECKINVARIANT unevalprop unevalprop unevalprop ';'
		{
		  /* $2 is start, $3 is nextstate, $4 is invariant */
		  check_invariant($2,$3,$4);
		}
	|	CTLMODELCHECK unevalprop unevalprop unevalprop ';'
		{
		  /* $2 is states to be checked */
		  /* $3 is nextstate */
		  /* $4 is CTL formula. */
		  /* Note that the nextstate relation is given, thus
			the nextstate relations embedded in the CTL
			formula are ignored. */
		  ctl_model_check($2,$3,$4);
		}
	|	CHECKWITHAPPROX unevalprop unevalprop unevalprop
		ASSUMING unevalprop ';'
		{
		  /* $2 is startstates, $3 is nextstate */
		  /* $4 is invariant to check */
		  /* $6 is assumption (upper approx of reachable) */
		  /* KLUDGE for $4 and $6, use (and ...) to give lists. */
		  check_with_approx($2,$3,$4,$6);
		}
	|	CHECKWITHSUPPORT unevalprop unevalprop unevalprop
		USING bitvectorlist ';'
		{
		  /* $2 is startstates, $3 is nextstate */
		  /* $4 is invariant to check */
		  /* $6 is assumption (upper approx of reachable) */
		  /* KLUDGE for $4, use (and ...) to give lists. */
		  check_with_support($2,$3,$4,$6);
		}
	|	CHECKINVFWD unevalprop unevalprop unevalprop ';'
		{
		  /* $2 is start, $3 is nextstate, $4 is invariant */
		  check_invariant_fwd($2,$3,$4);
		}
	|	PRINTPROP unevalprop ';'
		{
		  bdd_print_bdd(gbm,evaluate_proposition($2,FALSE)->bdd,
					bdd_naming_fn_full,
					bdd_terminal_id_fn_none,
					NULL,stdout);
		}
	|	PRINTSIZE unevalprop ';'
		{
		  printf("Unevaluated Size: %ld\n", count_bdd_nodes($2));
		}
	|	error ';'
		{
		  yyerrok;
		}
	;

typedescriptor :
		ID
		{
		  $$ = get_type_descriptor($1);
		}
	|	'(' BITS INTEGER optlabellist ')'
		{
		  $$ = build_bit_field_type(atoi($3),$4);
		}
	|	'(' optinterleave ARRAY INTEGER INTEGER typedescriptor ')'
		{
		  $$ = build_array_type($2,atoi($4),atoi($5),$6);
		}
	|	'(' optinterleave RECORD recordlist ')'
		{
		  $$ = build_record_type($2,$4);
		}
	;

optinterleave :	INTERLEAVED
		{
		  $$ = TRUE;
		}
	|	/* empty */
		{
		  $$ = FALSE;
		}
	;

optlabellist :
		optlabellist STRING
		{
		  $$ = append_to_string_list($2,$1);
		}
	|	/* empty */
		{
		  $$ = NULL;
		}
	;

recordlist :	ID typedescriptor
		{
		  $$ = append_to_record_list($1,$2,NULL);
		}
	|	recordlist ID typedescriptor
		{
		  $$ = append_to_record_list($2,$3,$1);
		}
	;

unevalproplist :
		unevalprop
		{
		  $$ = append_to_prop_list($1,NULL);
		}
	|	unevalproplist unevalprop
		{
		  $$ = append_to_prop_list($2,$1);
		}
	;

unevalprop :	T
		{
		  $$ = true_proposition;
		}
	|	F
		{
		  $$ = false_proposition;
		}
	|	'(' ID ')'
		{
		  $$ = get_proposition_descriptor($2);
		}
	|	'(' ID actuallist ')'
		{
		  $$ = instantiate_predicate(get_predicate_descriptor($2),$3);
		}
	|	PREVCUREQ
		{
		  $$ = previous_current_eq;
		}
	|	CURNEXTEQ
		{
		  $$ = current_next_eq;
		}
	|	PREVNEXTEQ
		{
		  $$ = previous_next_eq;
		}
	|	'(' EVAL unevalprop ')'
		{
		  $$ = evaluate_proposition($3,TRUE);
		}
	|	'(' AND unevalproplist ')'
		{
		  $$ = and_propositions($3);
		}
	|	'(' OR unevalproplist ')'
		{
		  $$ = or_propositions($3);
		}
	|	'(' NOT unevalprop ')'
		{
		  $$ = not_proposition($3);
		}
	|	'(' XOR unevalprop unevalprop ')'
		{
		  $$ = xor_propositions($3,$4);
		}
	|	'(' EQUIV unevalprop unevalprop ')'
		{
		  $$ = equiv_propositions($3,$4);
		}
	|	'(' IMPLIES unevalprop unevalprop ')'
		{
		  $$ = implies_propositions($3,$4);
		}
	|	'(' FORALL variablelist unevalprop ')'
		{
		  if (quantifiable_variables($3)) {
		    $$ = forall_proposition($3,$4);
		  } else {
		    yyerror("Illegal Variables in Quantifier");
		  }
		}
	|	'(' EXISTS variablelist unevalprop ')'
		{
		  if (quantifiable_variables($3)) {
		    $$ = exists_proposition($3,$4);
		  } else {
		    yyerror("Illegal Variables in Quantifier");
		  }
		}
	|	'(' HIDEPREV unevalprop ')'
		{
		  $$ = hide_previous_uneval($3);
		}
	|	'(' HIDECUR unevalprop ')'
		{
		  $$ = hide_current_uneval($3);
		}
	|	'(' HIDENEXT unevalprop ')'
		{
		  $$ = hide_next_uneval($3);
		}
	|	'(' IF unevalprop unevalprop ')'
		{
		  /* We might actually want if-then-else semantics here. */
		  $$ = implies_propositions($3,$4);
		}
	|	'(' IF unevalprop unevalprop unevalprop ')'
		{
		  $$ = if_then_else_uneval($3,$4,$5);
		}
	|	'(' REACHABLE unevalprop unevalprop ')'
		{
		  /* $3 is start state, $4 is nextstate */
		  $$ = reachable_uneval($3,$4);
		}
	|	'(' REACHGOAL unevalprop unevalprop unevalprop ')'
		{
		  /* $3 is goal, $4 is start state, $5 is nextstate */
		  $$ = reachgoal_uneval($3,$4,$5);
		}
	|	'(' AG unevalprop unevalprop ')'
		{
		  /* $3 is property, $4 is nextstate relation */
		  $$ = AG_uneval($3,$4);
		}
	|	'(' EG unevalprop unevalprop ')'
		{
		  /* $3 is property, $4 is nextstate relation */
		  $$ = EG_uneval($3,$4);
		}
	|	'(' AF unevalprop unevalprop ')'
		{
		  /* $3 is property, $4 is nextstate relation */
		  $$ = AF_uneval($3,$4);
		}
	|	'(' EF unevalprop unevalprop ')'
		{
		  /* $3 is property, $4 is nextstate relation */
		  $$ = EF_uneval($3,$4);
		}
	|	'(' AX unevalprop unevalprop ')'
		{
		  /* $3 is property, $4 is nextstate relation */
		  $$ = AX_uneval($3,$4);
		}
	|	'(' EX unevalprop unevalprop ')'
		{
		  /* $3 is property, $4 is nextstate relation */
		  $$ = EX_uneval($3,$4);
		}
	|	'(' AU unevalprop unevalprop unevalprop ')'
		{
		  /* $3 until $4; $5 is nextstate relation */
		  $$ = AU_uneval($3,$4,$5);
		}
	|	'(' EU unevalprop unevalprop unevalprop ')'
		{
		  /* $3 until $4; $5 is nextstate relation */
		  $$ = EU_uneval($3,$4,$5);
		}
	|	'(' SIMREL unevalprop unevalprop ')'
		{
		  $$ = simrel_uneval($3,$4);
		}
	|	'(' BECOMES variabledesignator bitvector ')'
		{
		  $$ = becomes_uneval($3,$4);
		}
	|	'(' CONSTRAIN variabledesignator unevalprop ')'
		{
		  $$ = constrain_uneval($3,$4);
		}
	|	'(' COMPOSE unevalproplist ')'
		{
		  $$ = compose_uneval($3);
		}
	|	'(' REPEAT INTEGER unevalprop ')'
		{
		  $$ = repeat_uneval(atoi($3),$4);
		}
	|	'(' GT bitvector bitvector ')'
		{
		  $$ = gt_uneval($3,$4);
		}
	|	'(' GE bitvector bitvector ')'
		{
		  $$ = ge_uneval($3,$4);
		}
	|	'(' EQ bitvector bitvector ')'
		{
		  $$ = eq_uneval($3,$4);
		}
	|	'(' LE bitvector bitvector ')'
		{
		  $$ = le_uneval($3,$4);
		}
	|	'(' LT bitvector bitvector ')'
		{
		  $$ = lt_uneval($3,$4);
		}
	|	'(' ANDCOMPRESS bitvector ')'
		{
		  $$ = and_compress_uneval($3);
		}
	|	'(' ORCOMPRESS bitvector ')'
		{
		  $$ = or_compress_uneval($3);
		}
	;

bitvector :	ID
		{
		  $$ = get_bit_vector($1);
		}
	|	optbitwidth INTEGER
		{
		  $$ = make_integer_bit_vector($1,atoi($2));
		}
	|	variabledesignator
		{
		  $$ = make_variable_bit_vector($1);
		}
	|	'(' VECTOR vectorizelist ')'
		{
		  $$ = vectorize($3);
		}
	|	'(' ADD bitvectorlist ')'
		{
		  $$ = add_bit_vectors($3);
		}
	|	'(' SUB bitvector bitvector ')'
		{
		  $$ = sub_bit_vectors($3,$4);
		}
	|	'(' LEFTSHIFT bitvector bitvector ')'
		{
		  $$ = left_shift_bit_vectors($3,$4);
		}
	|	'(' RIGHTSHIFT bitvector bitvector ')'
		{
		  $$ = right_shift_bit_vectors($3,$4);
		}
	|	'(' LEFTROTATE bitvector bitvector ')'
		{
		  $$ = left_rotate_bit_vectors($3,$4);
		}
	|	'(' RIGHTROTATE bitvector bitvector ')'
		{
		  $$ = right_rotate_bit_vectors($3,$4);
		}
	|	'(' BEQUIV bitvector bitvector ')'
		{
		  $$ = bitwise_equivalence($3,$4);
		}
	|	'(' BIMPLIES bitvector bitvector ')'
		{
		  $$ = bitwise_implies($3,$4);
		}
	|	'(' BXOR bitvector bitvector ')'
		{
		  $$ = bitwise_xor($3,$4);
		}
	|	'(' BAND bitvectorlist ')'
		{
		  $$ = bitwise_and($3);
		}
	|	'(' BOR bitvectorlist ')'
		{
		  $$ = bitwise_or($3);
		}
	|	'(' BNOT bitvector ')'
		{
		  $$ = bitwise_not($3);
		}
	|	'(' BIF unevalprop bitvector bitvector ')'
		{
		  $$ = bit_vector_if_then_else($3,$4,$5);
		}
	;

optbitwidth :	'\'' INTEGER '\''
		{
		  $$ = atoi($2);
		}
	|	/* Nothing */
		{
		  $$ = -1;	/* Signal that size not specified. */
		}
	;

vectorizelist :	unevalprop
		{
		  $$ = append_prop_to_vectorize_list($1,NULL);
		}
	|	bitvector
		{
		  $$ = append_bit_vector_to_vectorize_list($1,NULL);
		}
	|	vectorizelist unevalprop
		{
		  $$ = append_prop_to_vectorize_list($2,$1);
		}
	|	vectorizelist bitvector
		{
		  $$ = append_bit_vector_to_vectorize_list($2,$1);
		}
	;

bitvectorlist :	bitvector
		{
		  $$ = append_to_bit_vector_list($1,NULL);
		}
	|	bitvectorlist bitvector
		{
		  $$ = append_to_bit_vector_list($2,$1);
		}
	;

variablelist :	variablelist variabledesignator
		{
		  $$ = append_to_variable_list($2,$1);
		}
	|	/* empty */
		{
		  $$ = NULL;
		}
	;

variabledesignator :
		unphasedvardesig CUR
		{
		  $$ = add_phase_to_var_desig(CUR,$1);
		}
	|	unphasedvardesig PREV
		{
		  $$ = add_phase_to_var_desig(PREV,$1);
		}
	|	unphasedvardesig NEXT
		{
		  $$ = add_phase_to_var_desig(NEXT,$1);
		}
	;

unphasedvardesig :
		ID
		{
		  $$ = get_variable_descriptor($1);
		}
	|	unphasedvardesig arraydesig
		{
		  $$ = array_index_var_desig($1,$2);
		}
	|	unphasedvardesig recorddesig
		{
		  $$ = record_access_var_desig($1,$2);
		}
	;

arraydesig :	'[' bitvector ']'
		{
		  $$ = $2;
		}
	;

recorddesig :	'.' ID
		{
		  $$ = $2;
		}
	;

predicate :	'(' formallist ')'
		{
		  push_formals($2);
		}
		unevalprop
		{
		  $$ = build_macro_predicate($2,$5);
		  pop_scope();
		}
	|	LAMBDA '(' formallist ')'
		{
		  push_formals($3);
		}
		unevalprop
		{
		  $$ = build_lambda_predicate($3,$6);
		  pop_scope();
		}
	;

formallist :	formallist ID
		{
		  $$ = append_to_ID_list($2,$1);
		}
	|	ID
		{
		  $$ = append_to_ID_list($1,NULL);
		}
	;

actuallist :	bitvectorlist
		{
		  $$ = $1;
		}
	;
		

%%
