/* Copyright 1993 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Routines For CTL Model-Checking with Implicit Conjunctions */
/* By Alan J. Hu */
/* August 24, 1993 */
/* Copied into a separate file September 7, 1994. */

/* ************************ CHANGE LOG ************************** */
/*
24 August 1993	-- Started adding CTL model checking with implicitly
			conjoined invariants.  (Maybe only do ACTL first.)
*/
/* ************************************************************** */

#include "ever.h"
#include "everlist.h"
#include "everimage.h"
#include "bddajh.h"
#include "y.tab.h"	/* Defines constants for the tokens. */

/* Borrow the glue functions from everinv.c, which these routines used to
	be a part of. */
bdd simplify_bdd(bdd f, bdd c);
int converged(conjunct_list a, conjunct_list b);
conjunct_list conjunct_and_simplify(conjunct_list a, conjunct_list b);

conjunct_list disjunct_and_simplify_OLD(conjunct_list a, conjunct_list b)
	/* Modeled after c_and_s_4(). */
{
    int i, j;
    conjunct_list out, c;
    bdd temp1;

    /* Or the two lists. */
    c = or_conjunct_lists(a,b);

    /* Simplify everything by everything. */
    for (i=0; i < c->size; i++) {
	for (j=0; j < c->size; j++) {
	    if (i!=j) {
		temp1 = simplify_bdd(c->conjunct[i],c->conjunct[j]);
		bdd_free(gbm,c->conjunct[i]);
		c->conjunct[i] = temp1;
	    }
	}
    }

    /* Remove duplicates, special cases. */

    /* Munch the list down by evaluating conjunctions. */
    slow_munch_conjunct_list(c, 1.0 /* Grow Threshold */);
    out = c;

    return out;
} /* disjunct_and_simplify_OLD */



conjunct_list check_AG_with_conjunct_list(uneval_prop nextstate,
					  conjunct_list p)
	/* Returns a conjunct list for AG p */
{
    conjunct_list set, newset, temp1;

    set = copy_conjunct_list(p);
    do {
	temp1 = inv_image(nextstate,set);
	newset = conjunct_and_simplify(temp1,p);    /* Could use set instead.*/

	if (converged(newset,set)) break; /* Loop exit */

	free_conjunct_list(set);
	free_conjunct_list(temp1);
	set = newset;
    } while (TRUE);

    free_conjunct_list(set);
    free_conjunct_list(temp1);
    return newset;
}

conjunct_list check_AF_with_conjunct_list(uneval_prop nextstate,
					  conjunct_list p)
	/* Returns a conjunct list for AF p */
{
    conjunct_list set, newset, temp1;

    set = copy_conjunct_list(p);
    do {
	temp1 = inv_image(nextstate,set);
	newset = disjunct_and_simplify_OLD(temp1,p);    /* Could use set instead.*/

	if (converged(newset,set)) break; /* Loop exit */

	free_conjunct_list(set);
	free_conjunct_list(temp1);
	set = newset;
    } while (TRUE);

    free_conjunct_list(set);
    free_conjunct_list(temp1);
    return newset;
}

conjunct_list check_AU_with_conjunct_list(uneval_prop nextstate,
					  conjunct_list p,
					  conjunct_list q)
	/* Returns a conjunct list for A[p U q] */
{
    conjunct_list set, newset, temp1, temp2;

    set = copy_conjunct_list(q);
    do {
	temp1 = inv_image(nextstate,set);
	temp2 = conjunct_and_simplify(p,temp1);
	newset = disjunct_and_simplify_OLD(temp2,q);    /* Could use set instead.*/

	if (converged(newset,set)) break; /* Loop exit */

	free_conjunct_list(set);
	free_conjunct_list(temp1);
	free_conjunct_list(temp2);
	set = newset;
    } while (TRUE);

    free_conjunct_list(set);
    free_conjunct_list(temp1);
    free_conjunct_list(temp2);
    return newset;
}

conjunct_list ctl_formula_to_conjunct_list(uneval_prop nextstate,
					   uneval_prop ctl_formula)
	/* Returns a conjunct list for the set of states that
	   satisfy the given CTL formula. */
{
    conjunct_list result, temp1, temp2;
    bdd tempbdd1, tempbdd2;
    prop_list temppl;

    if (ctl_formula->type == EVAL) {
	/* Predefined formula. */
	return bdd_to_conjunct_list(ctl_formula->bdd);
    }

    if (ctl_formula->bdd != NULL) {
	/* Already been evaluated before. */
	return bdd_to_conjunct_list(ctl_formula->bdd);
    }

    switch (ctl_formula->type) {
    case AND:
	/* Concatenate the conjunct_lists.  Then crunch it down. */
	temppl = ctl_formula->prop_list1;
	result = bdd_to_conjunct_list(bdd_one(gbm));
	while (temppl!=NULL) {
	    temp1 = ctl_formula_to_conjunct_list(nextstate,temppl->data);
	    temp2 = concat_conjunct_lists(result, temp1);
	    free_conjunct_list(temp1);
	    free_conjunct_list(result);
	    result = temp2;
	    temppl = temppl->next;
	}
	slow_munch_conjunct_list(result,1.0);
	break;
    case OR:
	/* Multiply out the conjunct_lists.  Then crunch it down. */
	temppl = ctl_formula->prop_list1;
	result = bdd_to_conjunct_list(bdd_zero(gbm));
	while (temppl!=NULL) {
	    temp1 = ctl_formula_to_conjunct_list(nextstate,temppl->data);
	    temp2 = or_conjunct_lists(result,temp1);
	    free_conjunct_list(temp1);
	    free_conjunct_list(result);
	    /* Crunch on each iteration, since these can grow a lot. */
	    slow_munch_conjunct_list(temp2,1.0);
	    result = temp2;
	    temppl = temppl->next;
	}
	break;
    case NOT:
	/* This version is more general for long term, but slower for now. */
	/*8temp1 = ctl_formula_to_conjunct_list(nextstate,ctl_formula->prop1);
	tempbdd1 = conjunct_list_to_bdd(temp1);
	free_conjunct_list(temp1);
	tempbdd2 = bdd_not(gbm,tempbdd1);
	bdd_free(gbm,tempbdd1);
	result = bdd_to_conjunct_list(tempbdd2);
	bdd_free(gbm,tempbdd2);
	*/
	/* This version recognizes that anything under negation should
		be a plain formula. */
	evaluate_proposition(ctl_formula->prop1,FALSE);
	tempbdd1 = bdd_not(gbm,ctl_formula->prop1->bdd);
	result = bdd_to_conjunct_list(tempbdd1);
	bdd_free(gbm,tempbdd1);
	break;
    case XOR:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    case EQUIV:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    case IMPLIES:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    case IF:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    case AG:
	/* Note that the nextstate relation specified as part of the
		uneval_prop ctl_formula is ignored. */
	temp1 = ctl_formula_to_conjunct_list(nextstate,ctl_formula->prop1);
	result = check_AG_with_conjunct_list(nextstate,temp1);
	free_conjunct_list(temp1);
	break;
    case EG:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    case AF:
	temp1 = ctl_formula_to_conjunct_list(nextstate,ctl_formula->prop1);
	result = check_AF_with_conjunct_list(nextstate,temp1);
	free_conjunct_list(temp1);
	break;
    case EF:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    case AX:
	temp1 = ctl_formula_to_conjunct_list(nextstate,ctl_formula->prop1);
	result = inv_image(nextstate,temp1);
	free_conjunct_list(temp1);
	break;
    case EX:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    case AU:
	temp1 = ctl_formula_to_conjunct_list(nextstate,ctl_formula->prop1);
	temp2 = ctl_formula_to_conjunct_list(nextstate,ctl_formula->prop2);
	result = check_AU_with_conjunct_list(nextstate,temp1,temp2);
	free_conjunct_list(temp1);
	free_conjunct_list(temp2);
	break;
    case EU:
	fprintf(stderr,"DEBUG CTL unimplemented %d.\n",ctl_formula->type);
	break;
    default:
	evaluate_proposition(ctl_formula,FALSE);
	result = bdd_to_conjunct_list(ctl_formula->bdd);
    }

    return result;
} /* ctl_formula_to_conjunct_list */


void ctl_model_check(uneval_prop startstates,
		     uneval_prop nextstate,
		     uneval_prop ctl_formula)
	/* Note that the nextstate relation is given, thus the nextstate
	   relations given in ctl_formula (embedded with operators) are
	   INGORED. */
{
    conjunct_list good_states;
    int i;

    if (global_fdv_count>0) {
	/* I don't support FDVs with ctl_model_check.  Bail out. */
	fprintf(stderr,"Error:  Cannot use dependent variables with CTL model-checking.\n");
	return;
    }

    fprintf(stderr,"Finding states that satisfy CTL formula...\n");
    good_states = ctl_formula_to_conjunct_list(nextstate,ctl_formula);
    fprintf(stderr,"Done finding states that satisfy CTL formula...\n");

    fprintf(stderr,"Evaluating states to check the formula on...\n");
    evaluate_proposition(startstates,FALSE);

    fprintf(stderr,"Checking if specified states satisfy CTL formula...\n");
    i = find_violation(startstates->bdd, good_states);
    if (i<0) {
	printf("CTL formula holds on specified states.\n");
    } else {
	printf("CTL formula does not hold on all specified states.\n");
	printf("(Violates conjunct %d.)\n",i);	/* Useless info */
    }

    free_conjunct_list(good_states);
}
