/* Copyright 1991 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Evaluator Version 0.0 */
/* By Alan J. Hu */
/* August 6, 1991 */


/* ************************ CHANGE LOG ************************** */
/*
8 September 1991 AJH -- Changed all code for unevaluated bit vectors.
16 September 1991 AJH -- Added special code for constant array indices.
8 November 1991 AJH -- Deleted old code for the convoluted interleaving stuff.
11 December 1991 AJH -- Took out code for predicate instantiation, because
			I've decided to do it statically, instead.  This
			is less space-efficient, but cleaner (and correct
			for nested predicates).
4 January 1991 AJH -- Fixed some memory leakage in evaluate_proposition.
31 January 1992 AJH -- Added constrain operator for constrained assignment.
12 October 1992 AJH -- Added bif conditional operator.
15 October 1992 AJH -- Implemented vector operator.
19 November 1992 AJH -- Converted to David Long's BDD package
16 July 1993 AJH -- Modified building stability predicates in bdd_becomes()
		    for greater speed when doing large examples.  The old
		    method explicitly built up the stability predicate, bit
		    by bit, for all variables.  The new method starts with
		    current_next_eq->bdd, and "frees" (ORs in (x^n != x^c))
		    each variable that will actually be modified.
		    UNFORTUNATELY, THIS IS WRONG.  :-(
19 July 1993 AJH -- Did the correct modification to bdd_becomes().  I needed
		    to quantify the variables out, rather than using OR.
		    The new method could possibly be slower for very very
		    large (many elements) arrays with non-constant indices,
		    as the new method is always returning a complete next-state
		    relation on all variables.  I also added constant array
		    index range checking to bdd_becomes().
24 July 1993 AJH -- Made the same change to bdd_constrain().
26 July 1993 AJH -- Caught some memory leaks introduced in recent changes.
2 September 1993 AJH --Added leftshift operator.
4-6 December 1994 AJH -- Caught a bug in handling bnot.  Finished bitwise ops.
7 December 1996 AJH -- Fixed memory leak in eval_bit_vector_add(),
			eval_bit_vector_band(), and eval_bit_vector_bor().
22 December 1996 AJH -- Added eval for repeat operator. 
11 August 1997 AJH -- Added eval bitvector for rotate operators.
*/
/* ************************************************************** */


#include "ever.h"
#include "y.tab.h"



/* Internal Routines */



/* Helper function for eval_bit_vector_var_access() */

int var_desig_size(type_desc type_structure, var_accessor modifiers)
	/* Also used by routines in evermob.c. */
{
/*printf("DEBUG enter var_desig_size.  type_structure = %d\n",type_structure);*/
/*printf("size of type_structure is %d\n",type_structure->bdd_count);*/
    if (modifiers==NULL) return (type_structure->bdd_count)/3;

    if (modifiers->type==ARRAY) {
	if (type_structure->type!=ARRAY) {
	    yyerror("Illegal attempt to index a non-array.");
	    return 0;
	} else return var_desig_size(type_structure->subtype, modifiers->next);
    }

    if (modifiers->type==RECORD) {
	if (type_structure->type!=RECORD) {
	    yyerror("Illegal attempt to select a field of a non-record.");
	    return 0;
	} else {
	    record_list temp;

	    temp = type_structure->field_list;
	    while (temp!=NULL) {
		if (strcmp(temp->field_name,modifiers->field_select)==0) {
		    return var_desig_size(temp->field_type, modifiers->next);
		} else temp = temp->next;
	    }
	    yyerror("Undeclared field.");
	    return 0;
	}
    }

    yyerror("Impossible Error--Illegal modifier type in var_desig_size.");
    return 0;
}



bdd bdd_eq_int_bit_vector(int x, bit_vector y, int force); /* forward */

bdd build_rhs_var_prop(bdd *base,
				type_desc remaining_type,
				int pre_offset,
				int whichbit,
				int phase,
				var_accessor modchain,
				int force)
{
/*printf("DEBUG--Enter build_rhs_var_prop.\n");*/
/*printf("Base=%d, remaining=%d, pre_offset=%d, whichbit=%d, phase=%d, modchain=%d\n",base,remaining_type,pre_offset,whichbit,phase,modchain);*/
    if (modchain==NULL) {
	bdd new_node;

	/* I think these should be bdd_assigned.  --AJH 1/4/92 */
	if (phase==PREV) new_node = base[pre_offset+3*whichbit+0];
	if (phase==CUR) new_node = base[pre_offset+3*whichbit+1];
	if (phase==NEXT) new_node = base[pre_offset+3*whichbit+2];
	return new_node;
    }

    if ((modchain->type==ARRAY) && (!modchain->array_index->is_constant)) {
	int i;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;
	bdd result, temp1, temp2, temp3;

	result = bdd_one(gbm);
	for (i=lb; i<=ub; i++) {
	    temp1 = bdd_eq_int_bit_vector(i,modchain->array_index,force);
	    temp2 = build_rhs_var_prop(base,
				    element_type,
				    pre_offset+(i-lb)*element_type->bdd_count,
				    whichbit,
				    phase,
				    modchain->next,
				    force);
/*printf("Building non-constant rhs array access.\n");*/
	    temp3 = bdd_ite(gbm,temp1, temp2, bdd_one(gbm));
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);
	    temp1 = bdd_and(gbm,result,temp3);
	    bdd_free(gbm,result);
	    bdd_free(gbm,temp3);
	    result = temp1;
	}
	return result;
    }

    if ((modchain->type==ARRAY) && (modchain->array_index->is_constant)) {
	int i = modchain->array_index->constant_value;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

	if ((i<lb) || (i>ub))
		yyerror("Warning --- Constant array index out of range.");

	return	 build_rhs_var_prop(base,
				    element_type,
				    pre_offset+(i-lb)*element_type->bdd_count,
				    whichbit,
				    phase,
				    modchain->next,
				    force);
    }

    if (modchain->type==RECORD) {
	record_list temp = remaining_type->field_list;
	int offset = 0;

	while (temp!=NULL) {
	    if (strcmp(temp->field_name,modchain->field_select)==0) {
		return build_rhs_var_prop(base,
			temp->field_type,
			pre_offset+offset,
			whichbit,
			phase,
			modchain->next,
			force);
	    } else {
		offset += temp->field_type->bdd_count;
		temp = temp->next;
	    }
	}
	yyerror("Undeclared field.");
    }

    yyerror("Internal error in build_rhs_var_prop:  illegal modifier.");
    return bdd_zero(gbm);
}

bit_vector eval_bit_vector_var_access(bit_vector b, int force)
{
    int wordsize;
    int i;
    bit_vector new_vector;
    variable_desig v = b->var1;

    wordsize = var_desig_size(v->base_type, v->modifiers);
    if (wordsize > 0) {		/* Type checking worked out. */
	b->bit_list = (bdd *)calloc(wordsize,sizeof(bdd));

	/* Puts MSB at head of list. */
	for (i=0; i<wordsize; i++) {
	    b->bit_list[wordsize-1-i] =
				build_rhs_var_prop(v->bits,	/* base var */
						v->base_type,
						0,		/* offset */
						i,		/* which bit */
						v->phase,
						v->modifiers,
						force);
	}

	b->size = wordsize;
	return b;
    } else
	/* var_desig_size prints error message. */
	return make_integer_bit_vector(0, 0);
}



bit_vector eval_bit_vector_vectorize(bit_vector x, int force)
{
	int total_size = 0;
	int temp;
	vectorize_list temp_vec;
	bdd *result;

	/* Compute length of new vector and allocate storage. */
	temp_vec = x->vectorize_list1;
	while (temp_vec != NULL) {
	    if (temp_vec->type==DEFPROP) {
		/* It's an uneval_prop. */
		evaluate_proposition(temp_vec->prop_data,force);
		total_size++;
	    } else if (temp_vec->type==DEFVEC) {
		/* It's a bit_vector. */
		evaluate_bit_vector(temp_vec->vector_data,force);
		total_size += temp_vec->vector_data->size;
	    } else {
		yyerror("Error:  Illegal node type in vectorize.\n");
	    }

	    temp_vec = temp_vec->next;
	}
	result = (bdd *)calloc(total_size,sizeof(bdd));


	/* Copy bdds into new vector. */
	temp_vec = x->vectorize_list1;
	temp = total_size;	/* backwards since becomes flips order again.*/
	while (temp_vec != NULL) {
	    if (temp_vec->type==DEFPROP) {
		/* It's an uneval_prop. */
		evaluate_proposition(temp_vec->prop_data,force);
		temp--;
		result[temp] = bdd_identity(gbm,temp_vec->prop_data->bdd);
	    } else if (temp_vec->type==DEFVEC) {
		/* It's a bit_vector. */
		int i;

		evaluate_bit_vector(temp_vec->vector_data,force);
		temp -= temp_vec->vector_data->size;
		for (i=0; i < temp_vec->vector_data->size; i++) {
		    result[temp+i] = bdd_identity(gbm,
					temp_vec->vector_data->bit_list[i]);
		}
	    } else {
		fprintf(stderr,"Error:  Illegal node type in vectorize.\n");
	    }

	    temp_vec = temp_vec->next;
	}

	x->bit_list = result;
	x->size = total_size;
	return x;
}



bdd *add_bit_vectors_step(int asize, bdd *alist,
				  int bsize, bdd *blist)
	/* Helper function for add_bit_vectors. */
	/* Adds a into b, returns the sum, with the carry discarded. */
{
    bdd *result;
    bdd carry, sum;
    bdd temp1, temp2, temp3;
    int aindex, bindex, rindex;
    int rsize;

/*printf("Adding bit vectors\n");*/
    rsize = (asize>bsize)?asize:bsize;
    result = (bdd *)calloc(rsize,sizeof(bdd));

    carry = bdd_zero(gbm);	/* No carry at start. */
    for (aindex = asize-1, bindex = bsize-1, rindex = rsize-1;
	 rindex >=0;
	 aindex--, bindex--, rindex--) {

	if (aindex<0) {
	    sum = bdd_xor(gbm,blist[bindex], carry);
	    temp1 = bdd_and(gbm,blist[bindex],carry);
	    bdd_free(gbm,carry);
	    carry = temp1;
	} else if (bindex<0) {
	    sum = bdd_xor(gbm,alist[aindex], carry);
	    temp1 = bdd_and(gbm,alist[aindex],carry);
	    bdd_free(gbm,carry);
	    carry = temp1;
	} else {
	    temp1 = bdd_xor(gbm,alist[aindex],blist[bindex]);
	    sum = bdd_xor(gbm,temp1, carry);
	    bdd_free(gbm,temp1);
	    temp1 = bdd_and(gbm,alist[aindex],blist[bindex]);
	    temp2 = bdd_or(gbm,alist[aindex],blist[bindex]);
	    temp3 = bdd_ite(gbm,carry,temp2,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);
	    bdd_free(gbm,carry);
	    carry = temp3;
	}

	result[rindex] = sum;
    }
    bdd_free(gbm,carry);

    return result;
}

bit_vector eval_bit_vector_add(bit_vector b, int force)
	/* Note that addition is NOT associative the way we have
	   defined it.  In particular, (eq (add 1 1) 0) is true.
	   Therefore (add 1 1 1 1 1 100) is equal to 101, whereas
	   (add 100 1 1 1 1 1) equals 105.  This is because the
	   size of a sum is defined as the larger of the TWO operands.
	   The n-ary case is evaluated left-to-right. */
{
	bit_vector_list x;
	int xsize, sumsize;
	bdd *xlist, *sumlist, *newlist;
	int i;

	x = b->vec_list1;
	sumsize = 0;
	sumlist = (bdd *)calloc(sumsize,sizeof(bdd));
	while (x!=NULL) {
		evaluate_bit_vector(x->data, force);

		xsize = x->data->size;
		xlist = x->data->bit_list;

		/*printf("DEBUG 1 %d, %d\n",xsize,sumsize); fflush(stdout);*/
		newlist = add_bit_vectors_step(xsize, xlist, sumsize, sumlist);
		/* AJH 7 December 1996 */
		for (i=0; i<sumsize; i++) bdd_free(gbm,sumlist[i]);
		free(sumlist);
		sumlist = newlist;
		sumsize = (xsize>sumsize)?xsize:sumsize;

		x = x->next;
	}

	b->size = sumsize;
	b->bit_list = sumlist;
	return b;
}



bdd *sub_bit_vectors_step(int asize, bdd *alist,
				  int bsize, bdd *blist)
	/* Helper function for sub_bit_vectors. */
	/* Returns a-b with the borrow discarded. */
{
    bdd *result;
    bdd borrow, difference;
    bdd temp1, temp2, temp3;
    int aindex, bindex, rindex;
    int rsize;

/*printf("subtracting bit vectors\n");*/
    rsize = (asize>bsize)?asize:bsize;
    result = (bdd *)calloc(rsize,sizeof(bdd));

    borrow = bdd_zero(gbm);	/* No borrow at start. */
    for (aindex = asize-1, bindex = bsize-1, rindex = rsize-1;
	 rindex >=0;
	 aindex--, bindex--, rindex--) {

	if (aindex<0) {
	    difference = bdd_xor(gbm,blist[bindex], borrow);
	    temp1 = bdd_or(gbm,blist[bindex],borrow);
	    bdd_free(gbm,borrow);
	    borrow = temp1;
	} else if (bindex<0) {
	    difference = bdd_xor(gbm,alist[aindex], borrow);
	    temp1 = bdd_ite(gbm,alist[aindex],bdd_zero(gbm),borrow);
	    bdd_free(gbm,borrow);
	    borrow = temp1;
	} else {
	    temp1 = bdd_xor(gbm,alist[aindex],blist[bindex]);
	    difference = bdd_xor(gbm,temp1, borrow);
	    bdd_free(gbm,temp1);
	    temp1 = bdd_and(gbm,borrow,blist[bindex]);
	    temp2 = bdd_or(gbm,borrow,blist[bindex]);
	    temp3 = bdd_ite(gbm,alist[aindex],temp1,temp2);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);
	    bdd_free(gbm,borrow);
	    borrow = temp3;
	}

	result[rindex] = difference;
    }
    bdd_free(gbm,borrow);

    return result;
}

bit_vector eval_bit_vector_sub(bit_vector x, int force)
{
	bit_vector a = x->vector1;
	bit_vector b = x->vector2;
	prop_list temp;

	evaluate_bit_vector(a, force);
	evaluate_bit_vector(b, force);

	x->size = (a->size > b->size)?(a->size):(b->size);
	x->bit_list = sub_bit_vectors_step(a->size, a->bit_list,
					   b->size, b->bit_list);

	return x;
}


bit_vector eval_bit_vector_leftshift(bit_vector x, int force)
	/* Expands size of x to fit shifted result, rather than discarding
	   bits. */
{
	bit_vector a = x->vector1;
	bit_vector b = x->vector2;
	int i;

	evaluate_bit_vector(a, force);

	if (!b->is_constant) {
		fprintf(stderr,"Error:  Must leftshift by a constant. \n");
		return 0;
	}

	/* b is a constant */
	x->size = a->size + b->constant_value;
	x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));
	for (i=0; i < a->size; i++) {
		x->bit_list[i] = bdd_identity(gbm,a->bit_list[i]);
	}
	for (i=a->size; i < x->size; i++) {
		x->bit_list[i] = bdd_zero(gbm);
	}

	return x;
}

bit_vector eval_bit_vector_rightshift(bit_vector x, int force)
{
	bit_vector a = x->vector1;
	bit_vector b = x->vector2;
	int i;

	evaluate_bit_vector(a, force);

	if (!b->is_constant) {
		fprintf(stderr,"Error:  Must rightshift by a constant. \n");
		return 0;
	}

	/* b is a constant */
	x->size = a->size - b->constant_value;
	if (x->size < 0) x->size = 0;
	x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));
	for (i=0; i < x->size; i++) {
		x->bit_list[i] = bdd_identity(gbm,a->bit_list[i]);
	}

	return x;
}

bit_vector eval_bit_vector_leftrotate(bit_vector x, int force)
{
	bit_vector a = x->vector1;
	bit_vector b = x->vector2;
	int i,j;
	int rot;	/* rotation amount */

	evaluate_bit_vector(a, force);

	if (!b->is_constant) {
		fprintf(stderr,"Error:  Must leftrotate by a constant. \n");
		return 0;
	}

	/* b is a constant */
	rot = b->constant_value % a->size;
	x->size = a->size;
	x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));
	for (i=0,j=rot; j < a->size; i++,j++) {
		x->bit_list[i] = bdd_identity(gbm,a->bit_list[j]);
	}
	for (i=a->size-rot,j=0; i < x->size; i++,j++) {
		x->bit_list[i] = bdd_identity(gbm,a->bit_list[j]);
	}

	return x;
}

bit_vector eval_bit_vector_rightrotate(bit_vector x, int force)
{
	bit_vector a = x->vector1;
	bit_vector b = x->vector2;
	int i,j;
	int rot;	/* rotation amount */

	evaluate_bit_vector(a, force);

	if (!b->is_constant) {
		fprintf(stderr,"Error:  Must rightrotate by a constant. \n");
		return 0;
	}

	/* b is a constant */
	rot = b->constant_value % a->size;
	x->size = a->size;
	x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));
	for (i=0,j=rot; j < a->size; i++,j++) {
		x->bit_list[j] = bdd_identity(gbm,a->bit_list[i]);
	}
	for (i=a->size-rot,j=0; i < x->size; i++,j++) {
		x->bit_list[j] = bdd_identity(gbm,a->bit_list[i]);
	}

	return x;
}

bit_vector eval_bit_vector_bequiv(bit_vector x, int force)
{
    int i, j, k;
    bit_vector a = x->vector1;
    bit_vector b = x->vector2;
    bdd temp1, temp2;

    evaluate_bit_vector(a,force);
    evaluate_bit_vector(b,force);

    x->size = (a->size > b->size)?(a->size):(b->size);
    x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));

    for (i=a->size-1, j=b->size-1, k=x->size-1;
	    ((i>=0) || (j>=0));
	    i--, j--, k--) {
	temp1 = (i>=0) ? a->bit_list[i] : bdd_zero(gbm);
	temp2 = (j>=0) ? b->bit_list[j] : bdd_zero(gbm);
	x->bit_list[k] = bdd_xnor(gbm,temp1,temp2);
    }

    return x; /* I generally discard the return value, but... */
}

bit_vector eval_bit_vector_bimplies(bit_vector x, int force)
{
    int i, j, k;
    bit_vector a = x->vector1;
    bit_vector b = x->vector2;
    bdd temp1, temp2;

    evaluate_bit_vector(a,force);
    evaluate_bit_vector(b,force);

    x->size = (a->size > b->size)?(a->size):(b->size);
    x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));

    for (i=a->size-1, j=b->size-1, k=x->size-1;
	    ((i>=0) || (j>=0));
	    i--, j--, k--) {
	temp1 = (i>=0) ? a->bit_list[i] : bdd_zero(gbm);
	temp2 = (j>=0) ? b->bit_list[j] : bdd_zero(gbm);
	x->bit_list[k] = bdd_ite(gbm,temp1,temp2,bdd_one(gbm));
    }

    return x; /* I generally discard the return value, but... */
}

bit_vector eval_bit_vector_bxor(bit_vector x, int force)
{
    int i, j, k;
    bit_vector a = x->vector1;
    bit_vector b = x->vector2;
    bdd temp1, temp2;

    evaluate_bit_vector(a,force);
    evaluate_bit_vector(b,force);

    x->size = (a->size > b->size)?(a->size):(b->size);
    x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));

    for (i=a->size-1, j=b->size-1, k=x->size-1;
	    ((i>=0) || (j>=0));
	    i--, j--, k--) {
	temp1 = (i>=0) ? a->bit_list[i] : bdd_zero(gbm);
	temp2 = (j>=0) ? b->bit_list[j] : bdd_zero(gbm);
	x->bit_list[k] = bdd_xor(gbm,temp1,temp2);
    }

    return x; /* I generally discard the return value, but... */
}


bdd *band_bit_vectors_step(int asize, bdd *alist, int bsize, bdd *blist)
{
    int i, j, k;
    int rsize;
    bdd *result;

    rsize = (asize>bsize)?asize:bsize;
    result = (bdd *)calloc(rsize,sizeof(bdd));

    for (i=asize-1, j=bsize-1, k=rsize-1;
	    ((i>=0) || (j>=0));
	    i--, j--, k--) {
	if ((i<0) || (j<0)) {
	    result[k] = bdd_zero(gbm);
	} else {
	    result[k] = bdd_and(gbm,alist[i],blist[j]);
	}
    }

    return result;
}

bit_vector eval_bit_vector_band(bit_vector b, int force)
{
	bit_vector_list x = b->vec_list1;
	int xsize, resultsize;
	bdd *xlist, *resultlist, *newlist;
	int i;

	/* It's hard to write a clean band for zero-length list. */
	/* Fortunately, the syntax guarantees at least one bit vector. */
	evaluate_bit_vector(x->data, force);
	resultsize = x->data->size;
	resultlist = (bdd *)calloc(resultsize,sizeof(bdd));
	for (i=0; i<resultsize; i++)
		resultlist[i] = bdd_identity(gbm,x->data->bit_list[i]);
	x = x->next;

	while (x!=NULL) {
		evaluate_bit_vector(x->data, force);

		xsize = x->data->size;
		xlist = x->data->bit_list;

		newlist = band_bit_vectors_step(xsize, xlist,
						resultsize, resultlist);
		/* AJH 7 December 1996 */
		for (i=0; i<resultsize; i++) bdd_free(gbm,resultlist[i]);
		free(resultlist);
		resultlist = newlist;
		resultsize = (xsize>resultsize)?xsize:resultsize;

		x = x->next;
	}

	b->size = resultsize;
	b->bit_list = resultlist;
	return b;
}



bdd *bor_bit_vectors_step(int asize, bdd *alist, int bsize, bdd *blist)
{
    int i, j, k;
    int rsize;
    bdd *result;
    bdd temp1, temp2;

    rsize = (asize>bsize)?asize:bsize;
    result = (bdd *)calloc(rsize,sizeof(bdd));

    for (i=asize-1, j=bsize-1, k=rsize-1;
	    ((i>=0) || (j>=0));
	    i--, j--, k--) {
	temp1 = (i>=0) ? alist[i] : bdd_zero(gbm);
	temp2 = (j>=0) ? blist[j] : bdd_zero(gbm);
	result[k] = bdd_or(gbm,temp1,temp2);
    }

    return result;
}

bit_vector eval_bit_vector_bor(bit_vector b, int force)
{
	bit_vector_list x = b->vec_list1;
	int xsize, resultsize;
	bdd *xlist, *resultlist, *newlist;
	int i;

	resultsize = 0;
	resultlist = (bdd *)calloc(resultsize,sizeof(bdd));
	while (x!=NULL) {
		evaluate_bit_vector(x->data, force);

		xsize = x->data->size;
		xlist = x->data->bit_list;

		newlist = bor_bit_vectors_step(xsize, xlist,
						resultsize, resultlist);
		/* AJH 7 December 1996 */
		for (i=0; i<resultsize; i++) bdd_free(gbm,resultlist[i]);
		free(resultlist);
		resultlist = newlist;
		resultsize = (xsize>resultsize)?xsize:resultsize;

		x = x->next;
	}

	b->size = resultsize;
	b->bit_list = resultlist;
	return b;
}

bit_vector eval_bit_vector_bif(bit_vector x, int force)
{
    int i, j, k;
    uneval_prop cond = x->prop1;
    bit_vector thenpart = x->vector1;
    bit_vector elsepart = x->vector2;
    bdd temp1, temp2;

    evaluate_proposition(cond,force);
    evaluate_bit_vector(thenpart,force);
    evaluate_bit_vector(elsepart,force);

    x->size = (thenpart->size > elsepart->size)?
		    (thenpart->size):(elsepart->size);
    x->bit_list = (bdd *)calloc(x->size,sizeof(bdd));

    for (i=thenpart->size-1, j=elsepart->size-1, k=x->size-1;
	    ((i>=0) || (j>=0));
	    i--, j--, k--) {
	temp1 = (i>=0) ? thenpart->bit_list[i] : bdd_zero(gbm);
	temp2 = (j>=0) ? elsepart->bit_list[j] : bdd_zero(gbm);
	x->bit_list[k] = bdd_ite(gbm,cond->bdd,temp1,temp2);
    }

    return x;
}


bdd bdd_bit_vector_gt(bit_vector lhs, bit_vector rhs, int force)
{
    bdd iftie;
    bdd gt, eq, lt;
    bdd temp1, temp2;
    int lindex, rindex;

    evaluate_bit_vector(lhs,force);
    evaluate_bit_vector(rhs,force);

/*printf("bit vector gt\n");*/
    iftie = bdd_zero(gbm);	/* False if equality. */
    for (lindex = lhs->size-1, rindex = rhs->size-1;
	 (lindex >=0) || (rindex >=0);
	 lindex--, rindex--) {

	if (lindex<0) {
	    gt = bdd_zero(gbm);	/* lhs bit can't be bigger */
	    eq = bdd_not(gbm,rhs->bit_list[rindex]);
	    lt = bdd_identity(gbm,rhs->bit_list[rindex]);
	} else if (rindex<0) {
	    gt = bdd_identity(gbm,lhs->bit_list[lindex]);
	    eq = bdd_not(gbm,lhs->bit_list[lindex]);
	    lt = bdd_zero(gbm);	/* lhs bit can't be smaller */
	} else {
	    gt = bdd_ite(gbm,rhs->bit_list[rindex],
			 bdd_zero(gbm),
			 lhs->bit_list[lindex]);
	    eq = bdd_xnor(gbm,lhs->bit_list[lindex],rhs->bit_list[rindex]);
	    lt = bdd_ite(gbm,lhs->bit_list[lindex],
			 bdd_zero(gbm),
			 rhs->bit_list[rindex]);
	}

	bdd_free(gbm,lt);
	temp1 = bdd_and(gbm,eq,iftie);
	bdd_free(gbm,iftie);
	bdd_free(gbm,eq);
	temp2 = bdd_or(gbm,gt,temp1);
	bdd_free(gbm,gt);
	bdd_free(gbm,temp1);
	iftie = temp2;
    }

    return iftie;
}

bdd bdd_bit_vector_ge(bit_vector lhs, bit_vector rhs, int force)
{
    bdd iftie;
    bdd gt, eq, lt;
    bdd temp1, temp2;
    int lindex, rindex;

    evaluate_bit_vector(lhs,force);
    evaluate_bit_vector(rhs,force);

/*printf("bit vector ge\n");*/
    iftie = bdd_one(gbm);	/* True if equality. */
    for (lindex = lhs->size-1, rindex = rhs->size-1;
	 (lindex >=0) || (rindex >=0);
	 lindex--, rindex--) {

	if (lindex<0) {
	    gt = bdd_zero(gbm);	/* lhs bit can't be bigger */
	    eq = bdd_not(gbm,rhs->bit_list[rindex]);
	    lt = bdd_identity(gbm,rhs->bit_list[rindex]);
	} else if (rindex<0) {
	    gt = bdd_identity(gbm,lhs->bit_list[lindex]);
	    eq = bdd_not(gbm,lhs->bit_list[lindex]);
	    lt = bdd_zero(gbm);	/* lhs bit can't be smaller */
	} else {
	    gt = bdd_ite(gbm,rhs->bit_list[rindex],
			 bdd_zero(gbm),
			 lhs->bit_list[lindex]);
	    eq = bdd_xnor(gbm,lhs->bit_list[lindex],rhs->bit_list[rindex]);
	    lt = bdd_ite(gbm,lhs->bit_list[lindex],
			 bdd_zero(gbm),
			 rhs->bit_list[rindex]);
	}

	bdd_free(gbm,lt);
	temp1 = bdd_and(gbm,eq,iftie);
	bdd_free(gbm,iftie);
	bdd_free(gbm,eq);
	temp2 = bdd_or(gbm,gt,temp1);
	bdd_free(gbm,gt);
	bdd_free(gbm,temp1);
	iftie = temp2;
    }

    return iftie;
}

bdd bdd_bit_vector_eq(bit_vector lhs, bit_vector rhs, int force)
{
    int i, maxsize;
    bdd result;
    bdd temp1, temp2;

    evaluate_bit_vector(lhs, force);
    evaluate_bit_vector(rhs, force);

/*printf("bit vector eq\n");*/
    maxsize = lhs->size;
    if (rhs->size > maxsize) maxsize = rhs->size;

    result = bdd_one(gbm);

    for (i=0; i<maxsize; i++) {
	if (i > lhs->size-1) {
	    temp1 = bdd_not(gbm,rhs->bit_list[rhs->size-1-i]);
	    temp2 = bdd_and(gbm,result,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,result);
	    result = temp2;
	} else if (i > rhs->size-1) {
	    temp1 = bdd_not(gbm,lhs->bit_list[lhs->size-1-i]);
	    temp2 = bdd_and(gbm,result,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,result);
	    result = temp2;
	} else {
	    temp1 = bdd_xnor(gbm,rhs->bit_list[rhs->size-1-i],
			     lhs->bit_list[lhs->size-1-i]);
	    temp2 = bdd_and(gbm,result,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,result);
	    result = temp2;
	}
    }

    return result;
}

bdd bdd_bit_vector_le(bit_vector lhs, bit_vector rhs, int force)
{
    bdd iftie;
    bdd gt, eq, lt;
    bdd temp1, temp2;
    int lindex, rindex;

    evaluate_bit_vector(lhs,force);
    evaluate_bit_vector(rhs,force);

/*printf("bit vector le\n");*/
    iftie = bdd_one(gbm);	/* True if equality. */
    for (lindex = lhs->size-1, rindex = rhs->size-1;
	 (lindex >=0) || (rindex >=0);
	 lindex--, rindex--) {

	if (lindex<0) {
	    gt = bdd_zero(gbm);	/* lhs bit can't be bigger */
	    eq = bdd_not(gbm,rhs->bit_list[rindex]);
	    lt = bdd_identity(gbm,rhs->bit_list[rindex]);
	} else if (rindex<0) {
	    gt = bdd_identity(gbm,lhs->bit_list[lindex]);
	    eq = bdd_not(gbm,lhs->bit_list[lindex]);
	    lt = bdd_zero(gbm);	/* lhs bit can't be smaller */
	} else {
	    gt = bdd_ite(gbm,rhs->bit_list[rindex],
			 bdd_zero(gbm),
			 lhs->bit_list[lindex]);
	    eq = bdd_xnor(gbm,lhs->bit_list[lindex],rhs->bit_list[rindex]);
	    lt = bdd_ite(gbm,lhs->bit_list[lindex],
			 bdd_zero(gbm),
			 rhs->bit_list[rindex]);
	}

	bdd_free(gbm,gt);
	temp1 = bdd_and(gbm,eq,iftie);
	bdd_free(gbm,iftie);
	bdd_free(gbm,eq);
	temp2 = bdd_or(gbm,lt,temp1);
	bdd_free(gbm,lt);
	bdd_free(gbm,temp1);
	iftie = temp2;
    }

    return iftie;
}

bdd bdd_bit_vector_lt(bit_vector lhs, bit_vector rhs, int force)
{
    bdd iftie;
    bdd gt, eq, lt;
    bdd temp1, temp2;
    int lindex, rindex;

    evaluate_bit_vector(lhs,force);
    evaluate_bit_vector(rhs,force);

/*printf("bit vector lt\n");*/
    iftie = bdd_zero(gbm);	/* False if equality. */
    for (lindex = lhs->size-1, rindex = rhs->size-1;
	 (lindex >=0) || (rindex >=0);
	 lindex--, rindex--) {

	if (lindex<0) {
	    gt = bdd_zero(gbm);	/* lhs bit can't be bigger */
	    eq = bdd_not(gbm,rhs->bit_list[rindex]);
	    lt = bdd_identity(gbm,rhs->bit_list[rindex]);
	} else if (rindex<0) {
	    gt = bdd_identity(gbm,lhs->bit_list[lindex]);
	    eq = bdd_not(gbm,lhs->bit_list[lindex]);
	    lt = bdd_zero(gbm);	/* lhs bit can't be smaller */
	} else {
	    gt = bdd_ite(gbm,rhs->bit_list[rindex],
			 bdd_zero(gbm),
			 lhs->bit_list[lindex]);
	    eq = bdd_xnor(gbm,lhs->bit_list[lindex],rhs->bit_list[rindex]);
	    lt = bdd_ite(gbm,lhs->bit_list[lindex],
			 bdd_zero(gbm),
			 rhs->bit_list[rindex]);
	}

	bdd_free(gbm,gt);
	temp1 = bdd_and(gbm,eq,iftie);
	bdd_free(gbm,iftie);
	bdd_free(gbm,eq);
	temp2 = bdd_or(gbm,lt,temp1);
	bdd_free(gbm,lt);
	bdd_free(gbm,temp1);
	iftie = temp2;
    }

    return iftie;
}




bdd bdd_eq_int_bit_vector(int x, bit_vector y, int force)
{
    int mask;
    int which_bit;
    bdd result, temp1, temp2;

    evaluate_bit_vector(y, force);

/*printf("bit vector eq int\n");*/
    result = bdd_one(gbm);
    for (which_bit=y->size-1; which_bit>=0 ; which_bit--) {
	if (x & 1) {
	    temp1 = bdd_identity(gbm,y->bit_list[which_bit]);
	} else {
	    temp1 = bdd_not(gbm,y->bit_list[which_bit]);
	}
	temp2 = bdd_and(gbm,temp1, result);
	bdd_free(gbm,temp1);
	bdd_free(gbm,result);
	result = temp2;

	x /= 2;
    }

    if (x>0) {
	bdd_free(gbm,result);	/* x is bigger than y can possibly be. */
	return bdd_zero(gbm);
    } else return result;
}


bdd bdd_becomes_step(bdd *base,
				type_desc remaining_type,
				int pre_offset,
				var_accessor modchain,
				bdd *rhs_bits,
				int force)
{
    if (modchain==NULL) {
	/* We've reached the variable being accessed. */
	bdd result, temp1, temp2;
	bdd stability;
	bdd *unstable_list;
	bdd *temp_bit;
	int i;

	/* Build stability predicate. */
	temp1 = bdd_identity(gbm,current_next_eq->bdd);	/* Everything stable. */

	unstable_list = (bdd *)calloc(remaining_type->bdd_count/3 + 1,
					sizeof(bdd));
	unstable_list[remaining_type->bdd_count/3] = NULL;
						/* List Terminator. */
	for (i=0; i < remaining_type->bdd_count/3; i++) {
		/*unstable_list[2*i] = base[pre_offset+3*i+1]; /* Current phase */
		unstable_list[i] = base[pre_offset+3*i+2]; /* Next phase */
		/* I only have to quantify out the next phase variables. */
	}

	bdd_temp_assoc(gbm,unstable_list,0);
	bdd_assoc(gbm,-1);	/* temp association */
	stability = bdd_exists(gbm,temp1);    /* Destabilize selected list. */
	bdd_free(gbm,temp1);
	free(unstable_list);

	/* Build assignment. */
	i = remaining_type->bdd_count - 1; /* offset for the NEXT phase MSB */
	temp_bit = rhs_bits;
/*printf("building bdd becomes assignment\n");*/
	result = bdd_one(gbm);
	while (i>0) {
	    temp1 = bdd_xnor(gbm,base[pre_offset+i],*temp_bit);
	    temp2 = bdd_and(gbm,result,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,result);
	    result = temp2;
	    temp_bit++;
	    i -= 3;	/* skip to next variable */
	}

	temp1 = bdd_and(gbm,stability,result);
	bdd_free(gbm,stability);
	bdd_free(gbm,result);
	return temp1;
    }

    if ((modchain->type==ARRAY) && (!modchain->array_index->is_constant)) {
	bdd result, this_is_it, one_possibility;
	bdd temp1, temp2;
	int i, j;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

	result = bdd_zero(gbm);
	for (i=lb; i<=ub; i++) {
	    this_is_it = bdd_eq_int_bit_vector(i, modchain->array_index,force);
	    /* Suppose i==array_index, what happens? */
	    one_possibility = bdd_becomes_step(base,
				  element_type,
				  pre_offset+(i-lb)*element_type->bdd_count,
				  modchain->next,
				  rhs_bits,
				  force);
	    /* Assign value to these bits only when selected. */
	    temp1 = bdd_ite(gbm,this_is_it,one_possibility,bdd_zero(gbm));
	    bdd_free(gbm,this_is_it);
	    bdd_free(gbm,one_possibility);
	    temp2 = bdd_or(gbm,result,temp1);
	    bdd_free(gbm,result);
	    bdd_free(gbm,temp1);
	    result = temp2;
	}
	/* If array index is out-of-range, this will give no possible
	   successor state.  --AJH 19 July 1993 */

	return result;
    }

    if ((modchain->type==ARRAY) && (modchain->array_index->is_constant)) {
	bdd result;
	int i, j;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

	i = modchain->array_index->constant_value;

	/* Range checking constant value. */
	if ((lb > i) || (ub < i)) {
	    yyerror("Warning --- Constant array index out of range in lhs.");
	    result = bdd_zero(gbm);
	} else {
	    result = bdd_becomes_step(base,
			  element_type,
			  pre_offset+(i-lb)*element_type->bdd_count,
			  modchain->next,
			  rhs_bits,
			  force);
	}

	return result;
    }

    if (modchain->type==RECORD) {
	bdd result, temp1, temp2;
	record_list temp = remaining_type->field_list;
	int offset = 0;
	int i;

	result = bdd_one(gbm);
	while (temp!=NULL) {
	    if (strcmp(temp->field_name,modchain->field_select)==0) {
		/* This is the field being accessed. */
		temp1 = bdd_becomes_step(base,
					 temp->field_type,
					 pre_offset+offset,
					 modchain->next,
					 rhs_bits,
					 force);
		temp2 = bdd_and(gbm,result, temp1);
		bdd_free(gbm,temp1);
		bdd_free(gbm,result);
		result = temp2;
	    }

	    offset += temp->field_type->bdd_count;
	    temp = temp->next;
	}

	return result;
    }
}


bdd bdd_becomes(variable_desig lhs, bit_vector rhs, int force)
{
    bdd the_var;
    int i, lsize, adjuster;
    bdd *rhs_bits;

    /*fprintf(stderr,"\tEntering bdd_becomes\n"); fflush(stderr);*/
    if (lhs->fdv) {
	/* It's an assignment to a functionally dependent variable.
	   Therefore, ignore it when building transition relation. */
	/* GROSS HACK ALERT!!!  FIX THIS
	   This only works if no one relies on an updated value of
	   the dependent variable.  --AJH 10/18/92. */
	return bdd_identity(gbm,current_next_eq->bdd);
			/* bdd_identity added 16 Jul 1993 AJH */
    }

    /*fprintf(stderr,"\t\tEvaluating rhs\n"); fflush(stderr);*/
    evaluate_bit_vector(rhs, force);
    /*fprintf(stderr,"\t\tDone Evaluating rhs\n"); fflush(stderr);*/

    /* Make the left and right sides the same size. */
    lsize = var_desig_size(lhs->base_type, lhs->modifiers);
    rhs_bits = (bdd *)calloc(lsize,sizeof(bdd));
    adjuster = lsize - rhs->size;
    if (adjuster > 0)
	for (i=0; i<adjuster; i++) rhs_bits[i] = bdd_zero(gbm);
    for (i = (adjuster>0)?adjuster:0; i<lsize; i++)
	rhs_bits[i] = rhs->bit_list[i-adjuster];

    /*fprintf(stderr,"\t\tDoing the becomes step\n"); fflush(stderr);*/
    the_var = bdd_becomes_step(lhs->bits,	/* base var */
			       lhs->base_type,
			       0,		/* offset */
			       lhs->modifiers,	/* var accessor chain */
			       rhs_bits,	/* what we set it equal to */
			       force);
    /*fprintf(stderr,"\t\tDone doing the becomes step\n"); fflush(stderr);*/

    free(rhs_bits);	/* Dispose of chopped/padded copy of rhs->bit_list. */
    /*fprintf(stderr,"\t\tDone freeing rhs\n"); fflush(stderr);*/

    /*fprintf(stderr,"\t\tDone exiting bdd_becomes\n"); fflush(stderr);*/

    return the_var;
}



bdd bdd_constrain_step(bdd *base,
				type_desc remaining_type,
				int pre_offset,
				var_accessor modchain,
				bdd constraint,
				int force)
{
    if (modchain==NULL) {
	/* We've reached the variable being accessed. */
        bdd result, temp1;
        bdd stability;
        bdd *unstable_list;
        int i;

        /* Build stability predicate. */
        temp1 = bdd_identity(gbm,current_next_eq->bdd); /* Everything stable. */
/*fprintf(stderr,"DEBUG 1 temp1 = %d\n",temp1); fflush(stderr);*/

        unstable_list = (bdd *)calloc(remaining_type->bdd_count/3 + 1,
                                        sizeof(bdd));
/*fprintf(stderr,"DEBUG 2 unstable_list = %d, size = %d\n",unstable_list,remaining_type->bdd_count/3+1); fflush(stderr);*/
        unstable_list[remaining_type->bdd_count/3] = NULL;
                                                /* List Terminator. */
        for (i=0; i < remaining_type->bdd_count/3; i++) {
                /*unstable_list[2*i] = base[pre_offset+3*i+1];*//* Current phase
*/
                unstable_list[i] = base[pre_offset+3*i+2]; /* Next phase */
/*fprintf(stderr,"DEBUG 3 i = %d, pre_offset+3*i+2 = %d\n",i,pre_offset+3*i+2); fflush(stderr);*/
                /* I only have to quantify out the next phase variables. */
        }

	bdd_temp_assoc(gbm,unstable_list,0);
        bdd_assoc(gbm,-1);      /* temp association */
/*for (i=0; i<=remaining_type->bdd_count/3; i++) fprintf(stderr,"DEBUG \t %d\n",unstable_list[i]);
fflush(stderr);*/
        stability = bdd_exists(gbm,temp1);    /* Destabilize selected list. */
/*fprintf(stderr,"DEBUG 5 stability = %d, temp1 = %d\n",stability,temp1); fflush(stderr);*/
        bdd_free(gbm,temp1);
/*fprintf(stderr,"DEBUG 6 \n"); fflush(stderr);*/
        free(unstable_list);
/*fprintf(stderr,"DEBUG 7 constraint = %d\n",constraint); fflush(stderr);*/

	/* AND in the constraint. */
	result = bdd_and(gbm,stability,constraint);
	bdd_free(gbm,stability);

	return result;
    }

    if ((modchain->type==ARRAY) && (!modchain->array_index->is_constant)) {
	bdd result, this_is_it, one_possibility;
	bdd temp1, temp2;
	int i, j;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

	result = bdd_zero(gbm);	/* Result is FALSE if index out of range. */
	for (i=lb; i<=ub; i++) {
	    this_is_it = bdd_eq_int_bit_vector(i, modchain->array_index,force);
	    /* Suppose i==array_index, what happens? */
	    one_possibility = bdd_constrain_step(base,
				  element_type,
				  pre_offset+(i-lb)*element_type->bdd_count,
				  modchain->next,
				  constraint,
				  force);
            /* Assign value to these bits only when selected. */
            temp1 = bdd_ite(gbm,this_is_it,one_possibility,bdd_zero(gbm));
	    bdd_free(gbm,this_is_it);
	    bdd_free(gbm,one_possibility);

            temp2 = bdd_or(gbm,result,temp1);
            bdd_free(gbm,result);
            bdd_free(gbm,temp1);
            result = temp2;
        }

	return result;
    }

    if ((modchain->type==ARRAY) && (modchain->array_index->is_constant)) {
	bdd result;
	int i, j;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

        i = modchain->array_index->constant_value;

        /* Range checking constant value. */
        if ((lb > i) || (ub < i)) {
            yyerror("Warning --- Constant array index out of range in lhs.");
            result = bdd_zero(gbm);
        } else {
            result = bdd_constrain_step(base,
                          element_type,
                          pre_offset+(i-lb)*element_type->bdd_count,
                          modchain->next,
                          constraint,
                          force);
        }
	
	return result;
    }

    if (modchain->type==RECORD) {
	bdd result, temp1, temp2;
	record_list temp = remaining_type->field_list;
	int offset = 0;
	int i;

	result = bdd_one(gbm);
	while (temp!=NULL) {
	    if (strcmp(temp->field_name,modchain->field_select)==0) {
		/* This is the field being accessed. */
		temp1 = bdd_constrain_step(base,
					 temp->field_type,
					 pre_offset+offset,
					 modchain->next,
					 constraint,
					 force);
		temp2 = bdd_and(gbm,result, temp1);
		bdd_free(gbm,temp1);
		bdd_free(gbm,result);
		result = temp2;
	    }

	    offset += temp->field_type->bdd_count;
	    temp = temp->next;
	}

	return result;
    }
}


bdd bdd_constrain(variable_desig lhs, uneval_prop rhs, int force)
{
    bdd the_var;

    if (lhs->fdv) {
	/* It's an assignment to a functionally dependent variable.
	   Therefore, ignore it when building transition relation. */
	fprintf(stderr,"Warning:  constrain of dependent variable not implemented.\n");
	return bdd_identity(gbm,current_next_eq->bdd);
			/* bdd_identity added 16 Jul 1993 AJH */
    }

    evaluate_proposition(rhs, force);

    the_var = bdd_constrain_step(lhs->bits,	/* base var */
			       lhs->base_type,
			       0,		/* offset */
			       lhs->modifiers,	/* var accessor chain */
			       rhs->bdd,	/* constraint on that var */
			       force);

    return the_var;
}



bdd compose_nextstates(prop_list nextstates, int force)
{
    bdd temp1, temp2, temp3, temp4, temp5;
    bdd temp6, temp7, temp8, temp9, temp10;

    if (nextstates==NULL) return bdd_identity(gbm,current_next_eq->bdd);

    if (nextstates->next==NULL)
	return bdd_identity(gbm,
			evaluate_proposition(nextstates->data,force)->bdd);

    /* Functional composition is associative, last time I checked. */

    temp1 = evaluate_proposition(nextstates->data,force)->bdd;
    temp2 = compose_nextstates(nextstates->next, force);

/*printf("composing nextstate relations\n");*/
/*printf("\tprepending line %d\n",nextstates->data->lineno);*/
/*printf("\toriginal sizes %ld and %ld\n",bdd_size_1(temp1),bdd_size_1(temp2));*/
/*printf("\tchange to previous phase\n");*/
    bdd_assoc(gbm,cur_vars_assoc);
    temp4 = bdd_rel_prod(gbm,temp1,previous_current_eq->bdd);

/*printf("\tchange to current phase\n");*/
    bdd_assoc(gbm,next_vars_assoc);
    temp6 = bdd_rel_prod(gbm,temp4,current_next_eq->bdd);
	bdd_free(gbm,temp4);

/*printf("\tAND the two (sizes %ld and %ld)\n",bdd_size_1(temp2),bdd_size_1(temp6));*/
    bdd_assoc(gbm,cur_vars_assoc);
    temp8 = bdd_rel_prod(gbm,temp2,temp6);
	bdd_free(gbm,temp6);
	bdd_free(gbm,temp2);
/*printf("\t\t\toutput size %ld\n",bdd_size_1(temp8));*/

/*printf("\tchange to current phase\n");*/
    bdd_assoc(gbm,prev_vars_assoc);
    temp10 = bdd_rel_prod(gbm,temp8,previous_current_eq->bdd);
	bdd_free(gbm,temp8);
/*printf("\t\t\toutput size %ld\n",bdd_size_1(temp10));*/
/*printf("\tdone composing\n");*/

    return temp10;
}


bdd repeat_nextstate(int count, uneval_prop nextstate, int force)
	/* Creates transition relation that repeats nextstate count times. */
{
    bdd result, body;
    bdd temp1, temp2, temp3;

    if (nextstate==NULL) return bdd_identity(gbm,current_next_eq->bdd);

    if (count<=0) return bdd_identity(gbm,current_next_eq->bdd);

    body = evaluate_proposition(nextstate, force)->bdd;

    /* First repetition is special */
    result = bdd_identity(gbm,body);
    count--;
    while (count>0) {
	/* Loop Invariant:  result needs count more copies of body. */

	/* Change result to prev_cur phase*/
	bdd_assoc(gbm,cur_vars_assoc);
	temp1 = bdd_rel_prod(gbm,result,previous_current_eq->bdd);
	bdd_free(gbm,result);

	bdd_assoc(gbm,next_vars_assoc);
	temp2 = bdd_rel_prod(gbm,temp1,current_next_eq->bdd);
	bdd_free(gbm,temp1);

	/* Compose result and body */
	bdd_assoc(gbm,cur_vars_assoc);
	temp3 = bdd_rel_prod(gbm,temp2,body);
	bdd_free(gbm,temp2);

	/* Change result to cur_next phase*/
	bdd_assoc(gbm,prev_vars_assoc);
	result = bdd_rel_prod(gbm,temp3,previous_current_eq->bdd);
	bdd_free(gbm,temp3);

	count--;
    }
    /* Don't need to bdd_free body because it wasn't bdd_identity'ed. */

    return result;
}



/* Exported Routines */

uneval_prop evaluate_proposition(uneval_prop p, int force)
	/* force forces reevaluation */
{
    bdd bdd_result;
    bdd temp1;
    prop_list temppl;

    if (p->type == EVAL) {
	/*fprintf(stderr,"(Predefined)\n"); fflush(stderr);*/ /* DEBUG */
	return p;	/* Leaf node. */
    }

    if ((p->bdd != NULL)&&(!force)) {
	/*fprintf(stderr,"(Already Eval)\n"); fflush(stderr);*/ /* DEBUG */
	return p;	/* Already been evaluated. */
		/* Is it possible that we will want to reevaluate something? */
    }

    if (p->bdd != NULL) {
	bdd_free(gbm,p->bdd);
    }

    /*fprintf(stderr,"(Starting type %d\n",p->type); fflush(stderr);*/ /* DEBUG */
    switch (p->type) {
    case ID:	/* Delete this case, it will never happen. --AJH 12/11/91 */
	if (p->pred1->is_lambda) {
	    yyerror("Evaluator Error:  lambda predicates are unimplemented.");
	    return false_proposition;
	} else {
/*printf("plugging in actuals\n");*/
	    /*plug_in_actuals(p->pred1->bound_prop,p->vec_list1);*/
	    p->bdd = bdd_identity(gbm,
			evaluate_proposition(p->pred1->bound_prop,TRUE)->bdd
		     );
/*printf("unplugging actuals\n");*/
	    /*unplug_actuals(p->pred1->bound_prop);*//* Unsubstitute the args */
	}
	break;
    case AND:
	temppl = p->prop_list1;
	bdd_result = bdd_one(gbm);
	while (temppl!=NULL) {
/*printf("ANDing\n");*/
	    temp1 = bdd_and(gbm,bdd_result,
			    evaluate_proposition(temppl->data,force)->bdd);
	    bdd_free(gbm,bdd_result);
	    bdd_result = temp1;
	    temppl = temppl->next;
	}
	p->bdd = bdd_result;
	break;
    case OR:
	temppl = p->prop_list1;
	bdd_result = bdd_zero(gbm);
	while (temppl!=NULL) {
/*printf("ORing\n");*/
	    temp1 = bdd_or(gbm,bdd_result,
			   evaluate_proposition(temppl->data,force)->bdd);
	    bdd_free(gbm,bdd_result);
	    bdd_result = temp1;
	    temppl = temppl->next;
	}
	p->bdd = bdd_result;
	break;
    case NOT:
	p->bdd = bdd_not(gbm,evaluate_proposition(p->prop1,force)->bdd);
	break;
    case XOR:
/*printf("XORing\n");*/
	p->bdd = bdd_xor(gbm,evaluate_proposition(p->prop1,force)->bdd,
			 evaluate_proposition(p->prop2,force)->bdd);
	break;
    case EQUIV:
/*printf("XNORing\n");*/
	p->bdd = bdd_xnor(gbm,evaluate_proposition(p->prop1,force)->bdd,
			 evaluate_proposition(p->prop2,force)->bdd);
	break;
    case IMPLIES:
/*printf("Implication\n");*/
	p->bdd = bdd_ite(gbm,evaluate_proposition(p->prop1,force)->bdd,
			 evaluate_proposition(p->prop2,force)->bdd,
			 bdd_one(gbm));
	break;
    case IF:
/*printf("ite\n");*/
	p->bdd = bdd_ite(gbm,evaluate_proposition(p->prop1,force)->bdd,
			 evaluate_proposition(p->prop2,force)->bdd,
			 evaluate_proposition(p->prop3,force)->bdd);
	break;
    case GT:
	p->bdd = bdd_bit_vector_gt(p->vector1, p->vector2, force);
	break;
    case GE:
	p->bdd = bdd_bit_vector_ge(p->vector1, p->vector2, force);
	break;
    case EQ:
	p->bdd = bdd_bit_vector_eq(p->vector1, p->vector2, force);
	break;
    case LE:
	p->bdd = bdd_bit_vector_le(p->vector1, p->vector2, force);
	break;
    case LT:
	p->bdd = bdd_bit_vector_lt(p->vector1, p->vector2, force);
	break;
    case HIDEPREV:
/*printf("hideprev\n");*/
	evaluate_proposition(p->prop1,force);
	bdd_assoc(gbm,prev_vars_assoc);
	p->bdd = bdd_exists(gbm,p->prop1->bdd);
	break;
    case HIDECUR:
/*printf("hidecur\n");*/
	evaluate_proposition(p->prop1,force);
	bdd_assoc(gbm,cur_vars_assoc);
	p->bdd = bdd_exists(gbm,p->prop1->bdd);
	break;
    case HIDENEXT:
/*printf("hidenext\n");*/
	evaluate_proposition(p->prop1,force);
	bdd_assoc(gbm,next_vars_assoc);
	p->bdd = bdd_exists(gbm,p->prop1->bdd);
	break;
    case BECOMES:
	p->bdd = bdd_becomes(p->var1, p->vector1, force);
	break;
    case CONSTRAIN:
	p->bdd = bdd_constrain(p->var1, p->prop1, force);
	break;
    case COMPOSE:
	p->bdd = compose_nextstates(p->prop_list1, force);
	break;
    case REPEAT:
	p->bdd = repeat_nextstate(p->count,p->prop1, force);
	break;
    case FORALL:
    case EXISTS:
    case REACHABLE:
    case REACHGOAL:
    case AG:
    case EG:
    case AF:
    case EF:
    case AX:
    case EX:
    case AU:
    case EU:
    case SIMREL:
    case ANDCOMPRESS:
    case ORCOMPRESS:
    default:
	yyerror("Evaluator Error:  Unimplemented boolean operator.");
	fprintf(stderr,"-----Operator type = %d.\n",p->type);
	return false_proposition;
    }

    /*fprintf(stderr,"Done type %d, line %d, BDD Size %ld )\n",p->type,p->lineno,bdd_size(gbm,p->bdd,1)); fflush(stderr); /* DEBUG */
    return p;
}



bit_vector evaluate_bit_vector(bit_vector b, int force)
	/* force forces reevaluation */
{
    int i;

    if (b->type == EVAL) {
	return b;	/* Leaf node. */
    }

    if ((b->bit_list != NULL)&&(!force)) {
	return b;	/* Already been evaluated. */
    }

    if (b->bit_list != NULL) free(b->bit_list);

    switch (b->type) {
    case DEFVAR:	/* Variable access */
	b = eval_bit_vector_var_access(b,force);
	break;
    case VECTOR:
	b = eval_bit_vector_vectorize(b,force);
	break;
    case ADD:
	b = eval_bit_vector_add(b,force);
	break;
    case SUB:
	b = eval_bit_vector_sub(b,force);
	break;
    case LEFTSHIFT:
	b = eval_bit_vector_leftshift(b,force);
	break;
    case RIGHTSHIFT:
	b = eval_bit_vector_rightshift(b,force);
	break;
    case LEFTROTATE:
	b = eval_bit_vector_leftrotate(b,force);
	break;
    case RIGHTROTATE:
	b = eval_bit_vector_rightrotate(b,force);
	break;
    case BEQUIV:
	b = eval_bit_vector_bequiv(b,force);
	break;
    case BIMPLIES:
	b = eval_bit_vector_bimplies(b,force);
	break;
    case BXOR:
	b = eval_bit_vector_bxor(b,force);
	break;
    case BAND:
	b = eval_bit_vector_band(b,force);
	break;
    case BOR:
	b = eval_bit_vector_bor(b,force);
	break;
    case BNOT:
	b->vector1 = evaluate_bit_vector(b->vector1,force);
	b->size = b->vector1->size;
	b->bit_list = (bdd *)calloc(b->size,sizeof(bdd));
	for (i = 0; i < b->size; i++)
		b->bit_list[i] = bdd_not(gbm,b->vector1->bit_list[i]);
	break;
    case BIF:
	b = eval_bit_vector_bif(b,force);
	break;
    default:
	yyerror("Evaluator Error:  Unimplemented bit vector operator.");
	fprintf(stderr,"-----Operator type = %d.\n",b->type);
    }

    return b;
}
