/* Copyright 1991 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Routines For Image Computation */
/* By Alan J. Hu */
/* August 6, 1991 */
/* Split into Separate File on September 11, 1994. */

/* ************************ CHANGE LOG ************************** */
/*
11 May 1994 AJH -- Converted compute_backward_image() to use bdd_ite() for
		   the ternary if-then-else branch.
*/
/* ************************************************************** */

#include "ever.h"
#include "everlist.h"
#include "y.tab.h"	/* Defines constants for the tokens. */

/* More helper functions for printtrace.  These do next state computations. */


bdd compute_forward_image(bdd x, uneval_prop next, int force)
	/* Computes image next(x).  If you modify this, be sure to
	   update compute_backward_image as well. */
{
    bdd temp1, temp2, temp3;
    prop_list tp;

    if ((next->bdd != NULL) && (!force)) {
	/* Already evaluated, let's use it. */
	bdd_assoc(gbm,cur_vars_assoc);
	temp1 = bdd_rel_prod(gbm,x,next->bdd);

	return temp1;
    }

    if (next->type == OR) {
	/* Unevaluated disjunction. */
	temp1 = bdd_zero(gbm);
	tp = next->prop_list1;
	while (tp!=NULL) {
	    /* Compute forward image for each disjunct. */
	    temp2 = compute_forward_image(x, tp->data, force);
	    temp3 = bdd_or(gbm,temp1, temp2);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);
	    temp1 = temp3;
	    tp = tp->next;
	}

	return temp1;
    }

    if (next->type == IF) {
	/* Ternary if-then-else.  Should check here that condition only
	   operates on current phase, else this might not work. */
	evaluate_proposition(next->prop1,FALSE);

	temp1 = bdd_and(gbm,x,next->prop1->bdd);
	temp2 = compute_forward_image(temp1, next->prop2, force);/* then part */
	    bdd_free(gbm,temp1);
	temp1 = bdd_not(gbm,next->prop1->bdd);
	temp3 = bdd_and(gbm,x,temp1);
	    bdd_free(gbm,temp1);
	temp1 = compute_forward_image(temp3, next->prop3, force);/* else part */
	    bdd_free(gbm,temp3);

	temp3 = bdd_or(gbm,temp1, temp2);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);

	return temp3;
    }

    if (next->type == COMPOSE) {
	/* Function composition. */

	tp = next->prop_list1;
	temp1 = bdd_identity(gbm,x);
	while (tp != NULL) {
	    temp2 = compute_forward_image(temp1, tp->data, force);
		bdd_free(gbm,temp1);

	    /* If we're going to loop again, switch phase back. */
	    tp = tp->next;
	    if (tp != NULL) {
		bdd_assoc(gbm,next_vars_assoc);
		temp1 = bdd_rel_prod(gbm,temp2,current_next_eq->bdd);
		    bdd_free(gbm,temp2);
	    }
	}

	return temp2;
    }

    /* Talking to Jerry, got an idea 11/92.  You can do the BECOMES one
	bit at a time, if a single BECOMES becomes too large.  Maybe.  AJH */

    /* We don't know how to do this type as unevaluated, so do it directly. */
    fprintf(stderr,"Evaluating at type %d, line %d.\n",next->type,next->lineno);
    fflush(stderr);
    evaluate_proposition(next, force);
    bdd_assoc(gbm,cur_vars_assoc);
    temp2 = bdd_rel_prod(gbm,x,next->bdd);

    return temp2;
}


/* Forward declaration */
bdd compute_backward_image(bdd y, uneval_prop next);

bdd backward_image_compose_step(prop_list nextlist, bdd y)
	/* Helper function for compute_backward_image(). */
{
    bdd temp1, temp2;

/*printf("<"); fflush(stdout);	/* DEBUG */
    /*printf("SHRIMP:\tCompose Step Input %ld\n",bdd_size(gbm,y,1));*/
    if (nextlist==NULL) return bdd_identity(gbm,y);

/*printf("a"); fflush(stdout);	/* DEBUG */
    temp1 = backward_image_compose_step(nextlist->next, y);
/*printf("b"); fflush(stdout);	/* DEBUG */
    temp2 = compute_backward_image(temp1, nextlist->data);
/*printf("c"); fflush(stdout);	/* DEBUG */
    /*printf("SHRIMP:\tCompose Step %ld becomes %ld becomes %ld\n",bdd_size(gbm,y,1),
		bdd_size(gbm,temp1,1),bdd_size(gbm,temp2,1));*/
	bdd_free(gbm,temp1);
    bdd_assoc(gbm,cur_vars_assoc);
    temp1 = bdd_rel_prod(gbm,temp2,current_next_eq->bdd);
/*printf("d"); fflush(stdout);	/* DEBUG */
	bdd_free(gbm,temp2);
/*printf(">"); fflush(stdout);	/* DEBUG */
    return temp1;
}

bdd compute_backward_image(bdd y, uneval_prop next)
	/* Computes image pred(y).  If you modify this, be sure to
	   update compute_forward_image as well. */
{
    bdd temp1, temp2, temp3, temp4;
    prop_list tp;

    if (next->bdd != NULL) {
	/* Already evaluated, let's use it. */
/*printf("Dippy-Doo rel_prod with sizes %ld and %ld.\n",bdd_size(gbm,y,1),bdd_size(gbm,next->bdd,1));*/
	bdd_assoc(gbm,next_vars_assoc);
	temp1 = bdd_rel_prod(gbm,y,next->bdd);

/*printf("*"); fflush(stdout);*/	/* DEBUG */
	/*printf("SHRIMP:  Evaluated %ld becomes %ld\n",bdd_size(gbm,y,1),bdd_size(gbm,temp1,1));*/
	return temp1;
    }

    if (next->type == OR) {
	/* Unevaluated disjunction. */
	printf("SHRIMP:  Or %ld\n",bdd_size(gbm,y,1));
/*printf("["); fflush(stdout);*/	/* DEBUG */
	temp1 = bdd_zero(gbm);
	tp = next->prop_list1;
	while (tp!=NULL) {
	    /* Compute backward image for each disjunct. */
/*printf("."); fflush(stdout);*/	/* DEBUG */
	    temp2 = compute_backward_image(y, tp->data);
/*printf("o"); fflush(stdout);*/	/* DEBUG */
	    temp3 = bdd_or(gbm,temp1, temp2);
	    printf("SHRIMP:\t\t %ld %ld\n",bdd_size(gbm,temp2,1),bdd_size(gbm,temp3,1));
/*bdd_print_profile(gbm,temp3,bdd_naming_fn_full,NULL,80,stdout);*/
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);
	    temp1 = temp3;
	    tp = tp->next;
	}
	printf("SHRIMP:  End Or\n");
/*printf("]"); fflush(stdout);*/	/* DEBUG */

	return temp1;
    }

    if (next->type == IF) {
	/* Ternary if-then-else.  Should check here that condition only
	   operates on current phase, else this might not work. */
	/*printf("SHRIMP:  If %ld\n",bdd_size(gbm,y,1));*/
/*printf("("); fflush(stdout);*/	/* DEBUG */
	evaluate_proposition(next->prop1,FALSE);

	/* 11 May 1994 AJH */
/*printf("1"); fflush(stdout);	/* DEBUG */
	temp2 = compute_backward_image(y, next->prop2);	/* then part */
/*printf("2"); fflush(stdout);	/* DEBUG */
	temp3 = compute_backward_image(y, next->prop3);	/* else part */
/*printf("3"); fflush(stdout);	/* DEBUG */
	temp1 = bdd_ite(gbm,next->prop1->bdd,temp2,temp3);
	/*printf("SHRIMP:  Then %ld Else %ld Result %ld\n",bdd_size(gbm,temp2,1),
		bdd_size(gbm,temp3,1),bdd_size(gbm,temp1,1));*/
	    bdd_free(gbm,temp2);
	    bdd_free(gbm,temp3);

	/* OBSOLETE CODE
	/*temp1 = compute_backward_image(y, next->prop2); /* then part */
	/*temp2 = bdd_and(gbm,temp1, next->prop1->bdd);
	    /*bdd_free(gbm,temp1);
	/*temp1 = compute_backward_image(y, next->prop3); /* else part */
	/*temp3 = bdd_not(gbm,next->prop1->bdd);
	/*temp4 = bdd_and(gbm,temp1, temp3);
	    /*bdd_free(gbm,temp1);
	    /*bdd_free(gbm,temp3);
	/*temp1 = bdd_or(gbm,temp2, temp4);
	    /*bdd_free(gbm,temp2);
	    /*bdd_free(gbm,temp4);
	*/

	/*printf("SHRIMP:  End If\n");*/
/*printf(")"); fflush(stdout);	/* DEBUG */
	return temp1;
    }

    if (next->type == COMPOSE) {
	/* Function composition. */
	/*printf("SHRIMP:  Compose\n");*/
/*printf("{"); fflush(stdout);	/* DEBUG */

	temp1 = backward_image_compose_step(next->prop_list1, y);
/*printf(","); fflush(stdout);	/* DEBUG */
	/* Switch phase back to current phase. */
	bdd_assoc(gbm,next_vars_assoc);
/*printf(","); fflush(stdout);	/* DEBUG */
	temp2 = bdd_rel_prod(gbm,temp1, current_next_eq->bdd);
/*printf(","); fflush(stdout);	/* DEBUG */
	    bdd_free(gbm,temp1);
/*printf("}"); fflush(stdout);	/* DEBUG */
	/*printf("SHRIMP:  End Compose\n");*/
	return temp2;
    }

    /* We don't know how to do this type as unevaluated, so do it directly. */
    fprintf(stderr,"Evaluating at type %d, line %d.\n",next->type,next->lineno);
    fflush(stderr);
    evaluate_proposition(next, FALSE);
printf("Ha Ha rel_prod with sizes %ld and %ld.\n",bdd_size(gbm,y,1),bdd_size(gbm,next->bdd,1));
    bdd_assoc(gbm,next_vars_assoc);
    temp1 = bdd_rel_prod(gbm,y,next->bdd);

    return temp1;
}



conjunct_list over_approx_forward_image_conjunct_list(uneval_prop next,
						      conjunct_list x)
	/* Forward image of a conjunct list is not easy if we don't want
	   to evaluate the conjunction.  Instead, I'll over-approximate
	   Image(x1&x2&...&xn) as Image(x1)&...&Image(xn). --AJH 5 May 1994 */
{
	int i;
	bdd temp1;
	conjunct_list new = new_conjunct_list(x->size);

	printf("Zoom!");fflush(stdout);
	for (i=0; i < x->size; i++) {
		temp1 = compute_forward_image(x->conjunct[i],next,FALSE);

		/* Change phase back to current phase. */
		bdd_assoc(gbm,next_vars_assoc);
		new->conjunct[i] = bdd_rel_prod(gbm,temp1,current_next_eq->bdd);
		bdd_free(gbm,temp1);
	}
	printf("Vroom!");fflush(stdout);

	return new;
}


bdd pre_image_bdd(uneval_prop next, bdd x)
    /* Returns the PreImage of x under next. */
    /* Does the phase conversion and then calls compute_forward_image. */
{
    bdd new;
    bdd temp1, temp2;

    /* Convert phase to next phase. */
    bdd_assoc(gbm,cur_vars_assoc);
    temp1 = bdd_rel_prod(gbm, x, current_next_eq->bdd);

    new = compute_backward_image(temp1, next);
    bdd_free(gbm,temp1);

    return new;
}



bdd inv_image_bdd(uneval_prop next, bdd x)
	/* Returns the BackImage of x under next. */
	/* BackImage is the largest set all of whose successors are in x. */
{
    bdd new;
    bdd temp1, temp2;

    /* Can compute back_image(P) = not pre_image(not P). */
    /* Terminology has changed since I first wrote this stuff. */

    temp1 = bdd_not(gbm,x);

    /* Convert phase to next phase. */
    bdd_assoc(gbm,cur_vars_assoc);
    temp2 = bdd_rel_prod(gbm,temp1, current_next_eq->bdd);
    bdd_free(gbm,temp1);

    temp1 = compute_backward_image(temp2,next);
    bdd_free(gbm,temp2);
    new = bdd_not(gbm,temp1);
    bdd_free(gbm,temp1);

    return new;
}

conjunct_list inv_image(uneval_prop next, conjunct_list x)
	/* Returns the inverse image of x under next. */
	/* Inverse image is defined as the largest set all of whose
		successors are in x. */
{
    int i;
    conjunct_list new = new_conjunct_list(x->size);

    /* Inverse image distributes over conjunction. */
    for (i=0; i < x->size; i++) {
printf("DEBUG 1 %d\n",i);
	new->conjunct[i] = inv_image_bdd(next,x->conjunct[i]);
    }

    return new;
}

/* Forward declaration. */
conjunct_list inv_image_ic(uneval_prop next, conjunct_list x);


conjunct_list inv_image_ic_compose_step(prop_list next_list,
					conjunct_list x)
	/* Helper function for inv_image_ic(). */
{
    conjunct_list tail, result;

    /* Base Case */
    if (next_list==NULL) return copy_conjunct_list(x);

    /* Recursive BackImage on tail of list. */
    tail = inv_image_ic_compose_step(next_list->next, x);
    fast_munch_conjunct_list(tail);

    /* BackImage on head of list. */
    result = inv_image_ic(next_list->data, tail);
    free_conjunct_list(tail);
    fast_munch_conjunct_list(result);

    return result;
}

conjunct_list inv_image_ic(uneval_prop next, conjunct_list x)
	/* Computes "inverse image" (now called backward image) of
	   x under next, but recursively descends the nextstate relation,
	   and doesn't recombine intermediate results, yielding a really
	   long implicit conjunction as the result. */
{
    if (next->bdd != NULL) {
	/* Already been evaluated before. */
	return inv_image(next,x);
    }

    if (next->type == OR) {
	/* OR -- I can do each disjunct separately. */
	prop_list temppl;
	conjunct_list result, temp1, temp2;

	temppl = next->prop_list1;
	result = new_conjunct_list(0);
	while (temppl!=NULL) {
	    temp1 = inv_image_ic(temppl->data,x); /* Image under this rule. */
	    temp2 = concat_conjunct_lists(temp1, result);
	    free_conjunct_list(temp1);
	    free_conjunct_list(result);

	    /* Munch the list to keep from having to do a huge one later. */
	    /* Skip this for now. */
	    /*temp1 = evaluate_some_conjunctions(temp2);
	    free_conjunct_list(temp2);*/
	    temp1 = temp2;	/* Bypass for skipped code. */
	    fast_munch_conjunct_list(temp1);

	    result = temp1;

	    temppl = temppl->next;
	}

	return result;
    }

    if (next->type == IF) {
	/* Ternary if-then-else.  Should check here that condition only
	   operates on current phase, else this might not work. */
	/* BackImage(if b then f else g, x) =
		(b and BackImage(f,x)) or (~b and BackImage(g,x))  also =
		(~b or BackImage(f,x)) and (b or BackImage(g,x)) */
	/* Oooh, this is cool.  In general, if delta completely ignores x,
		then you want BackImage(delta,x)==x.  But, in this case,
		the parts of the BackImage will care about b, even if
		b, f, and g are completely orthogonal to x.  This can
		be a problem as you continue to propagate the b and ~b
		conjuncts separately.  Seems like we need to check for
		good BDD evals to do at this point. Dec 7, 1994 */

	conjunct_list thenpart, elsepart, result;
	bdd temp1;
	int i;

	evaluate_proposition(next->prop1,FALSE);

	/* then part */
	thenpart = inv_image_ic(next->prop2, x);
	for (i=0; i < thenpart->size; i++) {
	    temp1 = bdd_ite(gbm,next->prop1->bdd,
				thenpart->conjunct[i],
				bdd_one(gbm));
		    /* Computes ~b + thenpart[i] */
	    bdd_free(gbm,thenpart->conjunct[i]);
	    thenpart->conjunct[i] = temp1;
	}

	/* else part */
	elsepart = inv_image_ic(next->prop3, x);
	for (i=0; i < elsepart->size; i++) {
	    temp1 = bdd_or(gbm,next->prop1->bdd, elsepart->conjunct[i]);
	    bdd_free(gbm,elsepart->conjunct[i]);
	    elsepart->conjunct[i] = temp1;
	}

	result = concat_conjunct_lists(thenpart, elsepart);
	free_conjunct_list(thenpart);
	free_conjunct_list(elsepart);
	fast_munch_conjunct_list(result);
	medium_munch_conjunct_list(result);
	slow_munch_conjunct_list(result,1.0);

	return result;
    }

    if (next->type == COMPOSE) {
	/* Sequential Composition. */

	return inv_image_ic_compose_step(next->prop_list1,x);
    }

    /* Can't do anything special. */
    return inv_image(next,x);
}


conjunct_list inv_image_ic_OLD(uneval_prop next, conjunct_list x)
	/* Computes "inverse image" (now called backward image) of
	   x under next, but exploits top-level ORs in next to
	   expand the size of the returned image.  (Does each rule
	   separately. */
	/* This is now replaced by the new inv_image_ic, which computes
	   the backward image, but recursively descends the nextstate relation,
	   and doesn't recombine intermediate results, yielding a really
	   long implicit conjunction as the result. */
{
    if (next->bdd != NULL) {
	/* Already been evaluated before. */
	return inv_image(next,x);
    }

    if (next->type != OR) {
	/* Can't do anything special. */
	return inv_image(next,x);
    }

    /* Top level is OR, I can do each disjunct separately. */
    {
	prop_list temppl;
	conjunct_list result, temp1, temp2;

	temppl = next->prop_list1;
	result = new_conjunct_list(0);
	while (temppl!=NULL) {
	    temp1 = inv_image_ic(temppl->data,x); /* Image under this rule. */
	    temp2 = concat_conjunct_lists(temp1, result);
	    free_conjunct_list(temp1);
	    free_conjunct_list(result);

	    /* Munch the list to keep from having to do a huge one later. */
	    slow_munch_conjunct_list(temp2,1.0);

	    result = temp2;

	    temppl = temppl->next;
	}

	return result;
    }
}


/* Forward declaration. */
bdd inv_image_bdd_care_set(uneval_prop next, bdd x, conjunct_list c);


bdd inv_image_bdd_care_set_compose_step(prop_list next_list,
					bdd x,
					conjunct_list c)
	/* Helper function for inv_image_bdd_care_set(). */
	/* See my notes from May 4, 1994 for explanation. */
{
    conjunct_list care_image;
    bdd tail, result;

    /* Base Case */
    if (next_list==NULL) return simplify_bdd_by_conjunct_list(x,c);

    /* Compute forward image of care set. */
	/* Image of implicit conjunction is hard; therefore, approximate. */
    care_image = over_approx_forward_image_conjunct_list(next_list->data, c);

    /* Recursive BackImage on tail of list. */
    tail = inv_image_bdd_care_set_compose_step(next_list->next, x, care_image);
    free_conjunct_list(care_image);

    /* BackImage on head of list. */
    result = inv_image_bdd_care_set(next_list->data, tail, c);
    bdd_free(gbm,tail);

    return result;
}

bdd inv_image_bdd_care_set(uneval_prop next, bdd x, conjunct_list c)
	/* Computes BackImage (formerly "inverse image") of x under next,
	   using c as a care set on the result. */
{
    bdd temp1, temp2;
    prop_list tp;

    if (next->bdd != NULL) {
	/* Already evaluated; let's use it. */
	temp1 = inv_image_bdd(next, x);
	temp2 = simplify_bdd_by_conjunct_list(temp1, c);
	bdd_free(gbm,temp1);

	return temp2;
    }

    if (next->type == OR) {
	/* Unevaluated disjunction. */
	bdd temp3;
	conjunct_list care = copy_conjunct_list(c);

	printf("["); fflush(stdout);    /* DEBUG */

	temp1 = bdd_one(gbm);
	tp = next->prop_list1;
	while (tp!=NULL) {
	    /* Compute BackImage for each disjunct. */

	    printf("."); fflush(stdout);    /* DEBUG */
	    temp2 = inv_image_bdd_care_set(tp->data, x, care);

	    /* Result from each disjunct is care set for others. */
	    append_to_conjunct_list(temp2,care);

	    printf("o"); fflush(stdout);    /* DEBUG */
	    temp3 = bdd_and(gbm,temp1, temp2);
	    /*bdd_print_profile(gbm,temp3,bdd_naming_fn_full,NULL,80,stdout);*/
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);
	    temp1 = temp3;
	    tp = tp->next;
	}
	printf("]"); fflush(stdout);    /* DEBUG */

	free_conjunct_list(care);

	temp2 = simplify_bdd_by_conjunct_list(temp1,c);
	bdd_free(gbm,temp1);

	return temp2;
    }

    if (next->type == IF) {
	/* Ternary if-then-else.  Should check here that condition only
	   operates on current phase, else this might not work. */
	conjunct_list care;
	bdd thenpart, elsepart;

	evaluate_proposition(next->prop1,FALSE);

	/* then part */
	care = copy_conjunct_list(c);
	append_to_conjunct_list(next->prop1->bdd, care);
	thenpart = inv_image_bdd_care_set(next->prop2, x, care);
	free_conjunct_list(care);

	/* else part */
	care = copy_conjunct_list(c);
	temp1 = bdd_not(gbm,next->prop1->bdd);
	append_to_conjunct_list(temp1, care);
	bdd_free(gbm,temp1);
	elsepart = inv_image_bdd_care_set(next->prop3, x, care);
	free_conjunct_list(care);

	temp1 = bdd_ite(gbm, next->prop1->bdd, thenpart, elsepart);
	bdd_free(gbm,thenpart);
	bdd_free(gbm,elsepart);

	temp2 = simplify_bdd_by_conjunct_list(temp1,c);
	bdd_free(gbm,temp1);

	return temp2;
    }

    if (next->type == COMPOSE) {
	/* Sequential Composition. */

	/* This one is SLOW! */
	/*return inv_image_bdd_care_set_compose_step(next->prop_list1,x,c);*/
	/* Alternative:  FALL THROUGH and do normal BackImage. */
    }

    /* Don't know how to propagate don't-care info for this case. */
    temp1 = inv_image_bdd(next, x);
    temp2 = simplify_bdd_by_conjunct_list(temp1, c);
    bdd_free(gbm,temp1);

    return temp2;
}

conjunct_list inv_image_care_set(uneval_prop next,
				conjunct_list x,
				conjunct_list c)
	/* Computes BackImage (formerly "inverse image") of x under next,
	   using c as a care set on the result. */
{
    int i;
    bdd temp;
    conjunct_list new = new_conjunct_list(x->size);
    conjunct_list care = copy_conjunct_list(c);

    /* Inverse image distributes over conjunction. */
    for (i=0; i < x->size; i++) {
printf("DEBUG 1 %d\n",i);
	new->conjunct[i] = inv_image_bdd_care_set(next,x->conjunct[i],care);
	/* Result is additional care set on other conjuncts. */
	append_to_conjunct_list(new->conjunct[i],care);
    }

    free_conjunct_list(care);

    /* Additional simplification pass of results by original care set. */
    for (i=0; i < x->size; i++) {
	temp = simplify_bdd_by_conjunct_list(new->conjunct[i],c);
	bdd_free(gbm,new->conjunct[i]);
	new->conjunct[i] = temp;
    }

    return new;
}


bdd image_bdd(uneval_prop next, bdd x)
	/* Computes forward image of a bdd. */
{
    bdd new;
    bdd temp1;

    temp1 = compute_forward_image(x,next,FALSE);

    /* Convert phase back to current phase. */
    bdd_assoc(gbm,next_vars_assoc);
    new = bdd_rel_prod(gbm,temp1, current_next_eq->bdd);
    bdd_free(gbm,temp1);

    return new;
}

disjunct_list image(uneval_prop next, disjunct_list x)
	/* Computes the forward image of x through next.
	   Forward image distributes over disjunct_list. */
{
    int i;
    disjunct_list new = new_disjunct_list(x->size);

    /* Inverse image distributes over conjunction. */
    for (i=0; i < x->size; i++) {
	new->disjunct[i] = image_bdd(next,x->disjunct[i]);
    }

    return new;
}

disjunct_list image_id(uneval_prop next, disjunct_list x)
	/* Computes forward image of x, recursively descending the 
	   nextstate relation and not combining intermediate results to
	   yield a long implicit disjunction. */
{
    if (next->bdd != NULL) {
	/* Already been evaluated before. */
	return image(next,x);
    }

    if (next->type == OR) {
	/* OR -- I can do each disjunct separately. */
	prop_list temppl;
	disjunct_list result, temp1, temp2;

	temppl = next->prop_list1;
	result = new_disjunct_list(0);
	while (temppl!=NULL) {
	    temp1 = image_id(temppl->data,x); /* Image under this rule. */
	    temp2 = concat_disjunct_lists(temp1, result);
	    free_disjunct_list(temp1);
	    free_disjunct_list(result);

	    /* Munch the list to keep from having to do a huge one later. */
	    /* Skip this for now. */
	    /*slow_munch_disjunct_list(temp2);*/
	    fast_munch_disjunct_list(temp2);

	    result = temp2;

	    temppl = temppl->next;
	}

	return result;
    }

    if (next->type == IF) {
	/* Ternary if-then-else.  Should check here that condition only
	   operates on current phase, else this might not work. */
	disjunct_list condlist, thenpart, elsepart, result;
	bdd temp1;
	int i;

	evaluate_proposition(next->prop1,FALSE);

	/* then part */
	condlist = copy_disjunct_list(x);
	for (i=0; i < condlist->size; i++) {
	    temp1 = bdd_and(gbm, next->prop1->bdd, condlist->disjunct[i]);
	    bdd_free(gbm,condlist->disjunct[i]);
	    condlist->disjunct[i] = temp1;
	}
	thenpart = image_id(next->prop2, condlist);
	free_disjunct_list(condlist);

	/* else part */
	condlist = copy_disjunct_list(x);
	for (i=0; i < condlist->size; i++) {
	    /* Compute (not condition) and x */
	    temp1 = bdd_ite(gbm, next->prop1->bdd,
				 bdd_zero(gbm),
				 condlist->disjunct[i]);
	    bdd_free(gbm,condlist->disjunct[i]);
	    condlist->disjunct[i] = temp1;
	}
	elsepart = image_id(next->prop3, condlist);
	free_disjunct_list(condlist);

	result = concat_disjunct_lists(thenpart, elsepart);
	free_disjunct_list(thenpart);
	free_disjunct_list(elsepart);
	fast_munch_disjunct_list(result);

	return result;
    }

    if (next->type == COMPOSE) {
	/* Sequential Composition. */
	prop_list tp1;
	disjunct_list temp1, result;

	tp1 = next->prop_list1;
	result = copy_disjunct_list(x);
	while (tp1 != NULL) {
	    temp1 = image_id(tp1->data,result);	/* Image under one step. */
	    free_disjunct_list(result);
	    result = temp1;
	    tp1 = tp1->next;
	}

	return result;
    }

    /* Can't do anything special. */
    return image(next,x);
}
