/* Copyright 1993 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Type and Function Definitions for Basic Conjunct and Disjunst Lists */
/* By Alan J. Hu */
/* January 4, 1993 */

/* ************************ CHANGE LOG ************************** */
/*
7 September 1994 -- Created this file to split off code from everinv.c.
*/
/* ************************************************************** */

/* Private list of bdds type. */

typedef struct conjunct_list_node *conjunct_list;

struct conjunct_list_node   {
				int     size;
				bdd     *conjunct;
			    };

/* IMPORTANT -- Many routines rely on conjunct_list and disjunct_list
	being the same size and structure.  If you change this, be sure
	to change everything in everlist.c. AJH 12 September 1994 */

typedef struct disjunct_list_node *disjunct_list;

struct disjunct_list_node   {
				int	size;
				bdd	*disjunct;
			    };

/* Functions to work on conjunct_list */


conjunct_list new_conjunct_list(int size);
disjunct_list new_disjunct_list(int size);

void free_conjunct_list(conjunct_list x);
void free_disjunct_list(disjunct_list x);

conjunct_list copy_conjunct_list(conjunct_list x);
disjunct_list copy_disjunct_list(disjunct_list x);

conjunct_list concat_conjunct_lists(conjunct_list a, conjunct_list b);
disjunct_list concat_disjunct_lists(disjunct_list a, disjunct_list b);

conjunct_list or_conjunct_lists(conjunct_list a, conjunct_list b);
void append_to_conjunct_list(bdd x, conjunct_list a);
	/* DESTRUCTIVE LIST MANIPULATION! */

conjunct_list de_morgan_dl_to_cl(disjunct_list x);
disjunct_list de_morgan_cl_to_dl(conjunct_list x);
void de_morgan_in_place(conjunct_list x);

void fast_munch_conjunct_list(conjunct_list a);
	/* DESTRUCTIVE LIST MANIPULATION! */
void medium_munch_conjunct_list(conjunct_list a);
	/* DESTRUCTIVE LIST MANIPULATION! */
    /* Looks for duplicate conjuncts or complementary conjuncts. */
void phased_slow_munch_conjunct_list(conjunct_list a,
					float grow_threshold,
					int number_of_phases);
	/* DESTRUCTIVE LIST MANIPULATION! */
    /* Tries to shorten conjunct list a. */
    /* This is slightly greedier than slow_munch_conjunct_list.
       It goes through several phases.  At phase i, it uses a
       grow_threshold of grow_threshold*i/number_of_phases.
       During a phase, it combines any conjuncts that gives a growth
       ratio less than the current grow_threshold (rather than looking
       for the minimum growth ratio). */
void slow_munch_conjunct_list(conjunct_list a, float grow_threshold);
	/* DESTRUCTIVE LIST MANIPULATION! */
    /* Tries to shorten conjunct list a. */
    /* Implements the greedy heuristic from DAC 94 that always combines
	the pair which results in the least growth. */
    /* This implementation should use less space, and probably run faster
	than what I had used in evaluate_some_conjunctions(). */
void minmax_munch_conjunct_list(conjunct_list a);
	/* DESTRUCTIVE LIST MANIPULATION! */
	/* Combines conjuncts if the size of the result is no larger than
		the size of the larger of the two. */

void fast_munch_disjunct_list(disjunct_list a);
	/* DESTRUCTIVE LIST MANIPULATION! */
void medium_munch_disjunct_list(disjunct_list a);
	/* DESTRUCTIVE LIST MANIPULATION! */
void phased_slow_munch_disjunct_list(disjunct_list a,
					float grow_threshold,
					int number_of_phases);
	/* DESTRUCTIVE LIST MANIPULATION! */
void slow_munch_disjunct_list(disjunct_list a, float grow_threshold);
	/* DESTRUCTIVE LIST MANIPULATION! */
void minmax_munch_disjunct_list(disjunct_list a);
	/* DESTRUCTIVE LIST MANIPULATION! */
	/* Combines disjuncts if the size of the result is no larger than
		the size of the larger of the two. */

bdd simplify_bdd_by_conjunct_list(bdd f, conjunct_list c);

/* Various tests for convergence. */

int conjunct_list_component_equality(conjunct_list a, conjunct_list b);

int check_disjunct_list_tautology(conjunct_list a);
int conjunct_list_implies_bdd(conjunct_list a, bdd b);
int check_conjunct_list_implication_step(conjunct_list x, conjunct_list y);
    /* Checks whether x implies y. */
    /* 8 August 1994, AJH -- I realized that the way I had been decomposing
	this problem has two weaknesses:  (1) it might be easy to disprove
	implication for one conjunct, but not the others, so it'd be nice
	to detect the easier conjuncts faster and (2) the current version
	ends up redoing the Shannon decomposition of the conjunct list over
	and over again.  So, this is an attempt at doing it more directly. */
    /* MAY MODIFY ARGUMENTS.  Use check_conjunct_list_implication() as
	a wrapper to protect arguments. 21 August 1994 AJH */
int check_conjunct_list_implication(conjunct_list a, conjunct_list b);
    /* Checks whether a implies b. */
    /* Wrapper to reduce amount of memory alloc/dealloc. */
int check_disjunct_list_implication(disjunct_list a, disjunct_list b);

int check_conjunct_list_equality(conjunct_list a, conjunct_list b);
	/* This one is necessary and sufficient.  It works by checking
		implication in both directions. */

int check_disjunct_list_implies_conjunct_list(disjunct_list a, conjunct_list b);

/* This is for benchmarking data only.  Should be commented out. */
void print_conjunct_list_size(conjunct_list x, int d);
void print_disjunct_list_size(disjunct_list x, int d);


conjunct_list bdd_to_conjunct_list(bdd x);
bdd conjunct_list_to_bdd(conjunct_list x);

