/* Copyright 1991 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Main Routines Version 0.0 */
/* By Alan J. Hu */
/* August 6, 1991 */

/* ************************ CHANGE LOG ************************** */
/*
8 September 1991 AJH -- Changed bit_vector to be an unevaluate type.
15 September 1991 AJH -- Added unevaluated disjunctive nextstate relations.
8 November 1991 AJH -- Removed old interleaving code.
11 December 1991 AJH -- Added more aggressive unevaluated nextstate relations.
11 December 1991 AJH -- Completely redid predicate instantiation to
			a less-space efficient, but cleaner (and correct
			for nested calls, which the old one wasn't).  Now,
			there is no longer a predicate type.  Instead, when
			the predicate is instantiated, make a new copy of
			the bound proposition.
4 January 1992 AJH -- Fixed a bunch of memory leaks in the print trace routines.
31 January 1992 AJH -- Added constrain operator for constrained assignment.
17 October 1992 AJH -- Modified printtrace for functionally dependent variables
7 November 1992 AJH -- Converted to use David Long's BDD package.
15 August 1993 AJH -- Added catch for SIGINT to exit.
18 November 1993 AJH -- Added line to enable/disable dynamic reordering.
11 May 1994 AJH -- Converted compute_backward_image() to use bdd_ite() for
		   the ternary if-then-else branch.
11 September 1994 AJH -- Pulled the image computation stuff out into
		everimage.c.  Moved the printtrace stuff into everinv.c.
22 December 1996 AJH -- Added "repeat" operator to do rule repetition.
11 August 1997 AJH -- Added rotate operators.
14 August 1997 AJH -- Added code for rotate operators.
*/
/* ************************************************************** */

#include "ever.h"
#include "y.tab.h"	/* Defines constants for the tokens. */

#include "signal.h"	/* To catch SIGINT and exit. */
void exit();



/* YACC Interface Utilities */

void pline()
{
  fprintf(stderr, "Line %d: ", yylineno);
}

void yyerror(char *s)
{
  pline();
  fprintf(stderr, "%s (last token: %s)\n", s, yytext);
}

char *strpreserve(char *s)
{
  char *t = (char *) malloc(strlen(s)+1);
  return(strcpy(t,s));
}




/* List Manipulation Functions */
/*	WARNING:  These are destructive list manipulation functions! */

string_list	append_to_string_list(char *datum, string_list old_list)
{
	string_list	new_node, x;

	new_node = (string_list)malloc(sizeof(*new_node));
	new_node->data = datum;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}


record_list	append_to_record_list(char *id,
					type_desc field_type,
					record_list old_list)
{
	record_list	new_node, x;

	new_node = (record_list)malloc(sizeof(*new_node));
	new_node->field_name = id;
	new_node->field_type = field_type;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}

int	bdd_count_of_record_list(record_list x)
{
	if (x==NULL) return 0;
	else return x->field_type->bdd_count +
			bdd_count_of_record_list(x->next);
}

int	length_of_record_list(record_list x)
{
	if (x==NULL) return 0;
	else return 1 + length_of_record_list(x->next);
}

int	max_size_in_record_list(record_list x)
{
	int maxsize = -1;
	record_list t = x;

	while (t!=NULL) {
		if (t->field_type->bdd_count>maxsize)
			maxsize = t->field_type->bdd_count;
		t = t->next;
	}

	return maxsize;
}


prop_list	append_to_prop_list(uneval_prop new_item, prop_list old_list)
{
	prop_list	new_node, x;

	new_node = (prop_list)malloc(sizeof(*new_node));
	new_node->data = new_item;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}


vectorize_list	append_prop_to_vectorize_list(uneval_prop new_item,
						vectorize_list old_list)
{
	vectorize_list	new_node, x;

	new_node = (vectorize_list)malloc(sizeof(*new_node));
	new_node->type = DEFPROP;	/* For a proposition type node */
	new_node->prop_data = new_item;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}

vectorize_list	append_bit_vector_to_vectorize_list(bit_vector new_item,
						vectorize_list old_list)
{
	vectorize_list	new_node, x;

	new_node = (vectorize_list)malloc(sizeof(*new_node));
	new_node->type = DEFVEC;	/* For a vector type node */
	new_node->vector_data = new_item;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}


bit_vector_list	append_to_bit_vector_list(bit_vector new_item,
					bit_vector_list old_list)
{
	bit_vector_list	new_node, x;

	new_node = (bit_vector_list)malloc(sizeof(*new_node));
	new_node->data = new_item;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}


var_accessor	append_to_var_accessor(int type, bit_vector array_index,
					char *field_select,
					var_accessor old_list)
{
	var_accessor new_node, x;

	new_node = (var_accessor)malloc(sizeof(*new_node));
	new_node->type = type;
	new_node->array_index = array_index;
	new_node->field_select = field_select;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}


variable_list	append_to_variable_list(variable_desig new_item,
					variable_list old_list)
{
	variable_list	new_node, x;

	new_node = (variable_list)malloc(sizeof(*new_node));
	new_node->data = new_item;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}


id_list append_to_ID_list(char *new_item, id_list old_list)
{
	id_list	new_node, x;

	new_node = (id_list)malloc(sizeof(*new_node));
	new_node->data = new_item;
	new_node->next = NULL;

	if (old_list == NULL) return new_node;

	/* Append the new node. */
	x = old_list;
	while (x->next != NULL) x = x->next;
	x->next = new_node;

	return old_list;
}



/* Building Type Descriptors */

type_desc	build_bit_field_type(int size, string_list labels)
{
	type_desc	new_item;

	new_item = (type_desc)malloc(sizeof(*new_item));

	new_item->type = BITS;
	new_item->bits = size;
	new_item->bdd_count = 3*size;	/* One for each phase */
	new_item->output_labels = labels;

	return new_item;
}

type_desc	build_array_type(int interleave,
				    int lb,
				    int ub,
				    type_desc element_type)
{
	type_desc	new_item;

	new_item = (type_desc)malloc(sizeof(*new_item));

	new_item->type = ARRAY;
	new_item->interleaved = interleave;
	new_item->array_lb = lb;
	new_item->array_ub = ub;
	new_item->bdd_count = (ub-lb+1)*(element_type->bdd_count);
	new_item->subtype = element_type;

	return new_item;
}

type_desc	build_record_type(int interleave,
				    record_list fields)
{
	type_desc	new_item;

	new_item = (type_desc)malloc(sizeof(*new_item));

	new_item->type = RECORD;
	new_item->interleaved = interleave;
	new_item->bdd_count = bdd_count_of_record_list(fields);
	new_item->field_list = fields;

	return new_item;
}



/* Building uneval_prop Expression Trees */

/* Forward declarations for helper functions for instantiate_predicate. */
uneval_prop plug_in_actuals(uneval_prop bound, bit_vector_list args);
variable_desig plug_in_actuals_variable_desig(variable_desig x,
						bit_vector_list args);
vectorize_list plug_in_actuals_vectorize_list(vectorize_list x,
						bit_vector_list args);

uneval_prop instantiate_predicate(predicate pred, bit_vector_list args)
{
	uneval_prop	new_node;

	if (pred->is_lambda) {
	    yyerror("Warning:  Lambda predicates are unimplemented.");
	}

	/* Copies bound prop with args plugged in. */
	new_node = plug_in_actuals(pred->bound_prop, args);
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop and_propositions(prop_list props)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = AND;
	new_node->bdd = NULL;
	new_node->prop_list1 = props;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop or_propositions(prop_list props)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = OR;
	new_node->bdd = NULL;
	new_node->prop_list1 = props;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop not_proposition(uneval_prop p)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = NOT;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop xor_propositions(uneval_prop a, uneval_prop b)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = XOR;
	new_node->bdd = NULL;
	new_node->prop1 = a;
	new_node->prop2 = b;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop equiv_propositions(uneval_prop a, uneval_prop b)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = EQUIV;
	new_node->bdd = NULL;
	new_node->prop1 = a;
	new_node->prop2 = b;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop implies_propositions(uneval_prop a, uneval_prop b)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = IMPLIES;
	new_node->bdd = NULL;
	new_node->prop1 = a;
	new_node->prop2 = b;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop forall_proposition(variable_list quant_vars, uneval_prop p)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = FORALL;
	new_node->bdd = NULL;
	new_node->quant_vars = quant_vars;
	new_node->prop1 = p;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop exists_proposition(variable_list quant_vars, uneval_prop p)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = EXISTS;
	new_node->bdd = NULL;
	new_node->quant_vars = quant_vars;
	new_node->prop1 = p;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop hide_previous_uneval(uneval_prop p)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = HIDEPREV;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop hide_current_uneval(uneval_prop p)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = HIDECUR;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop hide_next_uneval(uneval_prop p)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = HIDENEXT;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop if_then_else_uneval(uneval_prop ifp,
				uneval_prop thenp,
				uneval_prop elsep)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = IF;
	new_node->bdd = NULL;
	new_node->prop1 = ifp;
	new_node->prop2 = thenp;
	new_node->prop3 = elsep;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop reachable_uneval(uneval_prop start, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = REACHABLE;
	new_node->bdd = NULL;
	new_node->prop1 = start;
	new_node->prop2 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop reachgoal_uneval(uneval_prop goal,
				uneval_prop start,
				uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = REACHGOAL;
	new_node->bdd = NULL;
	new_node->prop1 = goal;
	new_node->prop2 = start;
	new_node->prop3 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop AG_uneval(uneval_prop p, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = AG;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop EG_uneval(uneval_prop p, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = EG;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop AF_uneval(uneval_prop p, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = AF;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop EF_uneval(uneval_prop p, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = EF;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop AX_uneval(uneval_prop p, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = AX;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop EX_uneval(uneval_prop p, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = EX;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop AU_uneval(uneval_prop p, uneval_prop untilq, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = AU;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = untilq;
	new_node->prop3 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop EU_uneval(uneval_prop p, uneval_prop untilq, uneval_prop next)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = EU;
	new_node->bdd = NULL;
	new_node->prop1 = p;
	new_node->prop2 = untilq;
	new_node->prop3 = next;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop simrel_uneval(uneval_prop next1, uneval_prop next2)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = SIMREL;
	new_node->bdd = NULL;
	new_node->prop1 = next1;
	new_node->prop2 = next2;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop becomes_uneval(variable_desig lhs, bit_vector rhs)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = BECOMES;
	new_node->bdd = NULL;
	new_node->var1 = lhs;
	new_node->vector1 = rhs;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop constrain_uneval(variable_desig lhs, uneval_prop rhs)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = CONSTRAIN;
	new_node->bdd = NULL;
	new_node->var1 = lhs;
	new_node->prop1 = rhs;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop compose_uneval(prop_list nextstates)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = COMPOSE;
	new_node->bdd = NULL;
	new_node->prop_list1 = nextstates;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop repeat_uneval(int count, uneval_prop nextstate)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = REPEAT;
	new_node->bdd = NULL;
	new_node->prop1 = nextstate;
	new_node->count = count;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop gt_uneval(bit_vector lhs, bit_vector rhs)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = GT;
	new_node->bdd = NULL;
	new_node->vector1 = lhs;
	new_node->vector2 = rhs;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop ge_uneval(bit_vector lhs, bit_vector rhs)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = GE;
	new_node->bdd = NULL;
	new_node->vector1 = lhs;
	new_node->vector2 = rhs;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop eq_uneval(bit_vector lhs, bit_vector rhs)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = EQ;
	new_node->bdd = NULL;
	new_node->vector1 = lhs;
	new_node->vector2 = rhs;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop le_uneval(bit_vector lhs, bit_vector rhs)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = LE;
	new_node->bdd = NULL;
	new_node->vector1 = lhs;
	new_node->vector2 = rhs;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop lt_uneval(bit_vector lhs, bit_vector rhs)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = LT;
	new_node->bdd = NULL;
	new_node->vector1 = lhs;
	new_node->vector2 = rhs;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop and_compress_uneval(bit_vector x)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = ANDCOMPRESS;
	new_node->bdd = NULL;
	new_node->vector1 = x;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}

uneval_prop or_compress_uneval(bit_vector x)
{
	uneval_prop	new_node;

	new_node = (uneval_prop)malloc(sizeof(*new_node));
	new_node->type = ORCOMPRESS;
	new_node->bdd = NULL;
	new_node->vector1 = x;
	new_node->lineno = yylineno;	/* Line number of end of definition. */
	return new_node;
}



/* Building Bit Vectors */

bit_vector make_integer_bit_vector(int width, int i)
	/* Puts MSB at head of list! */
	/* If width is -1, makes the bit vector as narrow as possible. */
{
	int size = 0;
	bit_vector new_vector;
	int temp;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = EVAL;	/* Already been evaluated. */
	new_vector->is_constant = TRUE;
	new_vector->constant_value = i;

	if (width < 0) {
	    /* Determine the size of i. */
	    for (temp=i, size=0; temp>0; temp >>= 1) size++;
	} else size = width;

	/* Fill in the bits. */
	new_vector->bit_list = (bdd *)calloc(size, sizeof(bdd));
	for (temp=size-1; temp>=0; temp--, i >>= 1) {
	    if (i & 1) {
		new_vector->bit_list[temp] = bdd_one(gbm);
	    } else {
		new_vector->bit_list[temp] = bdd_zero(gbm);
	    }
	}

	new_vector->size = size;
	new_vector->formal_order = -1;	/* Important flag! */
	return new_vector;
}

bit_vector make_variable_bit_vector(variable_desig v)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = DEFVAR;	/* Code for variable access */
	new_vector->var1 = v;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector vectorize(vectorize_list list)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = VECTOR;	/* Code for vector construction. */
	new_vector->vectorize_list1 = list;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector add_bit_vectors(bit_vector_list operands)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = ADD;
	new_vector->vec_list1 = operands;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector sub_bit_vectors(bit_vector a, bit_vector b)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = SUB;
	new_vector->vector1 = a;
	new_vector->vector2 = b;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector left_shift_bit_vectors(bit_vector operand, bit_vector amount)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = LEFTSHIFT;
	new_vector->vector1 = operand;
	new_vector->vector2 = amount;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector right_shift_bit_vectors(bit_vector operand, bit_vector amount)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = RIGHTSHIFT;
	new_vector->vector1 = operand;
	new_vector->vector2 = amount;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector left_rotate_bit_vectors(bit_vector operand, bit_vector amount)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = LEFTROTATE;
	new_vector->vector1 = operand;
	new_vector->vector2 = amount;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector right_rotate_bit_vectors(bit_vector operand, bit_vector amount)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = RIGHTROTATE;
	new_vector->vector1 = operand;
	new_vector->vector2 = amount;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector bitwise_equivalence(bit_vector a, bit_vector b)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = BEQUIV;
	new_vector->vector1 = a;
	new_vector->vector2 = b;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector bitwise_implies(bit_vector a, bit_vector b)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = BIMPLIES;
	new_vector->vector1 = a;
	new_vector->vector2 = b;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector bitwise_xor(bit_vector a, bit_vector b)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = BXOR;
	new_vector->vector1 = a;
	new_vector->vector2 = b;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector bitwise_and(bit_vector_list operands)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = BAND;
	new_vector->vec_list1 = operands;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector bitwise_or(bit_vector_list operands)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = BOR;
	new_vector->vec_list1 = operands;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector bitwise_not(bit_vector a)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = BNOT;
	new_vector->vector1 = a;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}

bit_vector bit_vector_if_then_else(uneval_prop cond,
					bit_vector thenpart,
					bit_vector elsepart)
{
	bit_vector new_vector;

	new_vector = (bit_vector)malloc(sizeof(*new_vector));
	new_vector->type = BIF;
	new_vector->prop1 = cond;
	new_vector->vector1 = thenpart;
	new_vector->vector2 = elsepart;
	new_vector->is_constant = FALSE;
	new_vector->formal_order = -1;
	new_vector->bit_list = NULL;
	return new_vector;
}



/* Variable Designator Routines */

variable_desig	add_phase_to_var_desig(int phase, variable_desig x)
{
	if ((x->fdv) && (phase!=CUR)) {
	    fprintf(stderr,"Warning:  Non-current phase of dependent variable. (Line %d)\n", yylineno);
	    fflush(stderr);
	}

	x->phase = phase;
	return x;
}

variable_desig	array_index_var_desig(variable_desig var, bit_vector index)
{
	var->modifiers = append_to_var_accessor(ARRAY,index,
						NULL,var->modifiers);
	return var;
}

variable_desig	record_access_var_desig(variable_desig var, char *fieldname)
{
	var->modifiers = append_to_var_accessor(RECORD,NULL,
						fieldname,var->modifiers);
	return var;
}



/* Predicate Building Routines */

predicate	build_macro_predicate(id_list formals, uneval_prop x)
{
	predicate	new_item;

	new_item = (predicate)malloc(sizeof(*new_item));

	new_item->is_lambda = FALSE;
	new_item->formal_list = formals;
	new_item->bound_prop = x;

	return new_item;
}

predicate	build_lambda_predicate(id_list formals, uneval_prop x)
{
	predicate	new_item;

	new_item = (predicate)malloc(sizeof(*new_item));

	new_item->is_lambda = TRUE;
	new_item->formal_list = formals;
	new_item->bound_prop = x;

	return new_item;
}



/* Predicate Instantiation Routines */

bit_vector plug_in_actuals_bit_vector(bit_vector x, bit_vector_list args)
{
    prop_list p;
    bit_vector_list b, bnew;
    int i;
    bit_vector new_vector;

    if (x->formal_order>=0) {
        /* It's an argument. */
        /* Find the appropriate actual parameter. */
        b = args;
        for (i=0; i < x->formal_order; i++) b = b->next;

        /* Plug it in. */
        return b->data;
    }

    if (x->is_constant) return x;         /* End of the line. */

    new_vector = (bit_vector)malloc(sizeof(*new_vector));
    new_vector->type = x->type;
    new_vector->is_constant = x->is_constant;
    new_vector->constant_value = x->constant_value;
    new_vector->size = x->size;
    new_vector->formal_order = x->formal_order;
    new_vector->bit_list = NULL;

    /* It's not an argument, recursively descend through the tree. */
    switch (x->type) {
    case DEFVAR:        /* Variable access */
        new_vector->var1 = plug_in_actuals_variable_desig(x->var1, args);
        break;
    case VECTOR:
        new_vector->vectorize_list1 =
		plug_in_actuals_vectorize_list(x->vectorize_list1, args);
        break;
    case ADD:
    case BAND:
    case BOR:
	bnew = NULL;
        for (b = x->vec_list1; b!=NULL; b = b->next) {
            bnew = append_to_bit_vector_list(
			plug_in_actuals_bit_vector(b->data,args),
			bnew);
        }
	new_vector->vec_list1 = bnew;
        break;
    case SUB:
    case LEFTSHIFT:
    case RIGHTSHIFT:
    case LEFTROTATE:
    case RIGHTROTATE:
    case BEQUIV:
    case BIMPLIES:
    case BXOR:
        new_vector->vector1 = plug_in_actuals_bit_vector(x->vector1, args);
        new_vector->vector2 = plug_in_actuals_bit_vector(x->vector2, args);
        break;
    case BNOT:
        new_vector->vector1 = plug_in_actuals_bit_vector(x->vector1, args);
        break;
    case BIF:
        new_vector->prop1 = plug_in_actuals(x->prop1, args);
        new_vector->vector1 = plug_in_actuals_bit_vector(x->vector1, args);
        new_vector->vector2 = plug_in_actuals_bit_vector(x->vector2, args);
        break;
    default:
        yyerror("Evaluator Error:  Unimplemented operator while instantiating predicate.");
        fprintf(stderr,"-----Operator type = %d.\n",x->type);
    }

    return new_vector;
}


vectorize_list plug_in_actuals_vectorize_list(vectorize_list x,
						bit_vector_list args)
{
    vectorize_list new_list;

    new_list = NULL;
    for (; x!=NULL; x = x->next) {
        if (x->type==DEFPROP)
	    new_list = append_prop_to_vectorize_list(
				plug_in_actuals(x->prop_data,args),
				new_list);
        if (x->type==DEFVEC)
	    new_list = append_bit_vector_to_vectorize_list(
				plug_in_actuals_bit_vector(x->vector_data,args),
				new_list);
    }

    return new_list;
}


variable_desig plug_in_actuals_variable_desig(variable_desig x,
						bit_vector_list args)
{
    var_accessor v, vnew;
    variable_desig  new_desig;
    bit_vector new_array_index;

    new_desig = (variable_desig)malloc(sizeof(*new_desig));
    new_desig->base_type = x->base_type;
    new_desig->bitcount = x->bitcount;
    new_desig->fdv = x->fdv;
    new_desig->func_def = x->func_def;
    new_desig->bfv_offset = x->bfv_offset;
    new_desig->bits = x->bits;
    new_desig->phase = x->phase;

    vnew = NULL;
    for (v = x->modifiers; v!=NULL; v = v->next) {
        if (v->type==ARRAY) {
	    new_array_index = plug_in_actuals_bit_vector(v->array_index,args);
	    vnew = append_to_var_accessor(ARRAY, new_array_index, NULL, vnew);
	} else {
	    vnew = append_to_var_accessor(v->type,
					  v->array_index,
					  v->field_select,
					  vnew);
	}
    }
    new_desig->modifiers = vnew;

    return new_desig;
}


uneval_prop plug_in_actuals(uneval_prop bound, bit_vector_list args)
{
    prop_list tempp1, tempp2;
    bit_vector_list tempv1;
    uneval_prop	new_node;

    if (bound->type==EVAL) return bound;     /* Predefined proposition. */

    new_node = (uneval_prop)malloc(sizeof(*new_node));
    new_node->type = bound->type;
    new_node->bdd = NULL;
    new_node->lineno = bound->lineno;

    switch (bound->type) {
    case AND:                   /* n-ary */
    case OR:
    case COMPOSE:
        tempp1 = bound->prop_list1;
	tempp2 = NULL;
        while (tempp1!=NULL) {
            tempp2 = append_to_prop_list(plug_in_actuals(tempp1->data, args),
					 tempp2);
            tempp1 = tempp1->next;
        }
	new_node->prop_list1 = tempp2;
        break;
    case REPEAT:
	new_node->count = bound->count;
	new_node->prop1 = plug_in_actuals(bound->prop1,args);
	break;
    case NOT:                   /* unary */
    case FORALL:
    case EXISTS:
    case HIDEPREV:
    case HIDECUR:
    case HIDENEXT:
        new_node->prop1 = plug_in_actuals(bound->prop1,args);
        break;
    case XOR:                   /* binary */
    case EQUIV:
    case IMPLIES:
    case REACHABLE:
    case AG:
    case EG:
    case AF:
    case EF:
    case AX:
    case EX:
    case SIMREL:
        new_node->prop1 = plug_in_actuals(bound->prop1,args);
        new_node->prop2 = plug_in_actuals(bound->prop2,args);
        break;
    case IF:                    /* ternary */
    case REACHGOAL:
    case AU:
    case EU:
        new_node->prop1 = plug_in_actuals(bound->prop1,args);
        new_node->prop2 = plug_in_actuals(bound->prop2,args);
        new_node->prop3 = plug_in_actuals(bound->prop3,args);
        break;
    case GT:
    case GE:
    case EQ:
    case LE:
    case LT:
        new_node->vector1 = plug_in_actuals_bit_vector(bound->vector1,args);
        new_node->vector2 = plug_in_actuals_bit_vector(bound->vector2,args);
        break;
    case BECOMES:
        new_node->var1 = plug_in_actuals_variable_desig(bound->var1,args);
        new_node->vector1 = plug_in_actuals_bit_vector(bound->vector1,args);
        break;
    case CONSTRAIN:
	new_node->var1 = plug_in_actuals_variable_desig(bound->var1,args);
	new_node->prop1 = plug_in_actuals(bound->prop1,args);
	break;
    case ANDCOMPRESS:
    case ORCOMPRESS:
        new_node->vector1 = plug_in_actuals_bit_vector(bound->vector1,args);
        break;
    default:
        yyerror("Evaluator Error:  Unimplemented operator while instantiating predicate.");
        fprintf(stderr,"-----Operator type = %d.\n",bound->type);
    }

    return new_node;
}





/* Helper Functions */

/* Returns true iff all the variable designators in variable_list
   are OK to quantify on. */
int quantifiable_variables(variable_list the_vars)
{
	/* FIX THIS */
}




/* Global Data Structures */
bdd_manager gbm;

			/* BE CAREFUL NOT TO FREE THESE GLOBALS */
uneval_prop true_proposition, false_proposition;
uneval_prop previous_current_eq, current_next_eq, previous_next_eq;

bdd	*prev_vars, *cur_vars, *next_vars; /* Arrays for bdd_qnt. */
int	prev_vars_assoc, cur_vars_assoc, next_vars_assoc;
		/* For quantification and bdd_rel_prod */

int	global_bdd_count;
int	global_fdv_count;


/* ************************** Internal Routines ********************** */


extern FILE *yyin;			/* input file */


void init_globals()
{
  gbm = bdd_init();
  bdd_cache_ratio(gbm,256);

  /* Turn on/off dynamic variable reordering.  AJH 18 November 1993 */
  /*bdd_dynamic_reordering(gbm,bdd_reorder_sift);*/
  /*bdd_dynamic_reordering(gbm,bdd_reorder_stable_window3);*/

  true_proposition = (uneval_prop)malloc(sizeof(*true_proposition));
  true_proposition->type = EVAL;	/* Code for already evaluated. */
  true_proposition->bdd = bdd_one(gbm);

  false_proposition = (uneval_prop)malloc(sizeof(*false_proposition));
  false_proposition->type = EVAL;	/* Code for already evaluated. */
  false_proposition->bdd = bdd_zero(gbm);

  previous_current_eq = (uneval_prop)malloc(sizeof(*previous_current_eq));
  previous_current_eq->type = EVAL;	/* Code for already evaluated. */
  previous_current_eq->bdd = bdd_one(gbm);

  current_next_eq = (uneval_prop)malloc(sizeof(*current_next_eq));
  current_next_eq->type = EVAL;	/* Code for already evaluated. */
  current_next_eq->bdd = bdd_one(gbm);

  previous_next_eq = (uneval_prop)malloc(sizeof(*previous_next_eq));
  previous_next_eq->type = EVAL;	/* Code for already evaluated. */
  previous_next_eq->bdd = bdd_one(gbm);

  prev_vars = (bdd *)calloc(1,sizeof(bdd));
  prev_vars[0] = NULL;
  prev_vars_assoc = bdd_new_assoc(gbm, prev_vars, 0);

  cur_vars = (bdd *)calloc(1,sizeof(bdd));
  cur_vars[0] = NULL;
  cur_vars_assoc = bdd_new_assoc(gbm, cur_vars, 0);

  next_vars = (bdd *)calloc(1,sizeof(bdd));
  next_vars[0] = NULL;
  next_vars_assoc = bdd_new_assoc(gbm, next_vars, 0);

  global_bdd_count = 0;
  global_fdv_count = 0;
}


void my_exit()
{
	bdd_stats(gbm,stderr);
	fflush(stderr);
	fflush(stdout);
	exit();
}

/* ***** Main Program ***** */
void main(int argc, char *argv[])
{
  printf("Ever Version 0.1 by Alan J. Hu\n");
  printf("Use of this program implies acceptance of license terms.\n");
  printf("See the README file that came with this distribution for details.\n");
  printf("\n\n");

  /*signal(SIGINT,exit);*/	/* Catch interrupt and exit.  Lets profiling happen. */
  signal(SIGINT,my_exit);	/* Catch interrupt and exit.  Flush outputs. */
  init_globals();

  if (argc >= 2) {
    yyin = fopen( argv[1], "r" );	/* Doesn't deal with nested loads. */
    if (yyin==NULL) { 
      fprintf(stderr,"*** Error opening `%s' ***\n",argv[1]);
      fflush(stderr);
      yyin=stdin;
    } else printf("Taking input from %s...\n",argv[1]);
  }
  lexinit();
  yyparse();
  bdd_stats(gbm,stderr);
}
