/* Copyright 1992 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Multiset of Boolean Functional Vector (MOB) Code Version 0.0 */
/* By Alan J. Hu */
/* October 19, 1992 */

/* ************************ CHANGE LOG **************************** */
/*
18 November 1992 AJH	-- Converted to David Long's BDD Package.
11 March 1993 AJH	-- Caught a silly (non-BDD) memory leak.
1 October 1995 AJH	-- Caught a silly memory leak when building the MOB.
23 December 1996 AJH	-- Added repeat command.
*/
/* **************************************************************** */

#include "ever.h"
#include "y.tab.h"	/* Defines constants for tokens. */


typedef struct bdd_vector_node *bdd_vector;	/* Private bdd_vector type. */

struct bdd_vector_node	{
				int	size;
				bdd	*bits;
			};

typedef bdd *bool_func_vec;	/* Length is globally defined. */

typedef struct mob_node *mob;

struct mob_node {
			int		count;
			bool_func_vec	*bfv;
		};


int bfv_length;

bool_func_vec initial_bfv;	/* Starting point BFV (current state). */

mob global_mob=NULL;




/* BDD_VECTOR Routines */



void free_bdd_vector(bdd_vector x)
{
	int i;

	for (i=0; i < x->size; i++) {
		bdd_free(gbm,x->bits[i]);
	}
	free(x->bits);	/* AJH 3/11/93 */
	free(x);
}



bdd eval_prop_with_bfv(bool_func_vec old, uneval_prop p); /* forward */
bdd_vector eval_bit_vector_with_bfv(bool_func_vec old, bit_vector b);


bdd bdd_eq_int_bit_vector_with_bfv(bool_func_vec old,
					int x,
					bit_vector y)
{
    int mask;
    int which_bit;
    bdd result, temp1, temp2;
    bdd_vector y_eval;

    y_eval = eval_bit_vector_with_bfv(old,y);

    result = bdd_one(gbm);
    for (which_bit=y_eval->size-1; which_bit>=0; which_bit--) {
	if (x & 1) {
	    temp1 = bdd_identity(gbm,y_eval->bits[which_bit]);
	} else {
	    temp1 = bdd_not(gbm,y_eval->bits[which_bit]);
	}
	temp2 = bdd_and(gbm,temp1,result);
	bdd_free(gbm,temp1);
	bdd_free(gbm,result);
	result = temp2;

	x /= 2;
    }

    free_bdd_vector(y_eval);

    if (x>0) {
	bdd_free(gbm,result);
	return bdd_zero(gbm);	/* x is too big to be equal */
    } else return result;
}

bdd bit_vector_eq_with_bfv(bool_func_vec old,
				   bit_vector a,
				   bit_vector b)
{
    int i, maxsize;
    bdd_vector lhs, rhs;
    bdd result;
    bdd temp1, temp2;

    lhs = eval_bit_vector_with_bfv(old,a);
    rhs = eval_bit_vector_with_bfv(old,b);
    maxsize = (lhs->size > rhs->size)? lhs->size : rhs->size;

    result = bdd_one(gbm);

    for (i=0; i<maxsize; i++) {
	if (i > lhs->size-1) {
	    temp1 = bdd_not(gbm,rhs->bits[rhs->size-1-i]);
	    temp2 = bdd_and(gbm,result,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,result);
	    result = temp2;
	} else if (i > rhs->size-1) {
	    temp1 = bdd_not(gbm,lhs->bits[lhs->size-1-i]);
	    temp2 = bdd_and(gbm,result,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,result);
	    result = temp2;
	} else {
	    temp1 = bdd_xnor(gbm,rhs->bits[rhs->size-1-i],lhs->bits[lhs->size-1-i]);
	    temp2 = bdd_and(gbm,result,temp1);
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,result);
	    result = temp2;
	}
    }

    free_bdd_vector(lhs);
    free_bdd_vector(rhs);

    return result;
}



/* Helper function for eval_bit_vector_var_access_with_bfv() in evereval.c */
extern int var_desig_size(type_desc type_structure, var_accessor modifiers);

/* Helper function for eval_bit_vector_var_access_with_bfv(). */
bdd build_rhs_var_prop_with_bfv(bool_func_vec old,
					type_desc remaining_type,
					int pre_offset,
					int whichbit,
					var_accessor modchain)
{
/*printf("DEBUG--Enter build_rhs_var_prop_with_bfv, pre_offset=%d, whichbit=%d.\n",pre_offset, whichbit);*/
    if (modchain==NULL) {
	return bdd_identity(gbm,old[pre_offset+whichbit]);
    }

    if (modchain->type==RECORD) {
	record_list temp = remaining_type->field_list;
	int offset = 0;

	while (temp!=NULL) {
	    if (strcmp(temp->field_name,modchain->field_select)==0) {
		return build_rhs_var_prop_with_bfv(old,
						   temp->field_type,
						   pre_offset+offset,
						   whichbit,
						   modchain->next);
	    } else {
		offset += temp->field_type->bdd_count/3;
		temp = temp->next;
	    }
	}
	yyerror("Undeclared field.");
    }

    if ((modchain->type==ARRAY) && (modchain->array_index->is_constant)) {
	/* Constant array index. */
	int i = modchain->array_index->constant_value;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

	if ((i<lb) || (i>ub))
		yyerror("Warning --- Constant array index out of range.");

	return build_rhs_var_prop_with_bfv(old,
					   element_type,
					   pre_offset+
					    (i-lb)*element_type->bdd_count/3,
					   whichbit,
					   modchain->next);
    }

    if ((modchain->type==ARRAY) && (!modchain->array_index->is_constant)) {
	/* Non-constant array index. */
	int i = modchain->array_index->constant_value;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;
	bdd result, temp1, temp2, temp3;

	result = bdd_one(gbm);
	for (i=lb; i<=ub; i++) {
	    temp1 = bdd_eq_int_bit_vector_with_bfv(old,i,modchain->array_index);
	    temp2 = build_rhs_var_prop_with_bfv(old,
					   element_type,
					   pre_offset+
					    (i-lb)*element_type->bdd_count/3,
					   whichbit,
					   modchain->next);
	    temp3 = bdd_ite(gbm, temp1, temp2, bdd_one(gbm));
	    bdd_free(gbm,temp1);
	    bdd_free(gbm,temp2);
	    temp1 = bdd_and(gbm, result, temp3);
	    bdd_free(gbm,result);
	    bdd_free(gbm,temp3);
	    result = temp1;
	}
	return result;
    }

    yyerror("Internal error in build_rhs_var_prop_with_bfv: illegal modifier.");
    return bdd_zero(gbm);
}

bdd_vector eval_bit_vector_var_access_with_bfv(bool_func_vec old,
						bit_vector b)
{
    int wordsize;
    int i;
    bdd_vector new_vector;
    variable_desig v = b->var1;

/*printf("DEBUG--Enter eval_bit_vector_var_access.\n");*/
    /* Type checking and computes width of assignment. */
    wordsize = var_desig_size(v->base_type, v->modifiers);
    if (wordsize > 0) {		/* Type checking worked out. */
	new_vector = (bdd_vector)malloc(sizeof(*new_vector));
	new_vector->size = wordsize;
	new_vector->bits = (bdd *)calloc(wordsize,sizeof(bdd));

	/* Cheap hack, make sure rhs only uses current phase for
		BFV's. */
	if (v->phase!=CUR) {
	    fprintf(stderr,"Warning: Non-current phase in rhs expression is illegal for BFVs.\n");
	    fprintf(stderr,"\tAssuming current phase.\n");
	    fflush(stderr);
	}

	/* Puts MSB at head of list. */
	for (i=0; i<wordsize; i++) {
	    new_vector->bits[wordsize-1-i] =
		build_rhs_var_prop_with_bfv(old,
					v->base_type,
					v->bfv_offset,	/* offset to start */
					i,		/* which bit */
					v->modifiers);	/* modchain */
	}

	return new_vector;
    } else
	/* var_desig_size prints error message. */
	return NULL;	/* FIX THIS, should do something more robust. */
}

bdd_vector eval_bit_vector_vectorize_with_bfv(bool_func_vec old,
						vectorize_list v)
{
    bdd_vector final_result;

/*printf("DEBUG--Enter eval_bit_vector_vectorize.\n");*/
    if (v==NULL) {
	final_result = (bdd_vector)malloc(sizeof(*final_result));
	final_result->size = 0;
	final_result->bits = (bdd *)malloc(0);
	return final_result;
    }

    /* This recursion order puts things in backwards (first
	entry in vectorize list comes last) since becomes flips
	things around again. */
    final_result = eval_bit_vector_vectorize_with_bfv(old,v->next);
    if (v->type==DEFPROP) {
	/* It's an uneval_prop. */
	final_result->size++;
	final_result->bits = (bdd *)realloc(final_result->bits,
					final_result->size*sizeof(bdd));
	final_result->bits[final_result->size-1] =
		eval_prop_with_bfv(old,v->prop_data);
    } else if (v->type==DEFVEC) {
	/* It's bit_vector. */
	bdd_vector temp;
	int i;

	temp = eval_bit_vector_with_bfv(old,v->vector_data);
	final_result->size += temp->size;
	final_result->bits = (bdd *)realloc(final_result->bits,
					final_result->size*sizeof(bdd));
	for (i=0; i < temp->size; i++) {
	    final_result->bits[final_result->size - temp->size + i] =
		bdd_identity(gbm,temp->bits[i]);
	}

	free_bdd_vector(temp);
    } else {
	fprintf(stderr,"Error: Illegal node type in vectorize_with_bfv.\n");
    }

    return final_result;
}


bdd_vector add_bdd_vector(bdd_vector a, bdd_vector b)
{
	bdd_vector result;
	bdd carry, sum;
	bdd temp1, temp2, temp3;
	int aindex, bindex, rindex;

/*printf("DEBUG--Enter add_bdd_vector, a->size = %d, b->size = %d.\n",a->size, b->size);*/
	result = (bdd_vector)malloc(sizeof(*result));
	result->size = (a->size > b->size)?a->size:b->size;
	result->bits = (bdd *)calloc(result->size,sizeof(bdd));

	carry = bdd_zero(gbm);
	for (aindex = a->size-1, bindex = b->size-1, rindex = result->size-1;
	     rindex >= 0;
	     aindex--, bindex--, rindex--) {

	    if (aindex<0) {
		sum = bdd_xor(gbm,b->bits[bindex], carry);
		temp1 = bdd_and(gbm,b->bits[bindex],carry);
		bdd_free(gbm,carry);
		carry = temp1;
	    } else if (bindex<0) {
		sum = bdd_xor(gbm,a->bits[aindex], carry);
		temp1 = bdd_and(gbm,a->bits[aindex],carry);
		bdd_free(gbm,carry);
		carry = temp1;
	    } else {
		temp1 = bdd_xor(gbm,a->bits[aindex],b->bits[bindex]);
		sum = bdd_xor(gbm,temp1, carry);
		bdd_free(gbm,temp1);
		temp1 = bdd_and(gbm,a->bits[aindex],b->bits[bindex]);
		temp2 = bdd_or(gbm,a->bits[aindex],b->bits[bindex]);
		temp3 = bdd_ite(gbm,carry,temp2,temp1);
		bdd_free(gbm,temp1);
		bdd_free(gbm,temp2);
		bdd_free(gbm,carry);
		carry = temp3;
	    }

	    result->bits[rindex] = sum;
	}
	bdd_free(gbm,carry);

	return result;
}

bdd_vector eval_bit_vector_add_with_bfv(bool_func_vec old,
					    bit_vector b)
	/* Non-associative add.  See eval_bit_vector_add() in evereval.c. */
{
	bit_vector_list vec;
	bdd_vector sum, x, new;

/*printf("DEBUG--Enter eval_bit_vector_add_with_bfv.\n");*/
	vec = b->vec_list1;
	sum = (bdd_vector)malloc(sizeof(*sum));
	sum->size = 0;
	sum->bits = (bdd *)malloc(0);
	while (vec!=NULL) {
		x = eval_bit_vector_with_bfv(old,vec->data);
		new = add_bdd_vector(sum,x);
		free_bdd_vector(sum);
		free_bdd_vector(x);
		sum = new;

		vec = vec->next;
	}

	return sum;
}

bdd_vector eval_bit_vector_sub_with_bfv(bool_func_vec old,
					    bit_vector x)
{
    bdd_vector a = eval_bit_vector_with_bfv(old,x->vector1);
    bdd_vector b = eval_bit_vector_with_bfv(old,x->vector2);
    bdd_vector result;

    bdd borrow, difference;
    bdd temp1, temp2, temp3;
    int aindex, bindex, rindex;

    result = (bdd_vector)malloc(sizeof(*result));
    result->size = (a->size > b->size)? a->size : b->size;
    result->bits = (bdd *)calloc(result->size,sizeof(bdd));

    borrow = bdd_zero(gbm);	/* Assume no initial borrow. */
    for (aindex = a->size-1, bindex = b->size-1, rindex = result->size-1;
	 rindex >= 0;
	 aindex--, bindex--, rindex--) {

	if (aindex<0) {
            difference = bdd_xor(gbm,b->bits[bindex], borrow);
            temp1 = bdd_or(gbm,b->bits[bindex],borrow);
            bdd_free(gbm,borrow);
            borrow = temp1;
	} else if (bindex<0) {
            difference = bdd_xor(gbm,a->bits[aindex], borrow);
            temp1 = bdd_ite(gbm,a->bits[aindex],
				bdd_zero(gbm),
				borrow);
            bdd_free(gbm,borrow);
            borrow = temp1;
	} else {
            temp1 = bdd_xor(gbm,a->bits[aindex],b->bits[bindex]);
            difference = bdd_xor(gbm,temp1, borrow);
            bdd_free(gbm,temp1);
            temp1 = bdd_and(gbm,borrow,b->bits[bindex]);
            temp2 = bdd_or(gbm,borrow,b->bits[bindex]);
            temp3 = bdd_ite(gbm,a->bits[aindex],temp1,temp2);
            bdd_free(gbm,temp1);
            bdd_free(gbm,temp2);
            bdd_free(gbm,borrow);
            borrow = temp3;
	}

	result->bits[rindex] = difference;
    }
    bdd_free(gbm,borrow);

    free_bdd_vector(a);	/* AJH 3/11/1993 */
    free_bdd_vector(b);

    return result;
}

bdd_vector eval_bit_vector_bif_with_bfv(bool_func_vec old,
						bit_vector x)
{
    int i, j, k;
    bdd cond = eval_prop_with_bfv(old,x->prop1);
    bdd_vector thenpart = eval_bit_vector_with_bfv(old,x->vector1);
    bdd_vector elsepart = eval_bit_vector_with_bfv(old,x->vector2);
    bdd temp1, temp2;
    bdd_vector result;

    result = (bdd_vector)malloc(sizeof(*result));
    result->size = (thenpart->size > elsepart->size)?
			thenpart->size :
			elsepart->size;
    result->bits = (bdd *)calloc(result->size,sizeof(bdd));

    for (i=thenpart->size-1, j=elsepart->size-1, k=result->size-1;
            ((i>=0) || (j>=0));
            i--, j--, k--) {
        temp1 = (i>=0) ? thenpart->bits[i] : bdd_zero(gbm);
        temp2 = (j>=0) ? elsepart->bits[j] : bdd_zero(gbm);
        result->bits[k] = bdd_ite(gbm,cond,temp1,temp2);
    }

    free_bdd_vector(thenpart);	/* AJH 3/13/93 */
    free_bdd_vector(elsepart);
    bdd_free(gbm,cond);

    return result;
}

bdd eval_prop_with_bfv(bool_func_vec old, uneval_prop p)
{
    bdd bdd_result;
    bdd temp1, temp2;
    prop_list tp;

    switch (p->type) {
    case AND:
	tp = p->prop_list1;
	bdd_result = bdd_one(gbm);
	while (tp!=NULL) {
	    temp1 = eval_prop_with_bfv(old,tp->data);
	    temp2 = bdd_and(gbm,bdd_result,temp1);
	    bdd_free(gbm,bdd_result);
	    bdd_free(gbm,temp1);
	    bdd_result = temp2;
	    tp = tp->next;
	}
	break;
    case OR:
	tp = p->prop_list1;
	bdd_result = bdd_zero(gbm);
	while (tp!=NULL) {
	    temp1 = eval_prop_with_bfv(old,tp->data);
	    temp2 = bdd_or(gbm,bdd_result,temp1);
	    bdd_free(gbm,bdd_result);
	    bdd_free(gbm,temp1);
	    bdd_result = temp2;
	    tp = tp->next;
	}
	break;
    case NOT:
	temp1 = eval_prop_with_bfv(old,p->prop1);
	bdd_result = bdd_not(gbm,temp1);
	bdd_free(gbm,temp1);
	break;
    case EQ:
	bdd_result = bit_vector_eq_with_bfv(old,p->vector1,p->vector2);
	break;
    default:
	fprintf(stderr,"Error:  Unimplemented operator (%d) in eval_prop_with_bfv().\n",p->type);
	fflush(stderr);
    }

    return bdd_result;
}


bdd_vector eval_bit_vector_with_bfv(bool_func_vec old, bit_vector b)
{
    int i;
    bdd_vector result;

    if (b->type == EVAL) {
	/* It's a constant. */
	result = (bdd_vector)malloc(sizeof(*result));
	result->size = b->size;
	result->bits = (bdd *)calloc(result->size,sizeof(bdd));
	for (i=0; i < result->size; i++)
	    result->bits[i] = bdd_identity(gbm,b->bit_list[i]);
	return result;
    }

    switch (b->type) {
    case DEFVAR:	/* Variable access */
	result = eval_bit_vector_var_access_with_bfv(old,b);
	break;
    case VECTOR:
	result = eval_bit_vector_vectorize_with_bfv(old,b->vectorize_list1);
	break;
    case ADD:
	result = eval_bit_vector_add_with_bfv(old,b);
	break;
    case SUB:
	result = eval_bit_vector_sub_with_bfv(old,b);
	break;
    case BIF:
	result = eval_bit_vector_bif_with_bfv(old,b);
	break;
    default:
	fprintf(stderr,"Error:  Unimplemented operator (%d) in eval_bit_vector_with_bfv().\n",b->type);
	fflush(stderr);
    }

    return result;
}




/* BFV Routines */

bool_func_vec build_initial_bfv()
{
    string_list name_list = global_variable_names;
    variable_desig temp_desig;
    bool_func_vec result;
    int i, j;

    result = (bool_func_vec)calloc(bfv_length,sizeof(bdd));
    i = 0;
    while (name_list!=NULL) {
	temp_desig = get_variable_descriptor(name_list->data);

	/* Debugging sanity check. */
	if (i != temp_desig->bfv_offset) {
	    fprintf(stderr,"Internal Error!!!:  bfv_offset incorrect in build_initial_bfv().\n");
	    fflush(stderr);
	}

	for (j=1; j < temp_desig->bitcount; j+=3) {	/* curent phase only*/
	    result[i+(j/3)] = bdd_identity(gbm,temp_desig->bits[j]);
	}
	i += temp_desig->bitcount/3;
	free(temp_desig);
	name_list = name_list->next;
    }
    return result;
}

void free_bfv(bool_func_vec x)
{
	int i;

	for (i=0; i<bfv_length; i++) {
		bdd_free(gbm,x[i]);
	}
	free(x);
}

bool_func_vec copy_bfv(bool_func_vec x)
{
	bool_func_vec result;
	int i;

	result = (bool_func_vec)calloc(bfv_length,sizeof(bdd));
	for (i=0; i<bfv_length; i++) result[i] = bdd_identity(gbm,x[i]);
	return result;
}


void becomes_bfv_step(bool_func_vec old,	/* Current BFV */
		      bool_func_vec result,	/* Resulting new BFV */
		      type_desc remaining_type,	/* Variable type */
		      int pre_offset,		/* Offset to start of var */
		      var_accessor modchain,	/* var accessor chain */
		      bdd *rhs_bits,	/* value being assigned. */
		      bdd cond)		/* condition on assignment */
{
    if (modchain==NULL) {
	/* We've reached the variable being accessed. */
	int i = remaining_type->bdd_count/3 - 1;
	int j = 0;
	while (i>=0) {
	    result[pre_offset+i] = bdd_ite(gbm,cond,
						rhs_bits[j],
						old[pre_offset+i]);
	    j++;
	    i--;
	}
	return;
    }

    if (cond==bdd_zero(gbm)) {
	/* Short-circuit -- we know we won't assign. */
	int i = remaining_type->bdd_count/3 - 1;
	while (i>=0) {
	    result[pre_offset+i] = bdd_identity(gbm,old[pre_offset+i]);
	    i--;
	}
	return;
    }

    if (modchain->type==RECORD) {
	record_list temp = remaining_type->field_list;
	int offset = 0;
	int i;

	while (temp!=NULL) {
	    if (strcmp(temp->field_name,modchain->field_select)==0) {
		/* This is field being accessed. */
		becomes_bfv_step(old,
				 result,
				 temp->field_type,
				 pre_offset+offset,
				 modchain->next,
				 rhs_bits,
				 cond);
		offset += (temp->field_type->bdd_count/3);
		temp = temp->next;
	    } else {
		/* This field stays constant. */
		for (i=0; i < temp->field_type->bdd_count/3; i++) {
		    result[pre_offset+offset+i] =
			bdd_identity(gbm,old[pre_offset+offset+i]);
		}
		offset += (temp->field_type->bdd_count/3);
		temp = temp->next;
	    }
	}
	return;
    }

    if (modchain->type==ARRAY) {
	bdd this_is_it;
	bdd new_cond;
	int i;
	int lb = remaining_type->array_lb;
	int ub = remaining_type->array_ub;
	type_desc element_type = remaining_type->subtype;

	for (i=lb; i<=ub; i++) {
	    this_is_it = bdd_eq_int_bit_vector_with_bfv(old,i,
						modchain->array_index);
	    new_cond = bdd_and(gbm,cond,this_is_it);
	    bdd_free(gbm,this_is_it);
	    becomes_bfv_step(old,
			     result,
			     element_type,
			     pre_offset+(i-lb)*element_type->bdd_count/3,
			     modchain->next,
			     rhs_bits,
			     new_cond);
	    bdd_free(gbm,new_cond);
	}

	return;
    }

    fprintf(stderr,"Internal Error!!!:  Illegal modifier (%d) in becomes_bfv_step.\n",modchain->type);
    fflush(stderr);
}

bool_func_vec becomes_bfv(bool_func_vec old,	/* Current value of variables */
			variable_desig lhs,
			bit_vector rhs)
{
    string_list name_list;
    variable_desig temp_desig;
    int i, lsize, adjuster;
    bdd_vector rhs_value;
    bdd *rhs_bits;	/* Chopped/padded copy of rhs. */
    bool_func_vec result;

    result = (bool_func_vec)calloc(bfv_length,sizeof(bdd));

    /* Evaluate right hand side of assignment. */
    rhs_value = eval_bit_vector_with_bfv(old,rhs);

    /* Make everything else constant. */
    name_list = global_variable_names;
    while (name_list!=NULL) {
	temp_desig = get_variable_descriptor(name_list->data);
	if (temp_desig->bits!=lhs->bits) { /* Not the assigned variable. */
	    for (i=0; i<(temp_desig->bitcount/3); i++) {
		result[i+temp_desig->bfv_offset] =
			bdd_identity(gbm,old[i+temp_desig->bfv_offset]);
	    }
	}
	free(temp_desig);
	name_list = name_list->next;
    }

    /* Make the left and right sides the same size. */
    lsize = var_desig_size(lhs->base_type, lhs->modifiers);
    rhs_bits = (bdd *)calloc(lsize,sizeof(bdd));
    adjuster = lsize - rhs_value->size;
    if (adjuster > 0)
	for (i=0; i<adjuster; i++) rhs_bits[i] = bdd_zero(gbm);
    for (i = (adjuster>0)?adjuster:0; i<lsize; i++)
	rhs_bits[i] = rhs_value->bits[i-adjuster];

    becomes_bfv_step(old,		/* BFV for current values */
		     result,		/* Modifies result BFV */
		     lhs->base_type,
		     lhs->bfv_offset,	/* offset */
		     lhs->modifiers,	/* var accessor chain */
		     rhs_bits,		/* what we set it equal to */
		     bdd_one(gbm)); /* condition on assignment */

    free(rhs_bits);     /* Dispose of chopped/padded copy of rhs->bit_list. */
			/* The bdd's are illegal copies, so don't need to
				free them. */

    free_bdd_vector(rhs_value);

    return result;
}


bool_func_vec build_bfv(bool_func_vec old, uneval_prop next)
    /* Builds the bfv for a deterministic (assumed) next-state relation. */
    /* Old is the bfv for the assumed current values. */
{
    if (next->type==OR) {
	fprintf(stderr,"Error:  Cannot have non-determinism at this level (line %d).\n", next->lineno);
	fflush(stderr);
	return NULL;
    }

    /* Kludgey special case to allow me to continue to use AND to
       for an eval without resorting to setting force to 1. */
    if ((next->type==AND) &&
	(next->prop_list1!=NULL) &&
	(next->prop_list1->next==NULL)) {
	return build_bfv(old,next->prop_list1->data);
    }

    if (next==current_next_eq) {
	/* No change, make copy of old. */
	return copy_bfv(old);
    }

    if (next->type==COMPOSE) {
        /* Function composition. */
	prop_list tp;
	bool_func_vec	temp1, temp2;

        tp = next->prop_list1;
        temp1 = copy_bfv(old);
        while (tp != NULL) {
	    temp2 = build_bfv(temp1,tp->data);
	    free_bfv(temp1);
	    temp1 = temp2;

            tp = tp->next;
        }

        return temp1;
    }

    if (next->type==REPEAT) {
	int i;
	bool_func_vec	temp1, temp2;

	temp1 = copy_bfv(old);
	for (i=0; i < next->count; i++) {
	    temp2 = build_bfv(temp1,next->prop1);
	    free_bfv(temp1);
	    temp1 = temp2;
	}

	return temp1;
    }

    if (next->type==IF) {
	/* Should really check to make sure condition
		is on current phase only. */
	bdd cond;
	bool_func_vec temp1, temp2, temp3;
	int i;

	cond = eval_prop_with_bfv(old,next->prop1);
	temp1 = build_bfv(old,next->prop2);	/* then part */
	temp2 = build_bfv(old,next->prop3);	/* else part */
	temp3 = (bool_func_vec)calloc(bfv_length,sizeof(bdd));

	for (i=0; i<bfv_length; i++) {
		temp3[i] = bdd_ite(gbm,cond,temp1[i],temp2[i]);
	}
	bdd_free(gbm,cond);
	free_bfv(temp1);
	free_bfv(temp2);

	return temp3;
    }

    if (next->type==BECOMES) {
	return becomes_bfv(old, next->var1, next->vector1);
    }

    fprintf(stderr,"Error:  build_bfv doesn't know how to handle operator %d (line %d).\n",next->type, next->lineno);
    fflush(stderr);
    return NULL;
}




/* MOB Routines */


mob create_empty_mob()
	/* Allocates a MOB with no BFVs. */
{
	mob x;
	
	x = (mob)malloc(sizeof(struct mob_node));
	x->count = 0;
	x->bfv = (bool_func_vec *)malloc(0);
	return x;
}

void free_mob(mob x)
{
	int i;
    
	for (i=0; i < x->count; i++) {
		free_bfv(x->bfv[i]);
	}
	free(x->bfv);
	free(x);
}

void append_to_mob(mob a, mob b)
	/* Puts the BFVs in a after the BFVs in b. */
{
	int i;

	b->bfv = (bool_func_vec *)realloc(b->bfv,
				(a->count+b->count)*sizeof(bool_func_vec));
	for (i=0; i < a->count; i++) {
		b->bfv[b->count + i] = copy_bfv(a->bfv[i]);
	}
	b->count = a->count + b->count;
}




mob build_MOB_step(uneval_prop next)
	/* Helper function for build_MOB. */
{
    prop_list tp;
    mob temp1, temp2;

/*fprintf(stderr,"Enter build_MOB_step:  type = %d.\n",next->type);*/
    fprintf(stderr,"."); fflush(stderr);	/* Show some progress. */

    /* Kludgey special case to allow me to continue to use AND to
       for an eval without resorting to setting force to 1. */
    if ((next->type==AND) &&
	(next->prop_list1!=NULL) &&
	(next->prop_list1->next==NULL)) {
	return build_MOB_step(next->prop_list1->data);
    }

    if (next->type == OR) {
        /* Unevaluated disjunction. */
        temp1 = create_empty_mob();
        tp = next->prop_list1;
        while (tp!=NULL) {
            /* Compute MOB for each disjunct. */
            temp2 = build_MOB_step(tp->data);
	    append_to_mob(temp2, temp1);
	    free_mob(temp2);

            tp = tp->next;
        }

        return temp1;
    }

    /* Cheap Hack Alert!!!!!!!!!!!! */
    /* To avoid difficulties with sets of conditional BFVs,
	I'm going to assume all the ORs occur above anything
	else in the parse tree.  (e.g. OR tree of deterministic
	next-state relations) AJH 10/17/92 */

    temp1 = create_empty_mob();
    temp1->count = 1;
    temp1->bfv = (bool_func_vec *)realloc(temp1->bfv,1*sizeof(bool_func_vec));
    temp1->bfv[0] = build_bfv(initial_bfv,next);

    return temp1;
}


/* --------- Exported Routines --------- */

void build_MOB(uneval_prop next_state_rel) /* Builds MOB for next_state_rel*/
{
    /* Free global mob if already calculated. */
    if (global_mob!=NULL) free_mob(global_mob);

    bfv_length = (global_bdd_count + global_fdv_count) / 3;

    fprintf(stderr,"Building initial BFV..."); fflush(stderr);
    initial_bfv = build_initial_bfv();
    fprintf(stderr,"Done.\nBuilding MOB..."); fflush(stderr);
    global_mob = build_MOB_step(next_state_rel);
    fprintf(stderr,"Done.\nFreeing initial BFV..."); fflush(stderr);
    free_bfv(initial_bfv);
    fprintf(stderr,"Done.\n"); fflush(stderr);
}



bdd check_MOB(bdd reached) /* Returns set of states that can
						lead to violation. */
{
    int i;
    string_list name_list;
    variable_desig temp_desig;

    if (global_mob==NULL) {
	fprintf(stderr,"Internal Error:  check_MOB() called with NULL global_mob.\n");
	fflush(stderr);
	return bdd_zero(gbm);
    }

    fprintf(stderr,"Checking MOB"); fflush(stderr);
    for (i=0; i < global_mob->count; i++) {
	fprintf(stderr,"."); fflush(stderr);
	name_list = global_variable_names;
	while (name_list!=NULL) {
	    temp_desig = get_variable_descriptor(name_list->data);
	    if (temp_desig->fdv) {
		/* Need to check this fdv against this bfv. */
		uneval_prop prop1;
		bit_vector vec1;
		bdd temp1, temp2;

		/* GROSS HACK ALERT!!!!!!!!!!!!!!!! */
		/* To handle my net example better, I'm going to assume
			that the fdv is an array, and that the func_def
			is a vectorize with one entry per array element.
			AJH 10/21/92. */
		/*temp_desig->base_type = temp_desig->base_type->subtype;
		temp_desig->bitcount = temp_desig->base_type->bdd_count;
		temp_desig->func_def = temp_desig->func_def->vectorize_list1
						->vector_data;*/
		/* End of Gross Hack (gag, barf, vommit) */

		temp_desig->phase = CUR;

		/* Make a temporary bit_vector to compute new value of fdv. */
		vec1 = (bit_vector)malloc(sizeof(*vec1));
		vec1->type = DEFVAR;	/* Code for variable access */
		vec1->var1 = temp_desig;
		vec1->is_constant = FALSE;
		vec1->formal_order = -1;
		vec1->bit_list = NULL;

		/* Make a temporary uneval_prop for testing the equality. */
		prop1 = (uneval_prop)malloc(sizeof(*prop1));
		prop1->type = EQ;
		prop1->vector1 = vec1;
		prop1->vector2 = temp_desig->func_def;

		/* Are the defined and actual next values of fdv the same? */
		temp1 = eval_prop_with_bfv(global_mob->bfv[i], prop1);
		/* Only have to be the same on reachable states. */
		temp2 = bdd_ite(gbm,reached,temp1,bdd_one(gbm));
		bdd_free(gbm,temp1);
		if (temp2!=bdd_one(gbm)) {
		    /* We have a violation! */
		    temp1 = bdd_not(gbm,temp2);	/* Set of violating states. */
		    bdd_free(gbm,temp2);
		    free(prop1);	/* Dispose of temporaries. */
		    free(vec1);
		    free(temp_desig);
		    fprintf(stderr,"\n"); fflush(stderr);
		    return temp1;	/* Bail out with counter-example. */
		}

		/* This fdv checks ok. */
		bdd_free(gbm,temp2);

		free(prop1);
		free(vec1);
	    }
	    free(temp_desig);
	    name_list = name_list->next;
	}
    }
    fprintf(stderr,"\n"); fflush(stderr);
    return bdd_zero(gbm);
}
