#include <sys/types.h>
#include <sys/times.h>
#include "ever.h"
#include "y.tab.h"

char *memory_zero;
char *last_mem;

void reset_memory_meter()
{
    char *sbrk();

    memory_zero = sbrk(0);	/* Initialize the zero address. */
    last_mem = memory_zero;
}

long read_memory_meter()
{
    char *sbrk();

    /* Return additional bytes of heap used. */
    last_mem = sbrk(0);
    return (long) last_mem - (long) memory_zero;
}

int memory_grew()
{
	char *sbrk();

	return sbrk(0)!=last_mem;
}

int user_cpu_time()
{
    struct tms buffer;

    times(&buffer);
    return buffer.tms_utime;
}


/* Used to hold bdds for bdd_size. */
bdd	*countbuf;
int	countbufsize;

void count_bdd_nodes_step(uneval_prop p)
	/* Helper function to do the recursive search. */
{
    prop_list tempp1;

    if (p->bdd!=NULL) {
	countbuf[countbufsize] = p->bdd;
	countbufsize++;
	countbuf = (bdd *)realloc(countbuf,(countbufsize+1)*sizeof(bdd));
	countbuf[countbufsize] = NULL;
	return;
    }

    switch (p->type) {
    case AND:		/* n-ary */
    case OR:
    case COMPOSE:
        tempp1 = p->prop_list1;
        while (tempp1!=NULL) {
	    count_bdd_nodes_step(tempp1->data);
            tempp1 = tempp1->next;
        }
        break;
    case NOT:		/* unary */
	count_bdd_nodes_step(p->prop1);
        break;
    case XOR:		/* binary */
    case EQUIV:
    case IMPLIES:
	count_bdd_nodes_step(p->prop1);
	count_bdd_nodes_step(p->prop2);
        break;
    case IF:		/* ternary */
	count_bdd_nodes_step(p->prop1);
	count_bdd_nodes_step(p->prop2);
	count_bdd_nodes_step(p->prop3);
        break;
    case GT:		/* bit vector arguments */
    case GE:
    case EQ:
    case LE:
    case LT:
	printf("Error counting BDD nodes (relop)!\n"); fflush(stdout);
        break;
    case HIDEPREV:
    case HIDECUR:
    case HIDENEXT:
	printf("Error counting BDD nodes (hide)!\n"); fflush(stdout);
        break;
    case BECOMES:
	printf("Error counting BDD nodes (becomes)!\n"); fflush(stdout);
        break;
    case CONSTRAIN:
	printf("Error counting BDD nodes (constrain)!\n"); fflush(stdout);
        break;
    case FORALL:
    case EXISTS:
    case REACHABLE:
    case REACHGOAL:
    case AG:
    case EG:
    case AF:
    case EF:
    case AX:
    case EX:
    case AU:
    case EU:
    case SIMREL:
    case ANDCOMPRESS:
    case ORCOMPRESS:
    default:
	printf("Error counting BDD nodes (unimplemented)!\n"); fflush(stdout);
        yyerror("Evaluator Error:  Unimplemented boolean operator.");
        fprintf(stderr,"-----Operator type = %d.\n",p->type);
    }

    return;
}

long count_bdd_nodes(uneval_prop p)
	/* Recursively search an uneval_prop and compute the total
	   number of BDD nodes it uses.  This result correctly deals
	   with sharing (thanks to bdd_size()), but currently does
	   not handle unevaluated bit vectors. */
{
    long value;

    countbuf = (bdd *)malloc(sizeof(bdd));
    countbuf[0] = 0;
    countbufsize = 0;
    count_bdd_nodes_step(p);
    value = bdd_size_multiple(gbm,countbuf,1);
    free(countbuf);
    return value;
}


/* Count Satisfying Assignments */

/* Not used anymore.  Nov 19, 1992 AJH
#define BITSPERINT 32

int int_enumerate_reachable_states(bdd reachable,
				   int unconstrained_bits)
{
    bdd_terminal_option root = bdd_terminal(reachable);
    bdd temp;
    int count;

    if (unconstrained_bits<0)
        fprintf(stderr,"Warning:  Underflow while counting states.\n");

    if (root==BDD_IS_ONE) {
        if (unconstrained_bits>BITSPERINT) {
            fprintf(stderr,"Warning:  Overflow while counting states.  2^%d states not counted.\n",unconstrained_bits);
            return 0;
        } else return 1 << unconstrained_bits;
    } else if (root==BDD_IS_ZERO) {
        return 0;
    } else if ((root==BDD_POSVAR) || (root==BDD_NEGVAR)) {
        if (unconstrained_bits>BITSPERINT) {
            fprintf(stderr,"Warning:  Overflow while counting states.  2^%d states not counted.\n",unconstrained_bits);
            return 0;
        } else return 1 << (unconstrained_bits-1);
    } else {
	temp = bdd_then(reachable);
	count = int_enumerate_reachable_states(temp,unconstrained_bits-1);
	bdd_free(temp);
	temp = bdd_else(reachable);
	count += int_enumerate_reachable_states(temp,unconstrained_bits-1);
	bdd_free(temp);
	return count;
    }
}
*/

double enumerate_reachable_states(bdd reachable,
				  int unconstrained_bits)
{
    double exp(), log();

    if (unconstrained_bits<0)
        fprintf(stderr,"Warning:  Underflow while counting states.\n");

    return bdd_satisfying_fraction(gbm, reachable) *
		exp((double) unconstrained_bits * log(2.0));
}


/* Helper functions for bdd_naming_fn_full */

int find_variable_offset(variable_desig temp_desig, bdd bdd_var)
	/* Returns -1 if not in list. */
{
	int result = -1;
	int i;

	for (i=0; i < temp_desig->bitcount; i++) {
		if (temp_desig->bits[i] == bdd_var) {
			result = i;
			break;
		}
	}

	return result;
}

void find_qualifiers_for_offset(type_desc remaining_type, int offset, char *out)
{
    char scratch[80];

    if (remaining_type->type==BITS) {
	if (offset%3==0) sprintf(scratch,".%d^p",offset/3);
	if (offset%3==1) sprintf(scratch,".%d^c",offset/3);
	if (offset%3==2) sprintf(scratch,".%d^n",offset/3);

	strcat(out,scratch);
    } else if (remaining_type->type==ARRAY) {
	sprintf(scratch,"[%d]",
		offset/(remaining_type->subtype->bdd_count) +
			remaining_type->array_lb);

	strcat(out,scratch);

	find_qualifiers_for_offset(remaining_type->subtype,
				    offset%(remaining_type->subtype->bdd_count),
				    out);
    } else if (remaining_type->type==RECORD) {
	record_list fields;

	fields = remaining_type->field_list;
	while (fields!=NULL) {
	    if (offset < fields->field_type->bdd_count) break;

	    offset -= fields->field_type->bdd_count;
	    fields = fields->next;
	}

	if (fields==NULL) {
	    sprintf(scratch,".Illegal Field, Remaining Offset %d",offset);
	    strcat(out,scratch);
	} else {
	    sprintf(scratch,".%s",fields->field_name);
	    strcat(out,scratch);
	    find_qualifiers_for_offset(fields->field_type, offset, out);
	}
    } else {
	fprintf(stderr,"Internal error in find_qualifiers_for_offset!\n");
    }
}

/* Generates the variable name for a given BDD variable. */
char *bdd_naming_fn_full(bdd_manager bddm,
			 bdd bdd_var,
			 pointer env)
{
    static char	fullname[256];

    string_list name_list;
    variable_desig temp_desig;
    int var_index;

    /* Search the global_variable_names for the one with the given bdd_var. */
    name_list = global_variable_names;
    while (name_list!=NULL) {
	temp_desig = get_variable_descriptor(name_list->data);

	var_index = find_variable_offset(temp_desig,bdd_var);

	if (var_index > -1) break;

	free(temp_desig);
	name_list = name_list->next;
    }

    /* At this point, if name_list!=NULL, then name_list->data is the
       global variable containing bdd_var, and temp_desig->bits[var_index]
       is equal to bdd_var. */

    if (name_list==NULL) {
	return NULL;	/* Something's wrong.  We don't have a name for it. */
    }

    /* Figure out the name of the given bdd_var. */
    strcpy(fullname,name_list->data);	/* Get the base name. */

    find_qualifiers_for_offset(temp_desig->base_type,var_index,fullname);

    free(temp_desig);

    return fullname;
}


/* Prints the support set for a BDD */

void print_bdd_support(FILE *fp, char *(*naming_fn)(), bdd x)
{
    bdd *support_array, *temp;

    support_array = (bdd *)malloc(((long)global_bdd_count/3+1)*sizeof(bdd));

    bdd_support(gbm,x,support_array);

    temp = support_array;
    while (*temp != 0) {
	bdd_print_bdd(gbm, *temp, bdd_naming_fn_full, bdd_terminal_id_fn_none,
			NULL, fp);
	temp++;
    }

    free(support_array);
}
