/* Copyright 1991 by Alan J. Hu. */
/* All rights reserved. */

/* EVER Symbol Table Version 0.0 */
/* By Alan J. Hu */
/* August 6, 1991 */

/* ************************ CHANGE LOG **************************** */
/*
8 September 1991 AJH -- Changed bit vectors to be an unevaluate type.
8 November 1991 AJH -- Added call to interleave_variables().
31 August 1992	AJH -- Added code to allow backwards variable ordering.
13 October 1992 AJH -- Added code to handle functionally dependent variables.
19 November 1992 AJH -- Converted to David Long's BDD Package
14 May 1993 AJH -- Fixed careless memory leak in create_new_var()
*/
/* **************************************************************** */

#include "ever.h"

string_list	global_variable_names=NULL;	/* List of all variables. */


/* ***** Internal Stuff ***** */

struct var_desc_node	{
				type_desc	type_decl;
				int		fdv;	/* TRUE iff it's
							   a functionally
							   dependent variable */
				int		bfv_offset;
					/* Offset to start of var in bfv */
				bit_vector	func_def; /* functional
								dependency */
				int		bitcount;
				bdd		*bits;
			};
typedef struct var_desc_node *var_desc;

/* decl types */
#define SCOPEMARKER 0
#define TYPEDECL 1
#define VARDECL 2
#define PROPDECL 3
#define PREDDECL 4
#define VECDECL 5
#define FORMALDECL 6
struct ste	{
			char		*name;
			int		decl_type;
			type_desc	type_decl;
			var_desc	var_decl;
			uneval_prop	prop_decl;
			predicate	pred_decl;
			bit_vector	vec_decl;
			int		formal_order;
			struct ste	*next;
		};

struct ste	*symtab = NULL;		/* The Symbol Table */


var_desc create_new_var(type_desc its_type)
{
	var_desc	new_var;
	int		i;
	/* to compute phase equality bdds */
	bdd		temp_pc, temp_cn, temp_pn;
	bdd		temp1, temp2;

	new_var = (var_desc)malloc(sizeof(*new_var));
	new_var->fdv = FALSE;	/* Not a functionally dependent var */
	new_var->bfv_offset = (global_bdd_count+global_fdv_count)/3;
	new_var->func_def = NULL;
	new_var->type_decl = its_type;
	new_var->bitcount = its_type->bdd_count;
	new_var->bits = (bdd *)calloc(new_var->bitcount,
						sizeof(bdd));

	for (i=0; i < its_type->bdd_count; i++) {
	    /* Puts the new variable at end. */
	    new_var->bits[i] = bdd_new_var_last(gbm);
	    /* Alternative, puts the new variable at beginning. */
	    /*new_var->bits[i] = bdd_new_var_first(gbm);*/
	}

	/* Handle variable interleaving. */
	interleave_variables(new_var->bits, its_type);

	global_bdd_count += its_type->bdd_count;

	if (its_type->bdd_count % 3 != 0) {
	    yyerror("Internal Error!!!  bdd_count not divisible by 3 in create_new_var.");
	    return new_var;
	}


	/* Updates the phase equality propositions. */
	/* Also prev_vars, cur_vars, and next_vars (for quantification). */
	prev_vars = (bdd *)realloc(prev_vars,
					(global_bdd_count/3+1)*
					 sizeof(bdd));
	prev_vars[global_bdd_count/3] = 0;	/* Terminator */
	cur_vars = (bdd *)realloc(cur_vars,
					(global_bdd_count/3+1)*
					 sizeof(bdd));
	cur_vars[global_bdd_count/3] = 0;	/* Terminator */
	next_vars = (bdd *)realloc(next_vars,
					(global_bdd_count/3+1)*
					 sizeof(bdd));
	next_vars[global_bdd_count/3] = 0;	/* Terminator */

	temp_pc = bdd_one(gbm);	/* These will become the phase equality bdd */
	temp_cn = bdd_one(gbm);
	temp_pn = bdd_one(gbm);
	for (i=0; i < its_type->bdd_count/3; i++) {
	    /* previous_current_eq */
	    temp1 = bdd_xnor(gbm,new_var->bits[3*i],new_var->bits[3*i+1]);
	    temp2 = bdd_and(gbm,temp_pc,temp1);
	    bdd_free(gbm,temp_pc);
	    bdd_free(gbm,temp1);
	    temp_pc = temp2;

	    /* current_next_eq */
	    temp1 = bdd_xnor(gbm,new_var->bits[3*i+1],new_var->bits[3*i+2]);
	    temp2 = bdd_and(gbm,temp_cn,temp1);
	    bdd_free(gbm,temp_cn);
	    bdd_free(gbm,temp1);
	    temp_cn = temp2;

	    /* previous_next_eq */
	    temp1 = bdd_xnor(gbm,new_var->bits[3*i],new_var->bits[3*i+2]);
	    temp2 = bdd_and(gbm,temp_pn,temp1);
	    bdd_free(gbm,temp_pn);
	    bdd_free(gbm,temp1);
	    temp_pn = temp2;

	    prev_vars[(global_bdd_count-its_type->bdd_count)/3 + i] =
							new_var->bits[3*i];
	    cur_vars[(global_bdd_count-its_type->bdd_count)/3 + i] =
							new_var->bits[3*i+1];
	    next_vars[(global_bdd_count-its_type->bdd_count)/3 + i] =
							new_var->bits[3*i+2];
	}
	temp1 = bdd_and(gbm,previous_current_eq->bdd, temp_pc);
	bdd_free(gbm,previous_current_eq->bdd);
	bdd_free(gbm,temp_pc);
	previous_current_eq->bdd = temp1;

	temp1 = bdd_and(gbm,current_next_eq->bdd, temp_cn);
	bdd_free(gbm,current_next_eq->bdd);
	bdd_free(gbm,temp_cn);
	current_next_eq->bdd = temp1;

	temp1 = bdd_and(gbm,previous_next_eq->bdd, temp_pn);
	bdd_free(gbm,previous_next_eq->bdd);
	bdd_free(gbm,temp_pn);
	previous_next_eq->bdd = temp1;

	/* Update var associations */
	bdd_free_assoc(gbm,prev_vars_assoc);
	prev_vars_assoc = bdd_new_assoc(gbm,prev_vars,0);
	bdd_free_assoc(gbm,cur_vars_assoc);
	cur_vars_assoc = bdd_new_assoc(gbm,cur_vars,0);
	bdd_free_assoc(gbm,next_vars_assoc);
	next_vars_assoc = bdd_new_assoc(gbm,next_vars,0);

	return new_var;
}

var_desc create_new_dependent_var(type_desc its_type, bit_vector definition)
{
	var_desc	new_var;
	int		i;

	new_var = (var_desc)malloc(sizeof(*new_var));
	new_var->fdv = TRUE;	/* Functionally dependent var */
	new_var->bfv_offset = (global_bdd_count+global_fdv_count)/3;
	new_var->func_def = definition;
	new_var->type_decl = its_type;
	new_var->bitcount = its_type->bdd_count;
	new_var->bits = (bdd *)calloc(new_var->bitcount,
						sizeof(bdd));
			    /* MSB comes at end. */

	/* Evaluate the functional definition and create the variable. */
	evaluate_bit_vector(definition, FALSE);	/* Modifies definition */
	for (i=0; i < its_type->bdd_count; i++) {
	    if (i%3 == 1) {
		/* Current phase */
			/* i/3 is the bit number.  i/3=0 is LSB for variable*/
			/* LSB for bit_vector comes LAST, not first */
		int temp = definition->size - 1 - i/3; /* swaps LSB MSB */
		if (temp >= 0) {
		    new_var->bits[i] = bdd_identity(gbm,
					definition->bit_list[temp]);
		} else {
		    new_var->bits[i] = bdd_zero(gbm);
		}
	    } else {
		/* Other phases */
		/* What to do?  I don't think these guys should ever
			be accessed.  --AJH Oct 14, 1992
			I guess I really should make phase-changed
			versions of the bfv's.  FIX THIS */
		new_var->bits[i] = NULL;
	    }
	}

	global_fdv_count += its_type->bdd_count;

	if (its_type->bdd_count % 3 != 0) {
	    yyerror("Internal Error!!!  bdd_count not divisible by 3 in create_new_dependent_var.");
	    return new_var;
	}


	return new_var;
}

variable_desig make_var_desig(var_desc declaration)
{
	variable_desig	new_desig;

	new_desig = (variable_desig)malloc(sizeof(*new_desig));
	new_desig->base_type = declaration->type_decl;
	new_desig->bitcount = declaration->bitcount;
	new_desig->fdv = declaration->fdv;	/* Is it dependent variable?*/
	new_desig->func_def = declaration->func_def;
	new_desig->bfv_offset = declaration->bfv_offset;
	new_desig->bits = declaration->bits;
	new_desig->modifiers = NULL;
	return new_desig;
}


bit_vector make_formal_vec(int which_arg)
{
	bit_vector	new_vec;

	new_vec = (bit_vector)malloc(sizeof(*new_vec));
	new_vec->type = -1;
	new_vec->size = -1;
	new_vec->formal_order = which_arg; /* Flag that this is a formal. */
	new_vec->bit_list = NULL;

	return new_vec;
}


struct ste *find_ste_by_name(char *name)
{
	struct ste *x;

	x = symtab;

	while (x!=NULL) {
		if (strcmp(x->name,name)==0) return x;
		x = x->next;
	}

	return NULL;	/* Search failed. */
}

/* ***** Exported Routines ***** */

/* FIX THIS -- In all the define routines, if the definition is vacuous,
	don't make the definition.  This is for error handling. */

void define_type(char *id, type_desc its_type)
{
	struct ste *new_entry;

	new_entry = (struct ste *)malloc(sizeof(struct ste));
	new_entry->name = id;
	new_entry->decl_type = TYPEDECL;
	new_entry->type_decl = its_type;

	new_entry->next = symtab;
	symtab = new_entry;
}

void define_variable(char *id, type_desc its_type)
{
	struct ste *new_entry;

	new_entry = (struct ste *)malloc(sizeof(struct ste));
	new_entry->name = id;
	new_entry->decl_type = VARDECL;
	new_entry->var_decl = create_new_var(its_type);

	global_variable_names=append_to_string_list(id, global_variable_names);

	new_entry->next = symtab;
	symtab = new_entry;
}

void define_dependent_variable(char *id,
				type_desc its_type,
				bit_vector definition)
{
	struct ste *new_entry;

	new_entry = (struct ste *)malloc(sizeof(struct ste));
	new_entry->name = id;
	new_entry->decl_type = VARDECL;
	new_entry->var_decl = create_new_dependent_var(its_type,definition);

	global_variable_names=append_to_string_list(id, global_variable_names);

	new_entry->next = symtab;
	symtab = new_entry;
}

void define_proposition(char *id, uneval_prop value)
{
	struct ste *new_entry;

	new_entry = (struct ste *)malloc(sizeof(struct ste));
	new_entry->name = id;
	new_entry->decl_type = PROPDECL;
	new_entry->prop_decl = value;

	new_entry->next = symtab;
	symtab = new_entry;
}

void define_predicate(char *id, predicate value)
{
	struct ste *new_entry;

	new_entry = (struct ste *)malloc(sizeof(struct ste));
	new_entry->name = id;
	new_entry->decl_type = PREDDECL;
	new_entry->pred_decl = value;

	new_entry->next = symtab;
	symtab = new_entry;
}

void define_bitvector(char *id, bit_vector value)
{
	struct ste *new_entry;

	new_entry = (struct ste *)malloc(sizeof(struct ste));
	new_entry->name = id;
	new_entry->decl_type = VECDECL;
	new_entry->vec_decl = value;

	new_entry->next = symtab;
	symtab = new_entry;
}


type_desc	get_type_descriptor(char *id)
{
	struct ste *x;

	x = find_ste_by_name(id);

	if (x==NULL) {
		yyerror("Undeclared type.");
		return NULL;	/* May need to change this. */
	}

	if (x->decl_type!=TYPEDECL) {
		yyerror("Type expected.");
		return NULL;	/* May need to change this. */
	}

	return x->type_decl;
}

variable_desig	get_variable_descriptor(char *id)
{
	struct ste *x;

	x = find_ste_by_name(id);

	if (x==NULL) {
		yyerror("Undeclared variable.");
		return NULL;	/* May need to change this. */
	}

	if (x->decl_type!=VARDECL) {
		yyerror("Variable expected.");
		return NULL;	/* May need to change this. */
	}

	return make_var_desig(x->var_decl);
}

uneval_prop	get_proposition_descriptor(char *id)
{
	struct ste *x;

	x = find_ste_by_name(id);

	if (x==NULL) {
		yyerror("Undeclared proposition identifier.");
		return NULL;	/* May need to change this. */
	}

	if (x->decl_type==PROPDECL) {
		return x->prop_decl;
	}

	yyerror("Proposition identifier expected.");
	return NULL;	/* May need to change this. */
}

predicate	get_predicate_descriptor(char *id)
{
	struct ste *x;

	x = find_ste_by_name(id);

	if (x==NULL) {
		yyerror("Undeclared predicate identifier.");
		return NULL;	/* May need to change this. */
	}

	if (x->decl_type!=PREDDECL) {
		yyerror("Predicate identifier expected.");
		return NULL;	/* May need to change this. */
	}

	return x->pred_decl;
}

bit_vector	get_bit_vector(char *id)
{
	struct ste *x;

	x = find_ste_by_name(id);

	if (x==NULL) {
		yyerror("Undeclared bit vector identifier.");
		return NULL;	/* May need to change this. */
	}

	if (x->decl_type==FORMALDECL) {
		return make_formal_vec(x->formal_order);
	}

	if (x->decl_type!=VECDECL) {
		yyerror("Bit vector identifier expected.");
		return NULL;	/* May need to change this. */
	}

	return x->vec_decl;
}


void	push_formals(id_list formal_list)
{
	struct ste *new_entry;
	int order = 0;

	new_entry = (struct ste *)malloc(sizeof(struct ste));
	new_entry->name = "Plate of Shrimp";	/* Arbitrary unique string. */
	new_entry->decl_type = SCOPEMARKER;

	new_entry->next = symtab;
	symtab = new_entry;

	while (formal_list != NULL) {
	    new_entry = (struct ste *)malloc(sizeof(struct ste));
	    new_entry->name = formal_list->data;
	    new_entry->decl_type = FORMALDECL;
	    new_entry->formal_order = order++;

	    new_entry->next = symtab;
	    symtab = new_entry;
	    formal_list = formal_list->next;
	}
}

void	pop_scope()
{
	struct ste *tos;

	if (symtab == NULL) return;

	tos = symtab;

	if (tos->decl_type == SCOPEMARKER) {
		symtab = tos->next;
		free(tos);
		return;
	}

	/* Free the current ste. */
	/* FIX THIS -- Should free internal stuff of ste first. */
	symtab = tos->next;
	free(tos);

	pop_scope();	/* Continue the search for the SCOPEMARKER. */
	return;
}
