-- This is a sample Ever program.

-- This is the simple network example presented in my thesis and several
-- papers.  It illustrates a simple, common situation that causes normal
-- BDD algorithms to blow up.

-- Network Model
-- Alan J. Hu

-- This one uses a valid message bit instead of the message count.

-- Array of identical processors, each with an outstanding message count.
-- Network of messages, consisting of src and req/ack.
-- Processors send out messages and wait for replies, keeping count of
-- the number of messages they've sent.

-- Modified to use CurNextEq instead of FALSE --AJH 10/12/92


deftype integer (bits 4);
deftype boolean (bits 1 "False" "True");
deftype message (record msg_type (bits 1 "Ack" "Req") src integer);

defvar net (array 0 5 (record valid boolean data message));
defvar procs (array 0 5 (record cnt integer));

defpred Proc_Send_i (p i)	-- Proc p tries sending message i.
	(if (eq net[i].valid^c 0)		-- Is it available?
	    -- If yes, then send a request out.
	    (compose
		(becomes net[i].valid^n 1)	-- Mark as valid.
		(becomes net[i].data.msg_type^n 1) -- Mark as a request.
		(becomes net[i].data.src^n p)	-- Set return address.
		--Increment outstanding message counter.
		(becomes procs[p].cnt^n (add procs[p].cnt^c 1))
	    )
	    -- Else, block.
	    CurNextEq
	);

defpred Proc_Send (p)	-- Processor p sends a message.
	(or
	    (Proc_Send_i p '4'0)    -- Processor p sends message 0 in net.
	    (Proc_Send_i p '4'1)    -- Processor p sends message 1 in net.
	    (Proc_Send_i p '4'2)    -- etc.
	    (Proc_Send_i p '4'3)
	    (Proc_Send_i p '4'4)
	    (Proc_Send_i p '4'5)
	    --(Proc_Send_i p '4'6)
	    --(Proc_Send_i p '4'7)
	    --(Proc_Send_i p '4'8)
	    --(Proc_Send_i p '4'9)
	    --(Proc_Send_i p '4'10)
	    --(Proc_Send_i p '4'11)
	    --(Proc_Send_i p '4'12)
	    --(Proc_Send_i p '4'13)
	    --(Proc_Send_i p '4'14)
	    --(Proc_Send_i p '4'15)
	);

defpred Proc_Rcv_i (p i)	-- Proc p tries receiving message i.
	(if (and (eq net[i].valid^c 1)		-- Is it valid?
		 (eq net[i].data.msg_type^c 0)	-- An acknowledge?
		 (eq net[i].data.src^c p)	-- For me?
	    )
	    -- If yes, then process it.
	    (compose
		--Decrement outstanding message counter.
		(becomes procs[p].cnt^n (sub procs[p].cnt^c 1))
		--Delete message from network.
		(becomes net[i].valid^n 0)	-- Mark as invalid.
	    )
	    -- Else, block.
	    CurNextEq
	);

defpred Proc_Rcv (p)	-- Processor p receives a message.
	(or
	    (Proc_Rcv_i p '4'0)    -- Processor p receives message 0 in net.
	    (Proc_Rcv_i p '4'1)    -- Processor p receives message 1 in net.
	    (Proc_Rcv_i p '4'2)    -- etc.
	    (Proc_Rcv_i p '4'3)
	    (Proc_Rcv_i p '4'4)
	    (Proc_Rcv_i p '4'5)
	    --(Proc_Rcv_i p '4'6)
	    --(Proc_Rcv_i p '4'7)
	    --(Proc_Rcv_i p '4'8)
	    --(Proc_Rcv_i p '4'9)
	    --(Proc_Rcv_i p '4'10)
	    --(Proc_Rcv_i p '4'11)
	    --(Proc_Rcv_i p '4'12)
	    --(Proc_Rcv_i p '4'13)
	    --(Proc_Rcv_i p '4'14)
	    --(Proc_Rcv_i p '4'15)
	);

defpred Processor (p)	-- Processor p does something.
	(or
		(Proc_Send p)	-- Either processor p sends a message,
		(Proc_Rcv p)	-- Or processor p receives a message.
	);

defpred Server (i)	-- Server serves message i in network.
	(if (and (eq net[i].valid^c 1)		-- Is it a valid message?
		(eq net[i].data.msg_type^c 1))	-- And is it a request?
	    -- Yes, handle the request.
	    (compose
		-- Code to serve the request would go here.
		-- Just bounce the message back by changing msg_type.
		(becomes net[i].data.msg_type^n 0)
	    )
	    CurNextEq	-- Else block if not a valid request.
	);

defprop StartState (and
			(eq net[0].valid^c 0)
			(eq net[1].valid^c 0)
			(eq net[2].valid^c 0)
			(eq net[3].valid^c 0)
			(eq net[4].valid^c 0)
			(eq net[5].valid^c 0)
			--(eq net[6].valid^c 0)
			--(eq net[7].valid^c 0)
			--(eq net[8].valid^c 0)
			--(eq net[9].valid^c 0)
			--(eq net[10].valid^c 0)
			--(eq net[11].valid^c 0)
			--(eq net[12].valid^c 0)
			--(eq net[13].valid^c 0)
			--(eq net[14].valid^c 0)
			--(eq net[15].valid^c 0)
			--Zero out processors, too.
			(eq procs^c 0)	-- I think this'll do it.  AJH 10/16/92
		    );

defprop NextState	(and (or
				(Processor '4'0)
				(Processor '4'1)
				(Processor '4'2)
				(Processor '4'3)
				(Processor '4'4)
				(Processor '4'5)
				--(Processor '4'6)
				--(Processor '4'7)
				--(Processor '4'8)
				--(Processor '4'9)
				--(Processor '4'10)
				--(Processor '4'11)
				--(Processor '4'12)
				--(Processor '4'13)
				--(Processor '4'14)
				--(Processor '4'15)
				(Server '4'0)
				(Server '4'1)
				(Server '4'2)
				(Server '4'3)
				(Server '4'4)
				(Server '4'5)
				--(Server '4'6)
				--(Server '4'7)
				--(Server '4'8)
				--(Server '4'9)
				--(Server '4'10)
				--(Server '4'11)
				--(Server '4'12)
				--(Server '4'13)
				--(Server '4'14)
				--(Server '4'15)
			));

--Helpers for invariant checking

defpred Count_Msg (i p) -- Is net.data[i] a valid message to/from p?
	    (and
		(eq net[i].valid^c 1)
		(eq net[i].data.src^c p)
	    );

defpred Proc_Count_Ok (i) -- Does processor i have correct count?
	    (eq
		    procs[i].cnt^c
		    (add '4'0 -- set width of sum
			(vector (Count_Msg '4'0 i))
			(vector (Count_Msg '4'1 i))
			(vector (Count_Msg '4'2 i))
			(vector (Count_Msg '4'3 i))
			(vector (Count_Msg '4'4 i))
			(vector (Count_Msg '4'5 i))
			--(vector (Count_Msg '4'6 i))
			--(vector (Count_Msg '4'7 i))
			--(vector (Count_Msg '4'8 i))
			--(vector (Count_Msg '4'9 i))
			--(vector (Count_Msg '4'10 i))
			--(vector (Count_Msg '4'11 i))
			--(vector (Count_Msg '4'12 i))
			--(vector (Count_Msg '4'13 i))
			--(vector (Count_Msg '4'14 i))
			--(vector (Count_Msg '4'15 i))
		    ) -- add
	    );

defprop Ok
	(or (and
		(Proc_Count_Ok '4'0)
		--(Proc_Count_Ok '4'1)
		--(Proc_Count_Ok '4'2)
		--(Proc_Count_Ok '4'3)
		--(Proc_Count_Ok '4'4)
		--(Proc_Count_Ok '4'5)
		--(Proc_Count_Ok '4'6)
		--(Proc_Count_Ok '4'7)
		--(Proc_Count_Ok '4'8)
		--(Proc_Count_Ok '4'9)
		--(Proc_Count_Ok '4'10)
		--(Proc_Count_Ok '4'11)
		--(Proc_Count_Ok '4'12)
		--(Proc_Count_Ok '4'13)
		--(Proc_Count_Ok '4'14)
		--(Proc_Count_Ok '4'15)
	));

--printtrace (StartState) (NextState)
				    --(not (Ok));
					--(and (eq net[5].valid^c 1)
						--(eq net[5].data.msg_type^c 1)
						--(eq net[5].data.src^c 7));
checkinvariant (StartState) (NextState) (Ok);

--end;
